#############################################################################
##
#A  cycpol.g           CHEVIE library     Jean Michel 13-9-96, 
##                                         revised 17-10-98 then feb. 2000
##
#Y  Copyright (C)  1992-1996 Lehrstuhl D f\"ur  Mathematik, RWTH Aachen,
#Y  IWR  der   Universit\"at  Heidelberg,  University  of   Galway,  and
#Y  University Paris VII.
##
# The following function was provided by Thomas Breuer in may 1999
##############################################################################
##
#F  AsRootOfUnity( <c> )
##
##  Given a cyclotomic <c>  `AsRootOfUnity' returns the rational e/n if
##  c is equal to E(n)^e, and false otherwise.
##
##############################################################################
##  Suppose that  we know  the coefficients of  $\pm E(n)^i$  w.r.t. the
##  $n$-th  Zumbroich basis  (see~"ZumbroichBasis"). These  coefficients
##  are  either all  $1$  or all  $-1$. More  precisely,  they arise  in
##  the base  conversion from  (formally) successively  multiplying $\pm
##  E(n)^i$ by $1 = -  \sum_{j=1}^{p-1} E(n)^{jn/p}$, for suitable prime
##  diviors $p$  of $n$, and then  treating the summands $\pm  E(n)^{i +
##  jn/p}$ in  the same  way until  roots in the  basis are  reached. It
##  should be noted that all roots obtained this way are distinct.
##
##  Suppose  the  above  procedure  must   be  applied  for  the  primes
##  $p_1$,  $p_2$,   \ldots,  $p_r$.   Then  $   E(n)^i$  is   equal  to
##  $(-1)^r  \sum_{j_1=1}^{p_1-1}  \cdots  \sum_{j_r=1}^{p_r-1}  E(n)^{i
##  +  \sum_{k=1}^r  j_k  n/p_k}$.  The  number  of  summands  is  $m  =
##  \prod_{k=1}^r (p_k-1)$.  Since these root  are all distinct,  we can
##  compute the  sum $s$ of  their exponents  modulo $n$ from  the known
##  coefficients of  $ E(n)^i$, and  we get $s  \equiv m (  i + r  n/2 )
##  \pmod{n}$. Either  $m =  1$ or  $m$ is  even, hence  this congruence
##  determines $ E(n)^i$ at most up to its sign.
##
##  Now suppose  that $g  = \gcd(  m, n  )$ is  nontrivial. Then  $i$ is
##  determined only modulo $n/g$, and we have to decide which of the $g$
##  possible values $i$  is. This could be done by  computing the values
##  $j_{0,p}$  for one  candidate $i$,  where  $p$ runs  over the  prime
##  divisors $p$ of $n$ for which $m$ is the product of $(p-1)$.
##
##  (Recall that each  $n$-th root of unity is of  the form $\prod_{p\in
##  P} \prod_{k_p=1}^{\nu_p-1} E(n)^{j_{k,p}n/p^{k_p+1}}$, with $j_{0,p}
##  \in \{ 0, 1, \ldots, p-1 \}$, $j_{k,2}  \in \{ 0, 1 \}$ for $k > 0$,
##  and $j_{k,p} \in  \{ -(p-1)/2, \ldots, (p-1)/2 \}$ for  $p > 2$. The
##  root is in the  $n$-th Zumbroich basis if and only  if $j_{0,2} = 0$
##  and $j_{0,p} \not= 0$ for $p > 2$.)
##
##  But note that we do not have  an easy access to the decomposition of
##  $m$ into  factors $(p-1)$,  although in  fact this  decomposition is
##  unique for $n \leq 65000$.
##
##  So the exponent is identified by  dividing off one of the candidates
##  and then  identifying the quotient,  which is  a $g$-th or  $2 g$-th
##  root of unity. (Note that $g$  is small compared to $n$. $m$ divides
##  the product  of $(p-1)$ for  prime divisors  $p$ that divide  $n$ at
##  least with  exponent $2$. The maximum  of $g$ for $n  \leq 65000$ is
##  $40$, so $2$ steps suffice for these values of $n$.)

AsRootOfUnity := function( root )

    local coeffs,   # Zumbroich basis coefficients of `root'
          n,        # conductor of `n'
          sum,      # sum of exponents with nonzero coefficients
          num,      # number of nonzero coefficients
          i,        # loop variable
          val,      # one coefficient
          coeff,    # one nonzero coefficient (either `1' or `-1')
          exp,      # candidate for the exponent
          g,        # `Gcd( n, num )'
          groot;    # root in recursion

    # Added JM 19/5/2003: reject non-roots
    if root*GaloisCyc(root,-1)<>1 then return false;fi;

    # Handle the trivial cases that `root' is an integer.
    if root = 1 then return 0;
    elif root = -1 then return 1/2;
    fi;

    # Compute the Zumbroich basis coefficients,
    # and number and sum of exponents with nonzero coefficient (mod `n').
    coeffs:= COEFFSCYC( root );
    n:= Length( coeffs );
    sum:= 0;
    num:= 0;
    for i in [ 1 .. n ] do
      val:= coeffs[i];
      if val <> 0 then
        sum:= sum + i;
        num:= num + 1;
        coeff:= val;
      fi;
    od;
    sum:= sum - num;

    # `num' is equal to `1' if and only if `root' or its negative
    # belongs to the basis.
    # (The coefficient is equal to `-1' if and only if either
    # `n' is a power of $2$ or
    # `n' is odd and `root' is a primitive $2 `n'$-th root of unity.)
    if num = 1 then
      if coeff < 0 then
        if n mod 2 = 0 then
          sum:= sum + n/2;
        else
          sum:= 2*sum + n;
          n:= 2*n;
          sum:= sum mod n;
        fi;
      fi;
      if root=E(n)^sum then return (sum mod n)/n; else return false;fi;
    fi;

    # Let $N$ be `n' if `n' is even, and equal to $2 `n'$ otherwise.
    # The exponent is determined modulo $N / \gcd( N, `num' )$.
    g:= GcdInt( n, num );
    if g = 1 then

      # If `n' and `num' are coprime then `root' is identified up to its sign.
      exp:= ( sum / num ) mod n;
      if root <> E(n)^exp then
        exp:= 2*exp + n;
        n:= 2*n;
        exp:= exp mod n;
      fi;

    elif g = 2 then

      # `n' is even, and again `root' is determined up to its sign.
      exp:= ( sum / num ) mod ( n / 2 );
      if root <> E(n)^exp then
        exp:= ( exp + n / 2 ) mod n;
      fi;

    else

      # Divide off one of the candidates.
      # The quotient is a `g'-th or $2 `g'$-th root of unity,
      # which can be identified by recursion.
      exp:= ( sum / num ) mod ( n / g );
      groot:= AsRootOfUnity( root * E(n)^(n-exp) );
      if groot=false then return false;fi;
      if n mod 2 = 1 and Denominator(groot) mod 2 = 0 then
        exp:= 2*exp;
        n:= 2*n;
      fi;
      exp:= ( exp + Numerator(groot) * n / Denominator(groot) ) mod n;

    fi;

    if root=E(n)^exp then return (exp mod n)/n; else return false;fi;
end; 

#############################################################################
##  The  rest of  this  file contains  functions  dealing with  rational
##  fractions with  all poles and zeroes  equal to 0 or  roots of unity.
##  They are represented as 'CycPol', which are used in various files of
##  the CHEVIE package whenever possible.
##
##  A CycPol is a record of the form:
##  (coeff:=c,valuation:=v,vcyc:=vec,operations:=CycPolOps)
##
##  it represents  a rational fraction in  the variable q of  the form
##    c*q^v*product(terms for the entries in the list vec)
##
##  where:
##   c  can be anything (an element of a suitable ring...)
##      This  gives a  great flexibility  of  use (e.g.  one can  accept
##      arbitrary  polynomials by  putting  the  non-cyclotomic part  in
##      coeff)...
##   vec is a  list of entries. An  entry is of the form  [k/d, m] where
##       k<d are positive  integers and m is any  integer; it represents
##       (q-E(d)^k)^m
##
##  The   entries   [k/d,m]   are   sorted   as   [d,   k/d,   m],   and
##  CycPolOps.normalize makes sure all entries have k/d different.
##
##  The advantage  of representing  polynomials which  can be  as CycPol
##  should be obvious: less storage, faster multiplication, division and
##  evaluation, and not least trivial  to print as product of cyclotomic
##  polynomials. The only  disadvantage is that addition  is not defined
##  on CycPols :)
##
##       Jean Michel 13-9-96
#############################################################################

CycPolOps:= OperationsRecord("CycPolOps");

CycPolOps.\=:=function(a,b)
  if not IsRec(a) or not IsRec(b) or not IsBound(a.operations) or
   not IsBound(b.operations) or a.operations<>b.operations then
   return false;fi;
  return a.coeff=b.coeff and  a.valuation=b.valuation and a.vcyc=b.vcyc;
end;

CycPolOps.\*:=function(a,b)
  if IsCyc(a) then return rec(coeff:=a*b.coeff,valuation:=b.valuation,
                  vcyc:=ShallowCopy(b.vcyc),operations:=CycPolOps);
  fi;
  if IsCyc(b) then return rec(coeff:=b*a.coeff,valuation:=a.valuation,
                  vcyc:=ShallowCopy(a.vcyc),operations:=CycPolOps);
  fi;
  return CycPolOps.normalize(rec(coeff:=a.coeff*b.coeff,
                        valuation:=a.valuation+b.valuation,
   vcyc:=Concatenation(a.vcyc,b.vcyc),operations:=CycPolOps));
end;

CycPolOps.\/:=function(a,b)local res;
  if IsList(a) then return List(a,x->x/b);fi;
  if IsCyc(a) then return rec(coeff:=a/b.coeff,valuation:=-b.valuation,
               vcyc:=List(b.vcyc,v->[v[1],-v[2]]),operations:=CycPolOps);
  fi;
  if IsCyc(b) then return rec(coeff:=a.coeff/b,valuation:=a.valuation,
                  vcyc:=ShallowCopy(a.vcyc),operations:=CycPolOps);
  fi;
  return CycPolOps.normalize(rec(
    coeff:=a.coeff/b.coeff,
    valuation:=a.valuation-b.valuation,
    vcyc:=Concatenation(a.vcyc,List(b.vcyc,v->[v[1],-v[2]])),
    operations:=CycPolOps));
end;

CycPolOps.\^:=function(a,n)local res;
  res:=ShallowCopy(a);
  res.coeff:=res.coeff^n;res.valuation:=n*res.valuation;
  res.vcyc:=List(res.vcyc,x->[x[1],n*x[2]]);
  return res;
end;

CycPolOps.normalize:=function(a)local x,res,d,v,m,i,j;
  if a.coeff=0 then a.vcyc:=[];a.valuation:=0;fi;
  if a.vcyc=[] then return a;fi;
  res:=List(a.vcyc,x->[Denominator(x[1]),x[1],x[2]]);
  Sort(res);Add(res,[-1,-1,0]); # endmarker
  d:=0;a.vcyc:=[];
  for x in res do
    if d<>x[1] then 
      if d<>0 then
        for j in [1..i] do if m[j]<>0 then Add(a.vcyc,[v[j],m[j]]);fi;od;
      fi;
      d:=x[1];v:=[];m:=[];i:=0;
    fi;
    if i>0 and x[2]=v[i] then m[i]:=m[i]+x[3];
    else i:=i+1;v[i]:=x[2];m[i]:=x[3];
    fi;
  od;
  return a;
end;

CycPolOps.Value:=function(v,q)local res;
  if IsPolynomial(q) and q.baseRing=Rationals and 
     ForAll(v.vcyc,t->t[2]>=0) then
    res:=X(Cyclotomics)^v.valuation*Product(v.vcyc,
      t->(X(Cyclotomics)-E(Denominator(t[1]))^Numerator(t[1]))^t[2]);
    res:=Value(res,q);
  else res:=List(v.vcyc,t->(q-E(Denominator(t[1]))^Numerator(t[1]))^t[2]);
    if false in res then
      Error("Cannot evaluate the non-Laurent polynomial CycPol ",v);
    else res:=q^v.valuation*Product(res);
    fi;
  fi;
  if IsRec(v.coeff) and IsBound(v.coeff.operations) and
    IsBound(v.operations.Value) then return Value(v.coeff,q)*res;
  else return v.coeff*res;
  fi;
end;

# Computes CycPol(Value(p,q*e)) for a root of unity e
CycPolOps.EnnolaTwist:=function(v,e)local res,p,f;
 f:=function(x)if x<0 then return x+1;else return x;fi;end;
 res:=ShallowCopy(v);
 res.coeff:=res.coeff*e^Degree(res);
 e:=AsRootOfUnity(e);
 res.vcyc:=List(res.vcyc,p->[f(p[1]-e),p[2]]);
 return CycPolOps.normalize(res);
end;

# return CycPol(Value(p,q^n))
CycPolOps.DescentOfScalars:=function(p,n)local res;
  res:=ShallowCopy(p);res.valuation:=n*p.valuation;
  res.vcyc:=Concatenation(List(p.vcyc,x->List(x[1]/n+[0..n-1]/n,r->[r,x[2]])));
  return CycPolOps.normalize(res);
end;
    
CycPolOps.Degree:=function(a)local deg;
  deg:=a.valuation+Sum(a.vcyc,t->t[2]);
  if IsPolynomial(a.coeff) then deg:=deg+Degree(a.coeff);fi;
  return deg;
end;

# Valuation(p [,d]) : d-valuation i.e. returns power of (q-E(d))-factor of
# CycPol p (if d=0 return power of q)
CycPolOps.Valuation:=function(arg)local p,d;
  p:=arg[1];
  if Length(arg)=1 then d:=0;else d:=arg[2];fi;
  if d=0 then return p.valuation;
  elif d=1 then return Sum(Filtered(p.vcyc,x->x[1]=0),x->x[2]);
  else return Sum(Filtered(p.vcyc,x->x[1]=1/d),x->x[2]);
  fi;
end;

#f:=function(d)local p;p:=PrimeResidues(d);
#  d:=List(p,x->Jacobi(x,d)=1);d:=ListBlist(p,d);
#  return [d,Difference(p,d)];
#end;

# returns  the list  of  factors  of Phi(d)  which  are recognized  when
# formatting CycPols. Each  factor is return as the list  of i such that
# its roots are E(d)^i. The result is cached to go faster.
CycPolOps.PhiDecompositions:=function(d)local dec,decs,r,phi;
  if not IsBound(CycPolOps.decompositions) then CycPolOps.decompositions:=[];fi;
  if not IsBound(CycPolOps.decompositions[d]) then
    decs:=[,,,,,,,
      [[1,5],[3,7],[1,7],[3,5],[1,3],[5,7]],,,,  #8
      [[1,5],[7,11],[1,7],[5,11],[1,11],[5,7],[1],[5],[7],[11]],,,  #12
      [[1,4,11,14],[2,7,8,13],[1,4,7,13],[2,8,11,14],
       [1,4],[7,13],[11,14],[2,8]],  #15
      [[1,7,9,15],[3,5,11,13]],,,,  #16
      [[1,9,11,19],[3,7,13,17],[1,9,13,17],[3,7,11,19]],  #20
      [[1,4,10,13,16,19],[2,5,8,11,17,20]],,,  #21
      [[1,7,13,19],[5,11,17,23],[1,7,17,23],[5,11,13,19],  #24
       [1,5,19,23],[7,11,13,17],[1,11,17,19],[5,7,13,23],
       [7,13],[1,19],[5,23],[11,17]],,,,,,  
      [[1,11,19,29],[7,13,17,23],[1,7,13,19],[11,17,23,29],
       [11,29],[17,23],[1,19],[7,13]],,,,,,,,,,,,  #30
      [[1,13,19,25,31,37],[5,11,17,23,29,41]]];  #42
    dec:=[PrimeResidues(d)];
    if d=1 then dec:=[[1]];fi;
    if d>2 then r:=PrimitiveRootMod(d);
      if r<>false then
	phi:=Phi(d);
	Append(dec,[List([0,2..phi-2],i->r^i mod d),
		    List([1,3..phi-1],i->r^i mod d)]);
      fi;
    fi;
    if IsBound(decs[d]) then Append(dec,decs[d]);fi;
    CycPolOps.decompositions[d]:=dec;
  fi;
  return CycPolOps.decompositions[d];
end;

# dumps a CycPol to a vector: a list of
#  [n,integer d]     specifying Phi_d^n
#  [n,integer d,i]   specifying (i-th factor of Phi_d)^n
#  [n,i/d]           specifying (q-E(d)^i)^n
CycPolOps.analyze:=function(a)local n,res,den,t,d,v,pack;

  pack:=function(d,v)local dec,i,r,n,j;
    dec:=CycPolOps.PhiDecompositions(d);
    for i in [1..Length(dec)] do
      r:=dec[i]; n:=Minimum(v{r});
      if n>0 then v{r}:=v{r}-n;
      else n:=Maximum(v{r});if n<0 then v{r}:=v{r}-n; else n:=0;fi;
      fi;
      if n<>0 then if i=1 then Add(res,[n,d]);
      else Add(res,[n,d,i-1]);fi;fi;
    od;
    for i in [1..Length(v)] do if v[i]<>0 then Add(res,[v[i],i/d]);fi;od;
  end;

  res:=[];d:=0;
  for t in a.vcyc do
    den:=Denominator(t[1]);
    if den<>d then 
      if d<>0 then pack(d,v);fi;
      d:=den;v:=[1..d]*0;
    fi;
    n:=Numerator(t[1]);
    if n=0 then v[1]:=t[2]; else v[n]:=t[2];fi;
  od;
  if Length(a.vcyc)>0 then pack(d,v);fi;
  return res;
end;

CycPolOps.Used:=[];

# options:
# defaut print coeff.q^valuation.product of  Pi or P'i 
#            (see list of PhiDecompositions)
# TeX    same but in TeX
# GAP    print in extenso of each Pi
CycPolOps.Format:=function(a,option)local res,var;
  if IsBound(a.vname) then var:=a.vname;
  elif IsBound(X(Cyclotomics).name) then var:=X(Cyclotomics).name;
  else var:="q";
  fi;
  res:=List(CycPolOps.analyze(a),function(e)local p,n,c;
    n:=e[1];p:=e[2];
    if Length(e)=3 then
      if IsBound(option.GAP) then
	c:=Product(CycPolOps.PhiDecompositions(p)[e[3]+1],
	   l->X(Cyclotomics)-E(p)^l).coefficients;
	c:=Concatenation("(",StringPolynomial(c,0,var,option),")");
      elif IsBound(option.TeX) then 
	if e[3]>4 then c:=SPrint("{\\Phi^{(",e[3],")}_",TeXBracket(p),"}");
	else c:=SPrint("{\\Phi",List([1..e[3]],y->'\''),"_",TeXBracket(p),"}");
	fi;
        if not [p,e[3]] in CycPolOps.Used then
          AddSet(CycPolOps.Used,[p,e[3]]);
          Print("$",c,"=",
	    Format(Product(CycPolOps.PhiDecompositions(p)[e[3]+1],
	      l->Mvp("q")-E(p)^l),rec(LaTeX:=1,TeX:=1,GAP:=1)),"$\n");
        fi;
      elif e[3]>6 then c:=SPrint("P(",e[3],")",p);
      else c:=List([1..QuoInt(e[3],2)],y->'"');
	if e[3] mod 2=1 then Add(c,'\'');fi;
	c:=SPrint("P",c,p);
      fi;
    elif IsInt(p) then
      if IsBound(option.GAP) then
	c:=SPrint("(",StringPolynomial(CyclotomicPol(p),0,var,option),")");
      elif IsBound(option.TeX) then c:=SPrint("\\Phi_",TeXBracket(p));
      else c:=SPrint("P",p);
      fi;
    else 
      c:=Concatenation("(",
       StringPolynomial([-E(Denominator(p))^Numerator(p),1],0,var,option),")");
    fi;
    return StringMonomial(c,n,option);
  end);
  if IsBound(option.GAP) and not IsBound(option.TeX)then res:=Join(res,"*");
  else res:=Concatenation(res);fi;
  var:=StringMonomial(var,a.valuation,option);
  if var<>"1" then res:=Concatenation(var,res);fi;
  return StringCoefficient(a.coeff,res,option);
end;

CycPolOps.String:=a->Format(a);

CycPolOps.Print:=function(a)Print(String(a));end;

CycPolOps.ComplexConjugate:=function(a)
  a:=ShallowCopy(a);a.coeff:=ComplexConjugate(a.coeff);
  a.vcyc:=List(a.vcyc,function(v)
    v:=[1-v[1],v[2]];if v[1]=1 then v[1]:=0;fi;return v;end);
  return CycPolOps.normalize(a);
end;

###########################################################################
#  CycPol(vector v) returns a CycPol with vcyc:=v{3..Length(v)},
#     (Collected(v{3..Length(v)}) if elements of v are rationals)
#     coeff:=v[1] and valuation:=v[2]
#  CycPol(polynomial) returns the CycPol representing the polynomial
#        (special code to be fast for polynomials with <= 2 terms)
##
CycPol:=function(p)
  local res,d,a,l,i,r,q,e,n,q,tryorders,testcyc,testall,conductor;
  if IsList(p) then
    res:=rec(coeff:=p[1],valuation:=p[2],vcyc:=[],operations:=CycPolOps);
    if Length(p)>2 and IsList(p[3]) then res.vcyc:=p{[3..Length(p)]};
    else res.vcyc:=[];
      for a in p{[3..Length(p)]} do
        if IsInt(a) and a>0 then Append(res.vcyc,List(PrimeResidues(a),x->x/a));
        else Add(res.vcyc,a);
        fi;
      od;
      res.vcyc:=Collected(res.vcyc);
    fi;
  elif IsCyc(p) then
    res:=rec(coeff:=p,valuation:=0,vcyc:=[],operations:=CycPolOps);
  elif IsPolynomial(p) then
    res:=rec(coeff:=[],valuation:=p.valuation,vcyc:=[],operations:=CycPolOps);
    q:=X(Cyclotomics);
    if IsBound(q.name) then res.vname:=q.name;fi;
    if    # case 0
       Length(p.coefficients)=0 then res.coeff:=0;
    elif  # case ax^s
       Length(p.coefficients)=1 then res.coeff:=p.coefficients[1];
    elif  # case ax^s+bx^t
      ForAll(p.coefficients{[2..Length(p.coefficients)-1]},x->x=x*0) then
      d:=Length(p.coefficients)-1;
      res.coeff:=p.coefficients[d+1];
      r:=-p.coefficients[1]/res.coeff;
      a:=AsRootOfUnity(r);
      if a=false then res.coeff:=Value(p,q)/q^p.valuation;
      else res.vcyc:=List([0..d-1],i->[(a+i)/d,1]);
      fi;
    else
      tryorders:= # sorted by increasing Phi
      [1,2,3,4,6,5,8,10,12,7,9,14,18,15,16,20,24,30,11,22,13,21,26,28,36,42,
       32,40,48,60,35,45,56,70,72,84,90,64,80,96,120,63,108,126,105,112,140,
       144,168,180,210,192,240,135,216,252,270,288,336,360,420,189,378,315,
       504,540,630,720,756,1008,1260];
      res.coeff:=p.coefficients[Length(p.coefficients)];
      p:=Polynomial(Cyclotomics,p.coefficients/res.coeff); # now p is monic
      q:=X(Cyclotomics);
      testcyc:=function(i)local qr;
        if Degree(p)<Phi(i) then return false;fi;
	qr:=QuotientRemainder(p,CyclotomicPolynomial(Cyclotomics,i));
	if qr[2]=0*qr[2] then Append(res.vcyc,List(PrimeResidues(i),x->x/i));
          p:=qr[1];return testcyc(i);
        fi;
	return true;
      end;
      testall:=function(i)local r;
	if Degree(p)<Phi(i)/Phi(Gcd(i,conductor)) then return;fi;
	   # p is in F_conductor(x) so p cannot have a root in mu_i
        for r in PrimeResidues(i) do
	  while Value(p,E(i)^r)=0 do
	    p:=p/(q-E(i)^r);Add(res.vcyc,r/i);
	    conductor:=Lcm(List(p.coefficients,NofCyc));
	    if conductor=1 or Degree(p)<Phi(i)/Phi(Gcd(i,conductor)) then 
	       return;fi;
	  od;
	od;
      end;
      conductor:=Lcm(List(p.coefficients,NofCyc));
      i:=1;
      while i<Length(tryorders) and testcyc(tryorders[i]) do i:=i+1;od;
      i:=1;
      while i<Length(tryorders) and conductor<>1 do 
        testall(tryorders[i]);i:=i+1;
      od;
      res.vcyc:=Collected(res.vcyc);
      if Degree(p)=0 then res.coeff:=res.coeff*p.coefficients[1];
      else res.coeff:=res.coeff*p;
      fi;
    fi;
  fi;
  return CycPolOps.normalize(res);
end;
