#############################################################################
##
#A  symbol.g                CHEVIE library       Meinolf Geck, Frank L\"ubeck, 
#A                                                Jean Michel, G\"otz Pfeiffer
##
#A  $Id: symbol.g,v 1.1 1997/01/21 13:46:34 gap Exp $
##
#Y  Copyright (C) 1992 - 2005  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains a few functions dealing with multiple partitions and
##  symbols for unipotent characters of classical groups and complex
##  imprimitive group,  which are used in weylb.g, weyld.g and cmplximp.g,
##  in order to factor out some common code in these files.
##  
#############################################################################
##
#F  PartitionTupleToString( <mpart> ) . . . . . . . . . as the name says
##  
##  'IntListToString' is applied to the  partitions in <mpart> and
##  the results are concatenated by a separating '.':
##  
##     [[],[2,1,1]]      --->    ".211"
##     [[4,3],[2,1,1]]   --->    "43.211"
##     [[14,3],[2,1,1]]  --->    "(14,3).211"
##  
PartitionTupleToString:=p->String(Join(List(p,IntListToString),"."));

# same as above but the result is separated by ',' and bracketed
StringSymbol:=s->SPrint("(",Join(List(s,IntListToString)),")");

# shift a list of beta-numbers
ShiftBeta:=function(beta,n)
  if n>=0 then return Concatenation([0..n-1],beta+n);
  else return beta{[1-n..Length(beta)]}+n;
  fi;
end;

# symbol associated to multipartition p in series parametrized by cuspidal of
# shape s. Special case: when s is an integer it is interpreted as
# [s,0,0,...] so when p is a double partition one gets the symbol of
# 'defect' s associated to p.
SymbolPartitionTuple:=function(p,s)local e;
  e:=Length(p);
  if IsInt(s) then s:=[s];Append(s,[1..e-1]*0);fi;
  s:=List(p,Length)-s;
  s:=Maximum(s)-s;
  return List([1..e],i->ShiftBeta(BetaSet(p[i]),s[i]));
end;

# defect of a symbol with 2 lines
DefectSymbol:=d->Length(d[1])-Length(d[2]); 

# rank of a symbol
RankSymbol:=s->Sum(s,Sum)-QuoInt((Sum(s,Length)-1)*(Sum(s,Length)-Length(s)+1),
2*Length(s));

# gives the partition corresponding to a list b of beta-numbers
PartBeta:=function(b)
  if Length(b)=0 then return []; # needed since []+[] is an error!!!
  else return Reversed(Filtered(b-[0..Length(b)-1],x->x<>0));
  fi;
end;

# returns all the 2-line symbols of defect d and rank n
Symbols:=function(n,d)
  n:=n-QuoInt(d*d,4);
  if n<0 then return [];fi;
  if d>0 then return List(PartitionTuples(n,2),x->SymbolPartitionTuple(x,d));fi;
  return List(CHEVIE.RawData("CharParams","D")(n),
	      CHEVIE.RawData("symbolcharparam","D"));
end;

# Returns the fake degree of the unipotent character corresponding to
# symbol s as a 'CycPol' (see cycpol.g). Works for symbols of 
# G(e,1,r) (d=1) and G(e,e,r) (d=0). When given a second argument p which
# divides e, also works for the coset G(e,e,r).s_1^p
# See Malle, "Unipotente Grade", 2.11 and 5.7
#
CycPolFakeDegreeSymbol:=function(arg)
  local s,p,res,delta,theta,gamma,symm,r,d,e,q;
  s:=arg[1]; q:=X(Cyclotomics);
  delta:=S->Product(Filtered(Cartesian(S,S),x->x[1]>x[2]),
                                      x->CycPol(q^(e*x[1])-q^(e*x[2])));
  theta:=S->Product(Filtered(S,x->x>0),l->Product([1..l],h->CycPol(q^(e*h)-1)));

  e:=Length(s); r:=RankSymbol(s);
  if Length(s)=1 then d:=0; else d:=DefectSymbol(s); fi;
  if Length(arg)=2 then p:=arg[2];else p:=e;fi;

  if not d in [0,1] then return CycPol(0*q);fi;

  if d=1 then res:=theta([r]);
  elif d=0 then 
    if e=p then res:=theta([r-1])*CycPol(q^r-1);
    elif p=1 then res:=theta([r-1])*CycPol(q^r-E(e));
    fi;
  fi;

  res:=res*Product([1..e],i->delta(s[i])/theta(s[i]));
  gamma:=j->Sum([0..e-1],i->i*Sum(s[1+(i+j) mod e]));
  symm:=Number([1..e],i->s{List([1..e],j->1+(i+j-1) mod e)}=s);

  if d=1 then res.valuation:=res.valuation+gamma(0);
  elif d=0 then 
    if e=2 then 
      if p=1 then res:=res*CycPol(q^Sum(s[1])-q^Sum(s[2]))/symm;
      else res:=res*CycPol(q^Sum(s[1])+q^Sum(s[2]))/symm;
      fi;
    else res:=res*CycPol(Sum([0..e-1],j->E(e)^(j*p)*q^gamma(j)))/symm;
    fi;
  else res:=res*CycPol(q^Sum(s[1])+q^Sum(s[2]));
  fi;

  res.valuation:=res.valuation
                 -Sum(e*[1..QuoInt(Sum(s,Length),e)-1]+(d mod 2),x->x*(x-1)/2);
  if p=1 and r=2 and e>2 then res:=CycPolOps.EnnolaTwist(res,E(2*e));fi;
  return CycPolOps.normalize(res);
end;

# Generic degree of the unipotent character corresponding to symbol S
# as a 'CycPol' (see cycpol.g). Works for symbols for
# G(e,1,r) (d=1, defect=0), for G(e,e,r) (d=0, defect=0) 
# and for {}^2 G(e,e,2) (d=0, defect=1) (this includes 2Dn, 2B2, 2G2)
# here d=Inhalt mod. e
# See Malle's "Unipotente Grade", 3.9 and 6.4
CycPolGenericDegreeSymbol:=function(S)local sh,res,theta,m,r,d,e,defect,q;
  r:=RankSymbol(S); e:=Length(S); sh:=List(S,Length); 
  m:=QuoInt(Sum(sh),e); d:=Sum(sh)mod e; q:=X(Cyclotomics);
  defect:=(Binomial(e,2)*m-sh*[0..e-1]) mod e; 
  theta:=S->Product(S,l->Product([1..l]*e,h->CycPol(q^h-1)));

  # initialize with the q'-part of the group order
  if d=1 then res:=theta([r]);
  elif d=0 then res:=theta([r-1])*CycPol(q^r-E(e)^defect);
  fi;

  res:=res*(-1)^([0..e-1]*List(sh,x->Binomial(x,2)))*
   Product([0..e-1],i->Product([i..e-1],j->Product(List(S[i+1],l->q^l*E(e)^i),
    l->Product(Filtered(List(S[j+1],l->q^l*E(e)^j),
    m->i<j or Degree(m)<Degree(l)),m->CycPol(l-m)))))/
     (Product(S,theta)*(E(4)^Binomial(e-1,2)*ER(e^e))^m
     *CycPol(q^Sum(e*[1..m-1]+d,x->Binomial(x,2))));

  if d=0 then res:=res*First([1..e],i->S{1+List(i+[0..e-1],j->j mod e)}=S);fi;
  if defect=0 or r<>2 or e<=2 then return res;
  else return E(e)^-1*CycPolOps.EnnolaTwist(res,E(2*e));
  fi;
end;

# valuation of the generic degree of the unipotent character of symbol p 
# Works for symbols of 2Dn, and for symbols for G(e,1,r) and G(e,e,r)
LowestPowerGenericDegreeSymbol:=function(p)local m,e;
  e:=Length(p);p:=Concatenation(p); Sort(p); m:=Length(p);
  return p*[m-1,m-2..0]-QuoInt(m*(m-e)*(2*m-3-e),12*e);
end;

# degree of the generic degree of the unipotent character of symbol p 
# Works for symbols of 2Dn, and for symbols for G(e,1,r) and G(e,e,r)
HighestPowerGenericDegreeSymbol:=function(p)local m,r,e;
  r:=RankSymbol(p);e:=Length(p);
  p:=Concatenation(p); Sort(p); m:=Length(p);
  if (m mod e)=1 then r:=e*r*(r+1)/2;
  else r:=e*r*(r-1)/2+r;
  fi;
  return r+p*[0..m-1]-Sum(p,x->e*x*(x+1)/2)-QuoInt(m*(m-e)*(2*m-3-e),12*e);
end;

# list of standard tableaux of shape the partition-tuple S (that is, a filling
# of S with the numbers [1..Sum(S,Sum)] such that the numbers increase across
# the rows and down the columns).
Tableaux:=function(S)local rim,res,r,n,t,w;
  rim:=Concatenation(List([1..Length(S)],i->List(Filtered([1..Length(S[i])],
    j->0<>S[i][j] and (j=Length(S[i]) or  S[i][j+1]<S[i][j])), p->[i,p])));
  res:=[];
  w:=Sum(S,Sum);
  if w=0 then return [List(S,x->List(x,y->[]))];fi;
  for r in rim do
    n:=Copy(S);n[r[1]][r[2]]:=n[r[1]][r[2]]-1;
    n:=Tableaux(n);
    for t in n do Add(t[r[1]][r[2]],w);od;
    Append(res,n);
  od;
  return res;
end;

#############################################################################
##
#F  DifferencePartitions( <gamma>, <alpha> ) . . .  difference of partitions.
##
##  If <gamma>-<alpha> doesn't exist or  contains a 2x2 box then 'false'
##  is returned.  Otherwise this  difference is a  union of  skew hooks.
##  The  function returns  a  record with  fields  cc:=number of  hooks,
##  ll:=combined leg length and d:=the  contents of a box underneath the
##  last hook.
##  This function is used for the computation of single character values
##  for Hecke algebras of types A, B or D.
##  
DifferencePartitions:=function(gamma, alpha)local i, dp, old, new, int, inhook;

   dp:= rec(cc:= 0, ll:= 0);
   if Length(alpha) > Length(gamma) then return false; fi;
   old:= []; inhook:= false;  
   alpha:=Concatenation(alpha,[Length(alpha)+1..Length(gamma)]*0);
   for i in [1..Length(gamma)] do
      if alpha[i] > gamma[i] then return false; fi;
      new:= [alpha[i]+1..gamma[i]];
      int:= Intersection(old, new);
      if Length(int) > 1 then return false;
      elif Length(int) = 1 then dp.ll:= dp.ll + 1;
      elif inhook then dp.cc:= dp.cc + 1; dp.d:= old[1] - i; inhook:= false;
      fi;
      if new <> [] then inhook:= true; fi;
      old:= new;
   od;

   if inhook then dp.cc:= dp.cc + 1; dp.d:= old[1] - Length(gamma) - 1; fi;
   return dp;
end;

#############################################################################
##
#F  LessSymbols( <x>, <y> ) . . . . . . . . the '<' function for symbols
##  
##  A symbol  is smaller if it  has smaller defect, or  equal defect and
##  smaller shuffle-sequence for lexicographic  order where for a symbol
##  (a1,a2,.. |b1,b2... ) the shuffle sequence is a1,b1,a2,b2,...
##                                               
LessSymbols:=function(x,y) local dx, dy, lx, ly;             
  dx:=DefectSymbol(x); dy:=DefectSymbol(y);
  if dx<>dy then return dx<dy; fi;
  lx:=[];
  lx{2*[1..Length(x[1])]-1}:=x[1];
  lx{2*[1..Length(x[2])]}:=x[2];
  ly:=[];
  ly{2*[1..Length(y[1])]-1}:=y[1];
  ly{2*[1..Length(y[2])]}:=y[2];
  return lx<ly;
end;

# e-symbols of rank r, Malle-defect def and Inhalt=inh (mod e)
SymbolsDefect:=function(e,r,def,inh)
  local DefShape, ShapesSymbols, IsReducedSymbol;
  DefShape:=function(s)local e; e:=Length(s);
   return (Binomial(e,2)*QuoInt(Sum(s),e)-s*[0..e-1]) mod e;
  end;

  # possible shapes for cuspidal symbols of rank<=r, length e, Inhalt=d mod e
  ShapesSymbols:=function(r,e,d)local f,res,m,new;
    f:=function(lim2,sum,nb,max)local res,a;
      if nb=1 then 
	if sum=0 then return [[sum]];
	else return [];
	fi;
      fi;
      res:=[];a:=QuoInt(sum,nb-1);
      while a<=max and Binomial(a,2)<=lim2 and a<=sum do
	Append(res,List(f(lim2-Binomial(a,2),sum-a,nb-1,a),
	     x->Concatenation([a],x)));
	a:=a+1;
      od;
      return res;
    end;
    res:=[];m:=0;
    repeat new:=f(r+QuoInt((m*e+d-1)*(m*e+d-e+1),2*e),d+m*e,e,d+m*e);
	   Append(res,new); m:=m+1;
    until Length(new)=0;
    res:=Concatenation(List(res,x->Arrangements(x,e)));
    return Filtered(res,s->DefShape(s)=def and 
      ForAll(Rotations(s){[2..Length(s)]},x->DefShape(x)<>def or not x>s));
  end;

  IsReducedSymbol:=function(s)local e;e:=Length(s);
    return ForAll(List([1..e-1],i->s{1+List([i..i+e-1],x->x mod e)}),
       x->List(x,Length)<>List(s,Length) or not s>x);
  end;

  return Concatenation(List(ShapesSymbols(r,e,inh),s->
    Filtered(List(PartitionTuples(r-RankSymbol(List(s,x->[0..x-1])),e),
       x->SymbolPartitionTuple(x,s)),IsReducedSymbol)));
end;
