#############################################################################
##
#A  abscox.g              CHEVIE library       Frank Luebeck and  Jean Michel
##
#Y  Copyright (C) 1992 - 1999  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, and
#Y  Universite Paris VII.
##
##  This file contains  GAP functions for working with abstract Coxeter groups.
##
# They are represented by groups W which should have the following fields
# and methods defined:
#
#   .nbGeneratingReflections
#
#   .reflections      such that W.reflections{[1..W.nbGeneratingReflections]}
#                     are the Coxeter generators
#
#   .operations.IsLeftDescending(W,w,i)
#      for w in W and i in [1..W.nbGeneratingReflections], true iff
#      W.reflections[i]*w has a shorter length in the generators than w.
#
# plus the methods contained in

AbsCoxOps:=OperationsRecord("AbsCoxOps");

#############################################################################
# Once the above fields are defined, the function
#     AbsCoxOps.CompleteCoxeterGroupRecord
#
# Can be called to fill in some other fields/methods (the value provided by
# AbsCoxOps.CompleteCoxeterGroupRecord can be overridden by filling first
# the value when it is marked as 'default' below):
#
#   .nbGeneratingReflections
#       default:   Length(W.reflections)
#
#   .reflectionsLabels
#       default:   [1..W.nbGeneratingReflections]
#
#   ReflectionFromName
#       default:   x->Position(W.reflectionsLabels,x);
#
#   isCoxeterGroup
#       true
#
#   .coxeterMat
#       should be pre-filled when there is an infinite bond, otherwise
#       there will be an infinite loop. An infinite bond is marked by 0
#
#
# CompleteCoxeterGroupRecord also fills the following fields:
#
#   .EigenvaluesGeneratingReflections=List([1..nbGeneratingReflections],x->1/2)
#   .OrdersGeneratingReflections=List([1..nbGeneratingReflections],x->2)
#     [The eigenvalues and orders of the generators, filled in for upward 
#      compatibility with complex reflection groups]
#
#   .orbitRepresentative=List of length .nbGeneratingReflections indicating
#     for each reflection the smallest index of a reflection to which it is
#     conjugate
#
#    In addition, if W.isFinite is true,
#      W.longestElm
#      W.longestCoxeterWord
#      W.N  (number of reflections)
#
#    are computed by calling appropriate methods.

IsCoxeterGroup:=W->IsRec(W) and IsBound(W.isCoxeterGroup) and
                                W.isCoxeterGroup=true;

AbsCoxOps.CompleteCoxeterGroupRecord:=function(W)local i,j,ord;
  W.operations:=ShallowCopy(W.operations);Inherit(W.operations,AbsCoxOps);
  if not IsBound(W.nbGeneratingReflections) then
    W.nbGeneratingReflections:=Length(W.reflections);
  fi;
  W.generatingReflections:=[1..W.nbGeneratingReflections];
  W.EigenvaluesGeneratingReflections:=List(W.generatingReflections,x->1/2);
  W.isCoxeterGroup:=true;
  if not IsBound(W.reflectionsLabels) then
    W.reflectionsLabels:=W.generatingReflections;
  fi;
  if not IsBound(W.operations.ReflectionFromName) then
    W.operations.ReflectionFromName:=function(W,x)
      return Position(W.reflectionsLabels,x);end;
  fi;
  if not IsBound(W.rootInclusion) then
     W.rootInclusion:=W.generatingReflections;
     W.rootRestriction:=W.generatingReflections;
  fi;
  if not IsBound(W.orbitRepresentative) then
    ord:=DecomposedMat(List(CoxeterMatrix(W),x->List(x,y->y mod 2)));
    W.orbitRepresentative:=[];
    for i in [1..Length(ord)] do
      W.orbitRepresentative{ord[i]}:=ord[i]*0+W.rootInclusion[i];
    od;
  fi;
  if not IsBound(W.rank) then
    W.rank:=W.nbGeneratingReflections;
    W.semisimpleRank:=W.rank;
  fi;
  if not IsBound(W.OrdersGeneratingReflections) then
     W.OrdersGeneratingReflections:=List(W.generatingReflections,x->2);
  fi;
  if not IsBound(W.type) then
    i:=ReflectionType(W);
    if i<>false then W.isFinite:=true;W.type:=i;
    else W.isFinite:=false;
    fi;
  fi;
  if IsFinite(W) then W.N:=Length(LongestCoxeterWord(W));fi;
end;

###########################################################################
#
# N.B. a difference with older versions of Chevie is that IsLeftDescending and
# FirstLeftDescending  return  an  index  in  the  list  of  generators, not a
# reflectionName,  for  efficiency.  LeftDescentSet  still  returns  names  of
# reflections.
#
###########################################################################

##  generic reduction of FirstLeftDescending to IsLeftDescending
##  in practical implementations of Coxeter groups this routine can often
##  be overriden by a more efficient one.

AbsCoxOps.FirstLeftDescending:=function(W, x)local i,ILD;
  ILD:=W.operations.IsLeftDescending; # avoid dispatching overhead
  for i in W.generatingReflections do if ILD(W,x,i) then return i; fi; od;
  return false;
end;

#############################################################################
##
#F  LeftDescentSet( W, x)  the set of generators s such that sx < x
##
##  the generators returned as their reflectionName
##

AbsCoxOps.LeftDescentSet:=function(W, x)
  return W.reflectionsLabels{Filtered(W.generatingReflections,
        i->IsLeftDescending(W,x,i))};
end;

#############################################################################
##
#F  RightDescentSet( W, x)  the set of generators s such that xs < x
##

RightDescentSet:=function(W,w) return LeftDescentSet(W,w^-1); end;

#############################################################################
##
#F  CoxeterWord( <W> , <w> )  . . . . . . . . for Coxeter groups
##
##  'CoxeterWord' returns a reduced word in  the standard generators of W
##  determined by the element w of W. The generators are denoted by the
##  corresponding element of W.reflectionsLabels.
##

AbsCoxOps.CoxeterWord:=function(W, w)local  l, i;
  l := [];
  while true do
    i := W.operations.FirstLeftDescending(W, w);
    if i=false then return W.reflectionsLabels{l}; fi;
    Add(l,i);
    w := W.reflections[i]*w;
  od;
end;

#############################################################################
##
#F  EltWord ( <W> , <w> ) . .  convert a word to a group element.
##
##  Returns the element of the Coxeter group <W> corresponding to the
##  Coxeter word <w>.

AbsCoxOps.EltWord:=function(W,w)
  if Length(w)=0 then return W.identity;fi;
  return Product(W.reflections{List(w,a->W.operations.ReflectionFromName(W,a))});
end;

#############################################################################
##
#F  ReducedCoxeterWord( <W> , w )  . . . . . . . . a reduced word for w in W
##
##  'ReducedCoxeterWord' returns a reduced expression for the element w, given
##  as  an  arbitrary  list  of  W.reflectionsLabels
##
ReducedCoxeterWord:=function( W, w )return CoxeterWord(W, EltWord (W, w)); end;

#############################################################################
##
#F  CoxeterLength( <W> , <w> )  . . . . . length of element w of W
##
##  'CoxeterLength'  returns  the length of w in the generating reflections.
##
AbsCoxOps.CoxeterLength:=function(W, w)return Length(CoxeterWord(W, w));end;

#############################################################################
##
#F  CoxeterElements( <W>, <l> ) . . the set of elements of W of length l
##
##  Returns as a list of elements of W the set of elements of length l,
##  sorted (according to the default order on elemnts of W).
##

AbsCoxOps.CoxeterElements:=function(W,l)local e,i,H,x;
  if not IsBound(W.elts) then W.elts:=[[W.identity]];fi;
  if IsBound(W.elts[l+1])then return W.elts[l+1];fi;
  if IsFinite(W) then
    if l>W.N then 
      InfoChevie("#W no elements of that length\n");
      return [];
    fi;
    H:=ReflectionSubgroup(W,
                         W.reflectionsLabels{[1..W.nbGeneratingReflections-1]});
    W.elts[l+1]:=[];
    if not IsBound(W.rc) then 
      W.rc:=List([0..W.N],x->[]);
      for x in ReducedRightCosetRepresentatives(W,H) do
	 Add(W.rc[1+CoxeterLength(W,x)],x);
      od;
    fi;
    for i in [0..Minimum(l,H.N)] do
      e:=CoxeterElements(H,i);
      for x in W.rc[1+l-i] do Append(W.elts[l+1],e*x);od;
    od;
    W.elts[l+1]:=Set(W.elts[l+1]);
    InfoChevie2("#I Number of elements of length ",l,": ",
				     Length(W.elts[l+1]),"\n");
    if W.N-l<>l then
      W.elts[1+W.N-l]:=Set(W.elts[l+1]*LongestCoxeterElement(W));
    fi;
  else
    W.elts[l]:=CoxeterElements(W,l-1);
    W.elts[l+1]:=[];
    for e in W.elts[l] do
      for i in Filtered(W.generatingReflections,
			 x->not IsLeftDescending(W,e,x))
      do AddSet(W.elts[l+1],W.reflections[i]*e);od;
    od;
  fi;
  return W.elts[l+1];
end;

#############################################################################
##
## The elements are returned in increasing length order...
##

AbsCoxOps.Elements:=function(W)
  if not IsBound(W.N) then Error("W should be finite");fi;
  return Set(Concatenation(List([0..W.N],x->CoxeterElements(W,x))));
end;

#############################################################################
##
#F  CoxeterWords( <W> [, <l>] )  . . . . . .  words of W [ of length <l>]
##
##  'CoxeterWords' returns the list words in the Coxeter group W of length l.
##  If l is omitted then all words are returned, sorted by increasing length.
##

AbsCoxOps.CoxeterWords := function(arg)local W;
  W:=arg[1];
  if Length(arg)=1 then
    if not IsBound(W.N) then Error("W should be finite");
    else return Concatenation(List([0..W.N],i->CoxeterWords(W,i)));
    fi;
  else return List(CoxeterElements(W,arg[2]),w->CoxeterWord(W,w));
  fi;
end;

#############################################################################
##
#F  ReducedInRightCoset( <W> , <w> )  . . . . . reduced element in coset W.w
##
##  w is an automorphism of a parent Coxeter group of W.
##  'ReducedInRightCoset' returns the unique element in the right
##  coset W.w which is W-reduced.
##

AbsCoxOps.ReducedInRightCoset:=function(W,w)local i;
  while true  do
    i := FirstLeftDescending(W, w);
    if i=false then return w; fi;
    w := W.reflections[i] * w;
  od;
end;

#############################################################################
##
#F  ReducedRightCosetRepresentatives( <W>, <H> )  . . . . . . . . . . . . .
#F  . . . . . . . . . . . . . . .  distinguished right coset representatives
##
##  'ReducedRightCosetRepresentatives'  returns a list of reduced words in the
##  Coxeter group <W> that are distinguished right  coset representatives for
##  the right cosets H\W where  H is a reflection subgroup  of W.
##
ReducedRightCosetRepresentatives:=function(W, H)local res, totest, new;
  totest:=Set([W.identity]);
  res:=Set([W.identity]);
  repeat
    new:=Concatenation(List(totest,w->List(
      W.reflections{W.generatingReflections},s->ReducedInRightCoset(H, w*s))));
    UniteSet(res,totest);
    totest:=Difference(new,res);
  until Length(totest)=0;
  InfoChevie2("#I nb. of cosets: ",Length(res),"\n");
  SortParallel(List(res,x->CoxeterLength(W,x)),res);
  return res;
end;

#############################################################################
##
#F  PermsCosetsSubgroup ( <W>, <H> ) . . . . . . . . . . . . . .
#F  . . . .  permutation representation on the cosets of a reflection subgroup
##
##  'PermsCosetsSubgroup'  returns  the  list of permutations induced
##  by the standard generators of the Group  <W>  on the cosets  of  the
##  subgroup <H>. The cosets are in the order determined by the result of
##  the  function 'ReducedRightCosetRepresentatives( <W>, <H>)'.
##
PermCosetsSubgroup :=function(W,H)local D;
  D:=ReducedRightCosetRepresentatives(W, H);
  return List(W.reflections{W.generatingReflections},
    s->PermListList(List(D*s,x->ReducedInRightCoset(H,x)),D));
end;

#############################################################################
##
#F  Bruhat( <W>, <y>, <w>[, <ly>, <lw> ] ) . . . . . . . Bruhat partial order
##
##  'Bruhat'  returns true, if the element  <y>  is less than or equal to the
##  element <w> of the Coxeter group <W>, and false otherwise. Both <y> and
##  <w>  must be given as group elements.
##  additional arguments <ly> and <ly> can contain the lengths of <y> and <w>.
##
Bruhat := function ( arg ) local  s, i, x, y, lx, ly, W,ild,fld;
  W:=arg[1]; x:=arg[2]; y:=arg[3];
  if Length(arg)>3 then lx:=arg[4]; ly:=arg[5];
  else lx:=CoxeterLength(W,x); ly:=CoxeterLength(W,y);
  fi;

  ild:=W.operations.IsLeftDescending;#suppress dispatch overhead gains 40% speed
  fld:=W.operations.FirstLeftDescending;#this gains another 40% speed
  if lx=0 then return true;fi;
  while lx<ly do
    i:=fld(W,y);
    s := W.reflections[i];
    if ild(W,x,i) then
      if lx=1 then return true;fi;
      x:=s*x; lx:=lx-1;
    fi;
    y:=s*y; ly:=ly-1;
  od;
  return x=y;
end;

#############################################################################
##
#F  BruhatSmaller( <W>, <w>) . . List of elements Bruhat smaller than w
##
##  The result is a list whose i-th element is the list of elements of W
##  Bruhat smaller than w and of length i-1 (so the first element of the
##  list holds only W.identity and the CoxeterLength(W,w)-th element of
##  the lists holds only w).
##
BruhatSmaller:=function(W,w)local res,s,i,j;
  if w=W.identity then return [[w]];fi;
  i:=FirstLeftDescending(W,w);
  s:=W.reflections[i];
  res:=BruhatSmaller(W,s*w);Add(res,[]);
  for j in [1..Length(res)-1] do
    UniteSet(res[j+1],s*Filtered(res[j],x->not IsLeftDescending(W,x,i)));
  od;
  return res;
end;

# The following two methods will never finish if called for an infinite group

#############################################################################
##
#F  LongestCoxeterElement( <W> ) . . Longest element of W
##
##  LongestCoxeterElement returns as a group element the longest element of W
##

LongestCoxeterElement:=function( W )local i;
  if IsBound(W.longestElm) then return W.longestElm; fi;
  W.longestElm := W.identity;
  if W.nbGeneratingReflections=0 then return W.longestElm;fi;
  while true do
      i := 1;
      while IsLeftDescending(W, W.longestElm, i) do
	 i := i + 1;
	 if i>W.nbGeneratingReflections then return W.longestElm;fi;
      od;
      W.longestElm:=W.reflections[i]*W.longestElm;
  od;
end;

#############################################################################
##
#F  LongestCoxeterWord( <W> )  . . . . . . . . . .  the longest element in W
##
## 'LongestCoxeterWord' returns a reduced expression in the  standard
##  generators of the unique longest element of the Coxeter group W.
##

LongestCoxeterWord:=function(W)local save;
  if not IsBound(W.longestCoxeterWord) then
    W.longestCoxeterWord:=CoxeterWord( W, LongestCoxeterElement(W));
  fi;
  return W.longestCoxeterWord;
end;

CoxeterMatrix:=function(W)local ord,i,j,m;
  if not IsBound(W.coxeterMat) then
    if IsBound(W.cartan) then
      W.coxeterMat:=CoxeterMatrixFromCartanMat(W.cartan);
    else
      W.coxeterMat := IdentityMat(W.nbGeneratingReflections);
      for i in W.generatingReflections do
	for j in [1..i-1] do
	  m:=W.reflections[i]*W.reflections[j];
	  if not IsFinite(W) and # try to detect cheaply an infinite bond
	    IsMat(m) and TraceMat(m)=Length(m) then ord:=0;
	  else ord:=Order(W,W.reflections[i]*W.reflections[j]);
	  fi;
	  W.coxeterMat[i][j]:=ord;
	  W.coxeterMat[j][i]:=ord;
	od;
      od;
    fi;
  fi;
  return W.coxeterMat;
end;

CoxeterMatrixFromCartanMat:=function(m)local res,i,j,find,c;
  find:=function(c)local x;
    x:=[2,3,4,6,0];if c in [0..4] then return x[c+1];fi;
    x:=NofCyc(c);
    if c=2+E(x)+E(x)^-1 then return x; 
    elif c=2+E(2*x)+E(2*x)^-1 then return 2*x;
    else Error("not a Cartan matrix of a Coxeter group");
    fi;
  end;
  res:=m^0;
  for i in [2..Length(m)] do for j in [1..i-1] do
    c:=find(m[i][j]*m[j][i]);res[i][j]:=c;res[j][i]:=c;
  od;od;
  return res;
end;

CartanMatFromCoxeterMatrix:=m->List(m,l->List(l,
  function(c)if c=0 then return -2;else return -E(2*c)^-1-E(2*c);fi;end));

AbsCoxOps.CartanMat:=function(W)
  if not IsBound(W.cartan) then
    W.cartan:=CartanMatFromCoxeterMatrix(CoxeterMatrix(W));
  fi;
  return W.cartan;
end;

AbsCoxOps.ReflectionType:=function(W)local t;
  t:=ReflectionType(CartanMat(W));
  if ForAny(t,x->x=false) then return false;fi;
  W.operations:=ShallowCopy(W.operations);
  Inherit(W.operations,HasTypeOps);
  return t;
end;

AbsCoxOps.BraidRelations:=function(W)local m,p;
  p:=function(i,j,b)return List([1..b],k->i*(k mod 2)+j*((1-k)mod 2));end;
  m:=CoxeterMatrix(W);
  return Concatenation(List([1..Length(m)],i->List([1..i-1],
    j->[p(i,j,m[i][j]),p(j,i,m[i][j])])));
end;

#############################################################################
##
#F  ReflectionSubgoup( <W>, <J> ) . . . . .  standard parabolic W_J
##
## <J> is given as a list of reflection names for W
##

AbsCoxOps.ReflectionSubgroup:=function(W,J)local refs,P,inc;
  if not IsSubset(W.reflectionsLabels,J) then
    Error("Only standard parabolic subgroups implemented for abstract coxeter groups");
  fi;
  inc:=List(J,x->Position(W.reflectionsLabels,x));
  refs:=W.reflections{inc};
  P:=Subgroup(W,refs);
  P.reflections:=refs;
  P.rootInclusion:=W.rootInclusion{inc};
  P.rootRestriction:=[];
  P.rootRestriction{P.rootInclusion}:=[1..Length(J)];
  P.reflectionsLabels:=W.reflectionsLabels{inc};
  P.rank:=W.rank;
  P.semisimpleRank:=Length(inc);
  if IsBound(W.reflectionParent) then P.reflectionParent:=W.reflectionParent;
  else P.reflectionParent:=W;
  fi;
  P.operations.IsLeftDescending:=function(P,w,i)
    return IsLeftDescending(P.reflectionParent,w,P.rootInclusion[i]);end;
  AbsCoxOps.CompleteCoxeterGroupRecord(P);
  return P;
end;

AbsCoxOps.BrieskornNormalForm:=function(W,w)local l,I,found,IL,i;
  l:=[];IL:=W.operations.IsLeftDescending;
  while true do
    I:=Filtered(W.generatingReflections,i->IL(W,w,i));
    if I=[] then return l;fi;
    Add(l,W.rootInclusion{I});
    repeat
      found:=false;
      for i in I do if IL(W,w,i) then found:=true;w:=W.reflections[i]*w; fi; od;
    until not found;
  od;
end;

# Calls f on each element of W
ForEachElement:=function(W,f)local l,g;
  if not IsFinite(W) then Error("only for finite Coxeter groups");
  elif W.nbGeneratingReflections=0 then f(W.identity);return;
  fi;
  l:=List([0..W.nbGeneratingReflections],i->
        ReflectionSubgroup(W,W.reflectionsLabels{[1..i]}));
  l:=List([1..Length(l)-1],i->ReducedRightCosetRepresentatives(l[i+1],l[i]));
  g:=function(x,v)local y;
    if Length(v)=0 then f(x);
    else for y in v[1] do g(x*y,v{[2..Length(v)]});od;
    fi;
  end;
  g(W.identity,l);
end;

##############################################################################
##
## CoxeterHeckeAlgebraOps: operations for Hecke algebras for Coxeter groups.
##
CoxeterHeckeAlgebraOps:=OperationsRecord("CoxeterHeckeAlgebraOps",AbsHeckeOps);

ReadChv("prg/heckeelt"); # creates T Basis for CoxeterHeckeAlgebraOps

ReadChv("prg/heckemod");

AbsCoxOps.Hecke:=function(arg)local H;
  H:=ApplyFunc(PermRootOps.Hecke,arg);
  H.operations:=CoxeterHeckeAlgebraOps;
  return H;
end;
