#############################################################################
##
#A  affine.g              CHEVIE library        Jean Michel, Raphael Rouquier
##
##
#Y  Copyright (C) 1992 - 2000  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains functions for affine Weyl groups.
##  Since they are Coxeter groups, all functions for general Coxeter groups
##  like Hecke algebras, hecke modules etc.. are supported by the new scheme
##  (AbsCoxOps)
##

AffineCoxeterGroupOps:=OperationsRecord("AffineCoxeterGroupOps",GenCoxOps);

AffineCoxeterGroupOps.String:=W->SPrint("Affine(",W.linear,")");

AffineCoxeterGroupOps.Print:=function(W)Print(String(W));end;

#############################################################################
##
#F  Affine(<W>) . . . . . . . . . . . Makes the affine group corresponding
#F    to the irreducible Weyl group W
##
Affine:=function(W)local tmp,Wa,i,j,n;
  if Length(ReflectionType(W))>1 then
    Error("Affine(W) implemented only for irreducible W");
  fi;
# if not Forall(W.cartan,IsInt) then
#   Error("Affine(W) implemented only for crystallographic W");
# fi;
  tmp:=rec();
  tmp.operations:=AffineCoxeterGroupOps;
  tmp.nbGeneratingReflections:=W.nbGeneratingReflections+1;
  n:=tmp.nbGeneratingReflections;
  if
  Set(W.rootInclusion{[1..n-1]})<>[1..n-1]
  then Error("Affine Coxeter groups implemented only for Coxeter groups W such that\n",
  "Set(W.rootInclusion{[1..W.nbGeneratingReflections]})=[1..W.nbGeneratingReflections]");
  fi;
  tmp.reflectionsLabels:=[1..n];
##
## Extended Cartan matrix
##
  tmp.cartan:=List(W.cartan,ShallowCopy);
  for i in [1..W.nbGeneratingReflections] do 
               Add(tmp.cartan[i],-W.cartan[i]*W.roots[W.N]);od;
  Add(tmp.cartan,Concatenation(-W.coroots[W.N]*W.cartan,[2]));
##
## Action of the affine Weyl tmp group on the dual Cartan algebra
## (a n+1 dimensional vector space) following Kac, 1.1 and 3.7
##
  tmp.identity:=IdentityMat(n+1);
  tmp.reflections:=List([1..n],i->List(tmp.identity,ShallowCopy));
  for i in [1..n] do for j in [1..n] do 
    tmp.reflections[i][j][i]:=tmp.reflections[i][j][i]-tmp.cartan[i][j];
  od;od;
  tmp.reflections[n][n+1][n]:=-1;
  Wa:=ApplyFunc(Group,tmp.reflections);
  Inherit(Wa,tmp);
  Wa.linear:=W;
  Wa.isFinite:=false;
  if W.semisimpleRank=1 then Wa.coxeterMat:=[[1,0],[0,1]];fi;#~A1 infinite bond!
  AbsCoxOps.CompleteCoxeterGroupRecord(Wa);
  return Wa;
end;

## Given an affine Weyl group W, x a vector in the basis of 
## simple roots of W.linear and w in W, returns the image of x under w.
##
AffineRootAction:=function(W,w,x) local y;
  y:=Concatenation(x,[0,1])*w;
  return y{[1..Length(x)]}-y[Length(x)+1]*W.linear.roots[W.linear.N];
end;

AffineCoxeterGroupOps.IsLeftDescending:=function(W,w,i)
  return Sum(w[i])<0;
end;

#############################################################################
#F  PrintDiagram( <rec> ) . . prints Dynkin diagram of affine Coxeter
##
AffineCoxeterGroupOps.PrintDiagram:=function(W)
  local i, v, r, n, t, a, o;
  a:=W.linear.type[1];
  v:=W.reflectionsLabels;r:=a.rank;t:=a.series;
  if IsBound(a.bond) then 
    if a.bond mod 2 = 1 then
      Print(String("",7+Length(String(a.bond)))," ",v[3],"\n");
      Print("      ",a.bond," /",a.bond,"\\ ",a.bond,"\n");
      Print("I2(",a.bond,")~ ",v[1]," - ",v[2],"\n");
    elif a.bond mod 4 =2 then 
      Print(String("",7+Length(String(a.bond))),a.bond,"\n");
      Print("I2(",a.bond,")  ",v[1]," < ",v[2]," - ",v[3],"\n");
    else 
      Print(String("",8+Length(String(a.bond))),a.bond,
            String("",4-Length(String(a.bond))),a.bond,"\n");
      Print("I2(",a.bond,")  ",v[3]," > ",v[1]," < ",v[2],"\n");
    fi;
  elif t="B" then
    if IsBound(a.cartanType) and a.cartanType=1 then
      Print("C",r,"~   ",v[1]," > ",v[2]);
      for i in [3..r] do Print(" - ",v[i]);od;
      Print(" < ",v[r+1],"\n");
    else
      Print(String("",Length(String(r))+4*(r-1)+1),v[r+1],"\n");
      Print(String("",Length(String(r))+4*(r-1)+1),"|\n");
      Print("B",r,"~   ",v[1]," < ",v[2]);
      for i in [3..r] do Print(" - ",v[i]);od;Print("\n");
    fi;
  elif t="A" then
    if r=1 then Print("A1~  ",v[1]," oo ",v[2],"\n");
    else 
      Print("        ");
      for i in [1..r-3] do Print(" -");od;
      if r>2 then Print("  ",v[r+1],"  ");
      else Print(v[r+1]);fi;
      for i in [1..r-3] do Print("- ");od;
      Print("\n       /",String("",4*(r-2)+1),"\\\n");
    Print("A",r,"~   ",v[1]);
    for i in [2..r] do Print(" - ",v[i]);od;Print("\n");
    fi;
  elif t="D" then
    Print("D",r,"~  ",v[1],String("",4*(r-3)-1),v[r+1],"\n");
    Print(String("\\",Length(String(r))+6),String("",4*(r-3)-3),"/");
    Print(String("",Length(String(r))+6),"\n",
		  String("",Length(String(r))+6),v[3]);
    for i in [4..r-1] do Print(" - ",v[i]);od;Print("\n");
    Print(String("/",Length(String(r))+6),String("",4*(r-3)-3),"\\\n");
    Print(String("",Length(String(r))+4),v[2],String("",4*(r-3)-1),v[r],"\n");
  elif t="E" then
    if r=6 then
    Print("             ",v[7],"\n",
          "             |\n",
	  "             ",v[2],"\n",
          "             |\n",
          "E6~  ",v[1]," - ",v[3]," - ",v[4]," - ",v[5]," - ",v[6],"\n");
    elif r=7 then
    Print("                 ",v[2],"\n",
          "                 |\n",
          "E7~  ",v[8]," - ",v[1]," - ",v[3]," - ",v[4]," - ",v[5]," - ",
	  v[6]," - ",v[7],"\n");
    else 
    Print("             ",v[2],"\n",
          "             |\n",
          "E8~  ",v[1]," - ",v[3]," - ",v[4]," - ",v[5]," - ",
	  v[6]," - ",v[7]," - ",v[8]," - ",v[9],"\n");
    fi;
  elif t="G" then
    Print("G2~  ",v[3]," - ",v[1]," > ",v[2]," \n");
  elif t="F" then
    Print("F4~  ",v[5]," - ",v[1]," - ",v[2]," > ",v[3]," - ",v[4],"\n");
  elif t="H" then
    if r=3 then
      Print("     ",v[4],"\n",
            "    5 \\\n",
	    "H3~    ",v[2]," - ",v[3],"\n",
            "    5 /\n",
            "     ",v[1],"\n");
    else
      Print("        5           5\n");
      Print("H",r,"    ",v[1]," - ",v[2]," - ",v[3]," - ",v[4]," - ",v[5],"\n");
    fi;
  fi;
end;
