#############################################################################
##
#A  complexr.g  CHEVIE library    Jean Michel, David Bessis and Gunter Malle
##
#A  $Id: complexr.g,v 1.4 1997/04/14 09:08:54 werner Exp $
##
#Y  Copyright (C) 1992 - 1996  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains preliminary support for complex reflection
##  group in the setting of the chevie package.
##       Jean Michel and David Bessis 20-9-96
##  
#############################################################################
#
# Complex reflection groups are groups G acting on a vector space V,
# and generated by pseudo-reflexions in V. They are represented by
# finding a set of 'roots' in V invariant by G and such that all
# reflections in G are reflections w.r.t. some roots.
#
# They are represented as PermRootGroups with the additional field:
#  .type -- the types of irreducible components
# a component of type is of the form
# ["ST",W.generatingReflections,n]  for G(n)
# and
# ["ST",W.generatingReflections,p,q]  for G(p,q,r)

ComplexGroupOps:=OperationsRecord("ComplexGroupOps",PermRootOps);
Inherit(ComplexGroupOps,HasTypeOps);

ComplexGroupOps.Size:= W->Product(ReflectionDegrees(W));

ComplexGroupOps.Print:=function(W)Print(String(W));end;

ComplexGroupOps.String:=function(W)
  if IsBound(W.parent) then
    return SPrint("ReflectionSubgroup(",W.parent,", ",
            W.rootInclusion{W.generatingReflections},")");
  fi;
  return Join(List(W.type,function(t)local res;
    if IsBound(t.ST) then return SPrint("ComplexReflectionGroup(",t.ST,")");
    elif t.series="ST" then
      return SPrint("ComplexReflectionGroup(",t.p,",",t.q,",",t.rank,")");
    elif IsBound(t.bond) then 
      return SPrint("CoxeterGroup(\"",t.series,"\",",t.rank,",",t.bond,")");
    else return SPrint("CoxeterGroup(\"",t.series,"\",",t.rank,")");
    fi;
  end),"*");
end;

#############################################################################
##
#F  ComplexReflectionGroup( <args> ) . . . construct a complex refl. group
##  
# The function ComplexReflectionGroup has two forms:
#  ComplexReflectionGroup(d,p,n) returns the general imprimitive group g(d,p,n)
#  ComplexReflectionGroup(STnumber) returns the exceptional
#                     complex reflection group of Shephard-Todd number STnumber
#
ComplexReflectionGroup:=function(arg)local STnumber,res,p,q,r,type;
  if IsRec(arg[1]) and IsBound(arg[1].complexGroup) then
    return arg[1].complexGroup;
  fi;
  STnumber:=arg[1];
  if Length(arg)=3 then p:=arg[1];q:=arg[2];r:=arg[3];
# construct G(p,q,r) which is defined for p>1, q|p, r>0.
# for r=1 only value of q which makes sense is 1
    if not (r>0 and IsInt(p/q) and (r>1 or q=1)) then 
      Error("(p,q,r) must satisfy: p>1, q|p, r>0, and r=1 => q=1");
    fi;
    if p=2 and q=1 then return CoxeterGroup("B",r);
    elif p=2 and q=2 then return CoxeterGroup("D",r);
    elif p=1 and q=1 then return CoxeterGroup("A",r-1);
    elif r=2 and p=q then return CoxeterGroup("I",2,p);
    else type:=rec(series:="ST",p:=p,q:=q,rank:=r,operations:=ReflTypeOps);
      res:=rec(roots:=ShallowCopy(ReflTypeOps.GeneratingRoots(type)),
	 EigenvaluesGeneratingReflections:=ShallowCopy(
		 ReflTypeOps.EigenvaluesGeneratingReflections(type)));
    fi;
  elif STnumber in [23,28,30,35,36,37] then 
    type:=[["H",3],["F",4],["H",4],["E",6],["E",7],["E",8]];
    return ApplyFunc(CoxeterGroup,type[Position([23,28,30,35,36,37],STnumber)]);
  elif STnumber in Concatenation([4..22],[24..27],[29],[31..34]) then 
    type:=rec(series:="ST",ST:=STnumber, rank:=0, operations:=ReflTypeOps);
                                                          # enough to find data
    if STnumber in Concatenation([4..22],[24,27,33]) then
      res:=rec(roots:=ShallowCopy(ReflTypeOps.GeneratingRoots(type)),
	 simpleCoroots:=ShallowCopy(ReflTypeOps.GeneratingCoRoots(type)));
    else
      res:=rec(roots:=ShallowCopy(ReflTypeOps.GeneratingRoots(type)),
	 EigenvaluesGeneratingReflections:=ShallowCopy(
		 ReflTypeOps.EigenvaluesGeneratingReflections(type)));
    fi;
    type.rank:=Length(res.roots[1]);
  else Error("parameter must be in the interval [4..37]\n");
  fi;
  type.indices:=[1..Length(res.roots)];
  res.type:=[type];
  res.name:=ReflTypeOps.ReflectionName(type,rec());
  res:=PermRootGroup(res);
  res.N:=Sum(ReflectionCoDegrees(type),x->x+1);
  res.coN:=Sum(ReflectionDegrees(type),x->x-1);
  res.operations:=ComplexGroupOps;
  return res;
end;

ComplexGroupOps.\*:=function(W1,W2)local res,p,i,t;
  res:=PermRootOps.\*(W1,W2);
  if IsBound(W1.ReflectionCoDegrees) and IsBound(W2.ReflectionCoDegrees) then
    res.ReflectionCoDegrees:=Concatenation(W1.ReflectionCoDegrees,W2.ReflectionCoDegrees);
  fi;
  res.name:=Concatenation(ReflectionName(W1),"x",ReflectionName(W2));
  res.operations:=ComplexGroupOps;
  return res;
end;

##########################################################################
##
#F  Hecke( <G>, <para> ) . . . Hecke algebra for complex reflection groups
##
##  'Hecke' returns the cyclotomic Hecke algebra corresponding to the
##  complex reflection group <G>, with parameters <para>.
##
CycHeckeOps:=OperationsRecord("CycHeckeOps",AbsHeckeOps);

ComplexGroupOps.Hecke:=function(arg)local H;
  H:=ApplyFunc(AbsCoxOps.Hecke,arg);
  H.operations:=CycHeckeOps;
  return H;
end;

CycHeckeOps.CharTable:=function(H)local W,t,tmp,tbl;
  W:=Group(H);
  tmp:=List(ReflectionType(W),
    function(t)
      if t.series="ST" then return
       ReflTypeOps.HeckeCharTable(t,H.parameter{t.indices});
      else return
       ReflTypeOps.HeckeCharTable(t,H.parameter{t.indices},
          SublistUnbnd(H.rootParameter,t.indices));
      fi;
    end);
  if ForAny(tmp,x->x=false) then return false;fi;
  if Length(tmp)>0 then tbl:=tmp[1];
  else tbl:=CharTable(Group(()));
  fi;
  for t in tmp{[2..Length(tmp)]} do
    Unbind(tbl.powermap);Unbind(t.powermap);
# the above line because of a bug trying to get table of H(FFE)
    tbl:=CharTableDirectProduct(tbl,t); 
    Unbind(tbl.fusionsource); Unbind(tbl.fusions);
  od;
  tbl.parameter:=H.parameter; tbl.rootParameter:=H.rootParameter;
  if IsBound(W.cartan) then tbl.cartan:=W.cartan;fi;
  tbl.irredinfo:=List(CharParams(W),x->rec(charparam:=x,
                                           charname:=CharName(W,x)));
  tbl.operations.StringEntry := function(x)
    if x=0*x then return ".";else return Format(x);fi;end;
  Inherit(tbl,ChevieClassInfo(W));
  if not IsBound(tbl.name) then tbl.name:=tbl.identifier;fi;
  return tbl;
end;

CycHeckeOps.Representations:=function(arg)local H,l,inds,t,rep;
  H:=arg[1]; t:=ReflectionType(Group(H)); l:=List(t,NrConjugacyClasses);
  if Length(arg)=2 then inds:=arg[2];else inds:=[1..Product(l)];fi;
  rep:=function(i)local ind,reps,tensorat,dims,res;
    ind:=CartesianPosition(i,l);
    reps:=List([1..Length(t)],function(i)
      if t[i].series="ST" then return
       ReflTypeOps.HeckeRepresentation(t[i],H.parameter{t[i].indices},ind[i]);
      else return
       ReflTypeOps.HeckeRepresentation(t[i],H.parameter{t[i].indices},
          SublistUnbnd(H.rootParameter,t[i].indices),ind[i]);
      fi;
    end);
    if ForAny(reps,x->x=false) then return false;fi;
    if IsBound(H.unit) then reps:=reps*H.unit;fi;
    if Length(reps)=1 then return reps[1];fi;
    tensorat:=function(i,m,dims) # id(d1)tensor id(d2)..tensor(m) at i tensor..
      return List([1..Product(dims)],k->List([1..Product(dims)],
      function(l)local r,c;
	r:=CartesianPosition(k,dims); c:=CartesianPosition(l,dims);
	if ForAny([1..Length(dims)],j->r[j]<>c[j] and j<>i) then return 0*H.unit;
	else return m[r[i]][c[i]];fi;end));
    end;
    dims:=List(reps,x->Length(x[1]));res:=[];
    for i in [1..Length(t)] do
      res{t[i].indices}:=List(reps[i],m->tensorat(i,m,dims));
    od;
    return res;
  end;
  if IsList(inds) then return List(inds,rep);
  else return rep(inds);
  fi;
end;

CycHeckeOps.CharParams:=H->CharParams(Group(H));

############################################################################
##
#F  SchurElement( <H>, <p> ) .  Schur element of character with .charparam p
##
##  'SchurElement' returns $P/D_\phi$ where P is the
##  Poincare polynomial of <H>, and $\phi$ is the irreducible character
##  of $H$ with .charparam <p> and $D_\phi$ is the generic degree of $\phi$.
##
CycHeckeOps.SchurElement:=function(H,phi)local t;
  if not IsRec(H) or not IsBound(H.parameter) then
    Error("<H> should be a Hecke algebra");
  fi;
  t:=ReflectionType(Group(H));
  t:=List([1..Length(t)],function(i)
      if t[i].series="ST" then return
        CHEVIE.Data("SchurElement",t[i],phi[i],H.parameter{t[i].indices});
      else return
        CHEVIE.Data("SchurElement",t[i],phi[i],H.parameter{t[i].indices},
	      SublistUnbnd(H.rootParameter,t[i].indices));
      fi;
    end);
  if ForAny(t,x->x=false) then return false;
  else return Product(t);
  fi;
end;

CycHeckeOps.SchurElements:=H->List(CharParams(Group(H)),p->SchurElement(H,p));

# The function below, if a representation is known for hecke algebra H
# gives character values on element w of the Braid monoid, else returns false
CycHeckeOps.HeckeCharValues:=function(H,w)local p,W;
  W:=Group(H);
  p:=Position(ChevieClassInfo(W).classtext,w);
  if p<>false then return List(CharTable(H).irreducibles,x->x[p]);fi;
  return List(Representations(H),function(x)if x=false then return false;
   else return CharRepresentationWords(x,[W.rootRestriction{w}])[1];
   fi;end);
end;

###########################################################################
##
#F HeckeCentralMonomials( <H> ) . . . . . . . . scalars of central element 
## . . . . . . . . . . . . . . . . . . . . .  on irreduciblerepresentations
##
## 'HeckeCentralMonomials'  computes  the  scalars  by  which  the center
## of the pure braid group (the square of the longest element  in the
## Iwahori-Hecke case) acts on  the irreducible representations of the 
## Iwahori-Hecke algebra. Cf. BrMi, 4.16 for the formula used

HeckeCentralMonomials:=function(H)local v,W;
  W:=Group(H);v:=List(HyperplaneOrbits(W),ShallowCopy);
  return List(CharTable(W).irreducibles,irr->Product(v,function(C)local q,m;
    q:=H.parameter[W.rootRestriction[C.s]];
    m:=List([0..C.e_s-1],j->(irr[1]+Sum([1..C.e_s-1],
                   l->irr[C.classno[l]]*E(C.e_s)^(-j*l)))/C.e_s);
    return E(irr[1])^(-C.N_s*m*[0..C.e_s-1])*
           Product([1..C.e_s],j->q[j]^(C.N_s*C.e_s*m[j]/irr[1]));
      end));
end;
