#############################################################################
##
#A  coset.g                  CHEVIE library        Frank Luebeck, Jean Michel
##
#A  $Id: coset.g,v 1.1 1997/01/21 13:46:23 gap Exp $
##
#Y  Copyright (C) 1992 - 1996  Lehrstuhl D fur Mathematik, RWTH Aachen, IWR
#Y  der Universitat Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains general functions that deal with cosets of Coxeter 
##  groups.
##

#############################################################################
##
#F  CoxeterCosetOps . . . . . . .  operations record for Coxeter group cosets
##  
##  We  first copy  the  basic  functions   included in 'DomainOps'    and
##  overwrite some of them by more efficient ones.
##  
CoxeterCosetOps:=OperationsRecord("CoxeterCosetOps",SpetsOps);
Inherit(CoxeterCosetOps,HasTypeOps); # we can always classify them

CoxeterCosetOps.CoxeterGroup:=CoxeterCosetOps.Group;
# for compatibility with previous versions of CHEVIE

IsCoxeterCoset:=WF->IsRec(WF) and IsBound(WF.isCoxeterCoset)
  and WF.isCoxeterCoset;

#############################################################################
##
#F CoxeterCosetOps.ReflectionType(  <WF> ) .  . .  . . .  the reflection
#F                                                 type of a CoxeterCoset
##  
##  Gives  a list which describes  WF.  The list  has one element for each
##  orbit of <WF>.phi on the irreducible components of  W. This element is
##  a record with two components:
##  
##    .orbit  The ReflectionType of the orbit (see HasTypeOps.ReflectionType)
##    .twist  A permutation  of  .orbit[1].indices  which describes  the
##            effect of <WF>.phi^Length(orbit) on it.
##  
##  The indices in .orbit are given such that
##      .orbit([2].indices)=OnTuples(.orbit[1].indices,<WF>.phi), etc...
##  
##  If   the    components   in   .orbit    are   of   type    D_4   and
##  <WF>.phi^Length(orbit)  has   order  2  on   .orbit[1].indices  then
##  .orbit[1].indices   is  sorted   such  that   <WF>.phi^Length(orbit)
##  permutes the first two entries.
##  
##  If   the    components   in   .orbit    are   of   type    D_4   and
##  <WF>.phi^Length(orbit)  has   order  3  on   .orbit[1].indices  then
##  .orbit[1].indices   is  sorted   such  that   <WF>.phi^Length(orbit)
##  permutes the entries 1 -> 2 -> 4 -> 1.
##  
CoxeterCosetOps.ReflectionType:=function(WF)
  local W, c, i, type, rls, hi, phifus;

  W:=Group(WF);

  type:=ReflectionType(W);

  c:=List(type,a->Set(W.rootInclusion{a.indices}));
  c:=Cycles(Permutation(WF.phi,c,OnSets),[1..Length(type)]);
  type:=List(c,x->type{x});
  phifus:=WF.phi^MappingPermListList(W.rootInclusion,[1..Length(W.roots)]);
  for c in type do
    for i in [2..Length(c)] do
      c[i].indices:=OnTuples(c[i-1].indices,phifus);
    od;
  od;
  type:=List(type,c->rec(orbit:=c,
     twist:=RestrictedPerm(phifus^Length(c),c[1].indices),
     operations:=ReflTypeOps));

  # some adjustment such that in type ^2D_4 the first two simple
  # roots are permuted by res.twist, and in type ^3D_4 the permutation 
  # of the simple roots is 1 -> 2 -> 4 -> 1:

  type:=List(type,function(a) local b, rf, j,o,i;    
    i:=a.orbit[1].indices;o:=OrderPerm(a.twist);
    if o=1 or a.orbit[1].series<>"D" or Length(i)<>4 then return a;fi;
    if OrderPerm(a.twist)=2 then # ^2D_4
      rf:=Filtered([1..4],x->i[x]<>i[x]^a.twist);
      rf:=Concatenation(rf,[3],Difference([1,2,4],rf));
      b:=Copy(a); # necessary?
      for j in b.orbit do j.indices:=j.indices{rf};od;
      return b;
    fi;
    # triality group ^3D_4
    if i[1]^a.twist=i[2] then return a; fi;
    b:=Copy(a); # necessary?
    for j in b.orbit do j.indices:=j.indices{[1,4,3,2]};od;
    return b;
  end);

##    # some sorting: (now already done in ReflectionType(Group(WF)))
##    hi:=List(type,a->[a.orbit[1].rank,a.orbit[1].series]);
##    SortParallel(hi,type,function(a,b) 
##                      return a[1]>b[1] or (a[1]=b[1] and a[2]<b[2]); end);  
  return type;
end;

CoxeterCosetOps.CoxeterWord:=function(W,w)return CoxeterWord(Group(W),w); end;

CoxeterCosetOps.CoxeterLength:=function(W,w)return CoxeterLength(Group(W),w);
end;

CoxeterCosetOps.LeftDescentSet:=function(W,w)return LeftDescentSet(Group(W),w);
end;

CoxeterCosetOps.FirstLeftDescending:=function(W,w)return
  FirstLeftDescending(Group(W),w);
end;

CoxeterCosetOps.IsLeftDescending:=function(W,w,i)return
  IsLeftDescending(Group(W),w,i);
end;

CoxeterCosetOps.CoxeterElements:=function(arg)local W;
  W:=arg[1];arg[1]:=Group(W);
  return ApplyFunc(CoxeterElements,arg)*W.phi;
end;

CoxeterCosetOps.EltWord:=function(W,w)return EltWord(Group(W),w)*W.phi; end;

CoxeterCosetOps.ReflectionCharValue:=function(W,w)
  return CoxeterGroupOps.ReflectionCharValue(Group(W),w);
end;

CoxeterCosetOps.CoxeterWords:=function(arg)local a;
  a:=ShallowCopy(arg);a[1]:=Group(a[1]);
  return ApplyFunc(CoxeterWords,a);
end;

#############################################################################
##
#F  CoxeterCoset( <W>[, <F0Mat> or  <perm>] ) . . . .  create a CoxeterGroup
#F  coset record
#F  CoxeterCoset( <rec> ) . . . . . . . . . return component <rec>.coxeterCoset
##  
##  In the first form <W> must be a CoxeterGroup record.
##  <F0Mat> must be a square  matrix of rank <W>.rank such that:
##    - <F0Mat> is invertible
##    - <F0Mat> has finite order
##    - <F0Mat>: X->X lets invariant the set of roots of <W> and of Parent(<W>)
##    - TransposedMat(<F0Mat>): Y->Y lets invariant the set of coroots
##        of <W> and of Parent(<W>)
##  
##  If <W>.semisimpleRank  = <W>.rank it   is allowed to give an  argument
##  <perm> (which  is a permutation) instead  of <F0Mat>. Then  <F0Mat> is
##  computed as the  unique matrix which  maps the simple  roots of <W> on
##  the roots given by i^<perm>, i in [1..<W>.rank]. 
##  
##  If only  <W> is  given then the  default for  <F0Mat> is  the identity
##  matrix.
##  
##  'CoxeterCoset' returns a record with the following components:
##    .isDomain     set to true
##    .isFinite     set to true
##    .isCoset      set to true
##    .isCoxeterCoset  set to true
##    .isGroup      set to true, not quite correct, but necessary for using
##                  some dispatcher functions in the GAP library
##    .reflectionGroup      <W> from the argument
##    .F0Mat        as described above
##    .phi          unique element in W.<perm> stabilizing simple roots of <W>
##    .operations   CoxeterCosetOps
##    .callArgs     information used by 'Print'
##  
##  In the second form 'CoxeterCoset' returns the component <rec>.coxeterCoset,
##  if present.
##   
CoxeterCoset:=function(arg)local W, PW, perm, tmp, WF, i;

  # just return  component .coxeterCoset:
  if Length(arg)=1 and IsRec(arg[1]) and IsBound(arg[1].coxeterCoset) then
    return arg[1].coxeterCoset;
  fi;

  W:=arg[1];arg:=arg{[2..Length(arg)]};
  WF:=rec(isDomain:=true,
           isFinite:=true,
           isCoxeterCoset:=true,
           isCoset:=true,
           # unfortunately we have to set this to get the 'CharTable'
           # command working:
           isGroup:=true,
           reflectionGroup:=W);

  if Length(arg)>0 then
    if IsPerm(arg[1]) then 
         WF.F0Mat:=MatXPerm(Parent(W),arg[1]);
	 if W.rank=0 then perm:=();
	 else perm:=PermMatX(W,WF.F0Mat);
	 fi;
    else WF.F0Mat:=arg[1];
      # checking finite order of F0Mat:
      if W.rank>W.semisimpleRank then OrderMat(WF.F0Mat); fi;

      perm:=PermMatX(W,WF.F0Mat);
      if perm=false then
	Error("#I matrix for F0 must normalize set of roots.\n");
      fi;
    fi;
  else WF.F0Mat:=IdentityMat(W.rank);perm:=();
  fi;
  
  PW:=Parent(W);
  # checking if coroots of parent are normalized:
  if IsCoxeterGroup(PW) and PW.rank>0 and  # parent may be Spets
    not IsNormalizing(PW.coroots*PW.simpleCoroots,TransposedMat(WF.F0Mat)) then
    Error("#I transposed of matrix for F0 must normalize set of coroots",
	  " of parent.\n");
  fi;

  WF.phi:=ReducedInRightCoset(W,perm);
  if WF.phi<>perm then WF.F0Mat:=MatXPerm(Parent(W),WF.phi/perm)*WF.F0Mat;fi;

  WF.operations:=CoxeterCosetOps;
  
  # for printing:
  if Length(arg)=0 or W.rank=0 or WF.F0Mat=WF.F0Mat^0 then WF.callArgs:=[];
  else
    # check for simplification:
    tmp:=W.rootInclusion{[1..W.semisimpleRank]};
    if W.rank=W.semisimpleRank and Set(OnTuples(tmp,WF.phi))=Set(tmp) then
      WF.callArgs:=MappingPermListList(tmp, OnTuples(tmp,WF.phi));
    else
      WF.callArgs:=arg[1];
    fi;
  fi;

  WF.name:=ReflectionName(WF);

  # check for cached cosets. Especially useful for the trivial coset
  return CHEVIE.GetCached(W,"CoxeterCosets",WF,function(x,WF)return 
     x.callArgs=WF.callArgs;end);
end;

#############################################################################
##
#F  CoxeterSubCoset(  <WF>, <I>[, <Omega> ][, <w> ] ) . . . . . . subcoset 
#F  of a Coxeter coset
##   
##  The  arguments  <I>  and  (  optional)  <Omega>  must  be  as  in the call
##  'ReflectionSubgroup(Group(<WF>),  <I>[,  <Omega>  ]  )'.  <w>  must  be an
##  element    of    Group(<WF>)    such    that    the    root    system   of
##  'ReflectionSubgroup(Group(<WF>),  <I>[,  <Omega>  ]  )' is invariant under
##  <w>*<WF>.phi. The default for <w> is ().
##  
CoxeterSubCoset:=function(arg)local WF, W, w, tmp, Om, I, res, i;
  if Length(arg)=1 then arg:=arg[1]; fi;
  WF:=Parent(arg[1]);w:=arg[1].phi/WF.phi;W:=Group(WF);
  I:=arg[2];
  if Length(arg)>2 then
    if IsGroup(arg[3]) then Om:=arg[3];
    elif arg[3] in W then w:=arg[3]*w;
    else Error("must give w in Group(WF).\n");
    fi;
  fi;
  if Length(arg)>3 and arg[4] in W then w:=arg[4]*w; fi;
  res:=rec(isDomain:=true, isFinite:=true,
           # unfortunately we have to set this to get through the dispatcher
           # function for 'CharTable':
           isGroup:=true, isCoxeterCoset:=true, isCoset:=true, parent:=WF);
  
  if IsBound(Om) then res.reflectionGroup:=ReflectionSubgroup(W,I,Om);
  else res.reflectionGroup:=ReflectionSubgroup(W,I);
  fi;

  # checking, if w*WF.phi normalizes subroot system:
  tmp:=Set(res.reflectionGroup.rootInclusion);
  if not OnSets(tmp,w*WF.phi)=tmp then
    Error("must give w, such that w * WF.phi normalizes subroot system.\n");
  fi;
  
  res.phi:=ReducedInRightCoset(res.reflectionGroup,w*WF.phi);
  res.F0Mat:=MatXPerm(W,res.phi/WF.phi)*WF.F0Mat;
  res.operations:=CoxeterCosetOps;

  return CHEVIE.GetCached(W,"SubCosets",res,function(x,res)return 
         res.reflectionGroup=x.reflectionGroup and res.F0Mat=x.F0Mat;end);
end;

CoxeterCosetOps.ReflectionSubgroup:=CoxeterSubCoset;
#############################################################################
##
#F  Frobenius( <W> )  . . . . . . . .  returns function which gives the action
#F  of W.phi on a permutation or on the set of roots.
##  
##  
##  The function f returned by 'Frobenius( <W> )' can be applied to:
##    a permutation x:           then f(x) = x^(<W>.phi^-1)
##    a list l of integers:      then f(l) = List(l, i->i^(<W>.phi^-1))
##    a record r with component   
##        .operations.Frobenius: then f(r) = r.operations.Frobenius(<W>,r)
##  
##  If the simple roots are   invariant under <W>.phi then the   second
##  form gives  the action of the Frobenius  on elements of the Coxeter group
##  written as words in the generators.
##  
Frobenius:=function(W)
  local finv;
  finv:=W.phi^-1;
  return function(x)
    if IsPerm(x) then 
      return x^finv;
    elif IsList(x) then 
      return OnTuples(x,finv);
    elif IsRec(x) then
      if IsBound(x.operations) and  IsBound(x.operations.Frobenius) then 
	return x.operations.Frobenius(W,x);
      else
        Error("no method Frobenius for x.\n");
      fi;
   fi;
  end;
end;

# to trick the auto system to do the right thing
CoxeterCosetOps.Hecke:=CoxeterCosetOpsHecke;
CoxeterCosetOps.RelativeCoset:=CoxeterCosetOpsRelativeCoset;
CoxeterCosetOps.ClassInvariants:=CoxeterCosetOpsClassInvariants;
