#############################################################################
##
#A  garside.g                 CHEVIE library                      Jean Michel.
##
#Y  Copyright (C) 2002 -   The CHEVIE Team
##
##  This file contains functions to work with elements of (locally) Garside
##  monoids
#  Maybe the best illustration is the session log below
#  
#  gap> w:=DualBraid(CoxeterGroup("A",4))(1,2,3,4,3);
#  1234.3
#  gap> w*w;
#  1234.1249.34
#  gap> w^4;
#  1234.1249.1249.1249.267.4.
#
#  Elements are stored  as delta^i w_1.w_2...w_n where i  is positive or
#  negative and w_1.w_2...w_n is the  left-greedy canonical form: w_1 is
#  the  longest simple  which is  a left  prefix, then  the rest  of the
#  element is  recursively decomposed  in the same  way. This  is called
#  "left-greedy normal  form". There is no  delta if the monoid  is only
#  locally Garside
#
#  Elements are represented as records with fields:
#
#   .pd       the power of delta (positive or negative -- may be absent)
#   .elm      the canonical left-greedy decomposition (a list of simples)
#   .monoid   a pointer to the monoid info
#   .operations=GarsideEltOps  see below
#
#############################################################################

GarsideEltOps:=OperationsRecord("GarsideEltOps");

# suppress delta and identity
GarsideEltOps.Normalize:=function(b)local res,s,delta; 
  if IsBound(b.monoid.delta) then delta:=b.monoid.delta;else delta:=0;fi;
  res:=[];
  for s in b.elm do
    if s=delta then b.pd:=b.pd+1;
    elif s<>b.monoid.identity then Add(res,s);
    fi;
  od;
  b.elm:=res;return b;
end;

########################################################################
##
#F  GarsideAlpha(b) returns the largest simple (for divisiblity) dividing b
##
GarsideAlpha:=function(b)
  if IsBound(b.pd) then
    if b.pd<0 then Error("GarsideAlpha: element should be positive");
    elif b.pd>0 then return b.monoid.delta;
    fi;
  fi;
  if Length(b.elm)>0 then return b.elm[1];
  else return b.monoid.identity;
  fi;
end;

########################################################################
##
#F  GarsideOmega(b) returns Alpha(b)^-1 * b
##
GarsideOmega:=function(b)
  if IsBound(b.pd) then
    if b.pd<0 then Error("GarsideAlpha: element should be positive");
    elif b.pd>0 then b:=ShallowCopy(b);b.pd:=b.pd-1;return b;
    fi;
  fi;
  if Length(b.elm)>0 then 
    b:=ShallowCopy(b);b.elm:=b.elm{[2..Length(b.elm)]};return b;
  else return b.monoid.identity;
  fi;
end;

########################################################################
##
#F LeftGcd(a,b) returns triple [d,d^-1a,d^-1b] where d=left gcd(a,b)
##
LeftGcd:=function(a,b)local x,res,M,i,chop,m;
  if not IsBound(a.pd) then 
    Error("LeftGcd: both elements should in a Garside group");
  fi;
  m:=Minimum(a.pd,b.pd);
  if m<0 then a:=ShallowCopy(a);a.pd:=a.pd-m;b:=ShallowCopy(b);b.pd:=b.pd-m;fi;
  M:=a.monoid;res:=M.Elt([]);
  chop:=function(b)
    if IsBound(b.pd) and b.pd>0 then return M.Elt(b.elm,b.pd-1);
                                else return M.Elt(b.elm{[2..Length(b.elm)]});
    fi;
  end;
  while true do
    x:=M.LeftGcdSimples(GarsideAlpha(a),GarsideAlpha(b));
    if x[1]=M.identity then 
      if m<0 then res.pd:=res.pd+m;a.pd:=a.pd+m;b.pd:=b.pd+m;fi;
      return [res,a,b];
    else res:=res*M.Elt([x[1]]);
         a:=M.Elt([x[2]])*chop(a);b:=M.Elt([x[3]])*chop(b);
    fi;
  od;
end;

########################################################################
##
#F  LeftLcm(a,b) returns triple [m,ma^-1,mb^-1] where m=Left lcm(a,b)
##
LeftLcm:=function(a,b)local M,compl,m,x;
  if not IsBound(a.pd) then 
    Error("LeftLcm: both elements should be in a Garside group");
  fi;
  M:=a.monoid;
  compl:=function(b,m)local k; # a such that a*b=delta^m
   k:=Length(b.elm);
   return M.Elt(List([k,k-1..1],
     i->M.DeltaAction(M.RightComplementToDelta(b.elm[i]),-b.pd-i)),m-b.pd-k);
  end;
  m:=Maximum(a.pd+Length(a.elm),b.pd+Length(b.elm));
  x:=LeftGcd(compl(a,m),compl(b,m));
  return [x[2]*a,x[2],x[3]];
end;

#########################################################################
##
#F ReversedWord
##
ReversedWord:=function(b)
  if not IsBound(b.monoid.Reverse) then
    Error(b.monoid," does not have a Reverse operation");
  fi;
  return b.monoid.Reverse(b);
end;

########################################################################
##
#F  LeftLcm(a,b) returns triple [m,a^-1m,b^-1m] where m=right lcm(a,b)
##
RightLcm:=function(a,b)
  return List(LeftLcm(ReversedWord(a),ReversedWord(b)),ReversedWord);
end;

########################################################################
##
#F  RightGcd(a,b) returns triple [d,ad^-1,bd^-1] where d=right gcd(a,b)
##
RightGcd:=function(a,b)
  return List(LeftGcd(ReversedWord(a),ReversedWord(b)),ReversedWord);
end;

#############################################################################
##
#F  AsFraction(b) . . . . . . . . . . . .  Fraction normal form. Returns
##  a pair of elements x,y of monoid with trivial left gcd and b=x^-1 y
##
AsFraction:=function(b)local neg,M;
  M:=b.monoid;
  if not IsBound(M.delta) then Error("AsFraction: monoid should be Garside");fi;
  neg:=function(elm)local k;
    k:=Length(elm);
    return List([k,k-1..1],i->M.DeltaAction(M.RightComplementToDelta(elm[i]),k-i));
  end;
  if b.pd>0 then return [M.Elt([]),b];
  elif b.pd>=-Length(b.elm) then return
  [M.Elt(neg(b.elm{[1..-b.pd]})),M.Elt(b.elm{[1-b.pd..Length(b.elm)]})];
  else  return [ M.Elt(List(neg(b.elm),x->M.DeltaAction(x,b.pd+Length(b.elm))),
    -b.pd-Length(b.elm)),M.Elt([])];
  fi;
end;

############################################################################
##
#F  AsWord(w) . . . . . . . . . . . . . . . .  sequence of atoms
##
##  w  is  an  element  of  a  Garside  monoid.  AsWord(w)  returns  the
##  expression of w as  a list of atoms given as  a sequence of integers
##  (the index in the list of  atoms). As a special case (symmetric with
##  the input methods of 'Braid') an  element which is not in the monoid
##  is returned in fraction normal form,  the inverse of the atoms coded
##  as negative integers.

AsWord:=function(b)local res,i,M;
  M:=b.monoid;
  res:=[];
  if IsBound(b.pd) then
    if b.pd<0 then 
      b:=AsFraction(b);
      return Concatenation(-Reversed(AsWord(b[1])),AsWord(b[2]));
    fi;
    for i in [1..b.pd] do Append(res,M.AtomListSimple(M.delta));od;
  fi;
  for i in b.elm do Append(res,M.AtomListSimple(i));od;
  return res;
end;

GarsideEltOps.Format:=function(b,option)local M,res,p,SeqCollected,W;
  M:=b.monoid;
  SeqCollected:=function(l)local res,v,i;
    if l=[] then return [];fi;
    res:=[];v:=[l[1],1];
    for i in [2..Length(l)] do
      if l[i]=v[1] then v[2]:=v[2]+1; else Add(res,v);v:=[l[i],1]; fi;
    od;
    Add(res,v);
    return res;
  end;
  p:=function(b)local l;
    if IsBound(option.Brieskorn) then
      return Join(List(SeqCollected(BrieskornNormalForm(b)),function(x)local s;
        W:=M.group;
        if x[1]=W.rootInclusion{W.generatingReflections} then s:=M.stringDelta;
	elif Length(x[1])=1 then s:=String(x[1][1]);
	else s:=SPrint("w_",IntListToString(x[1]));
	fi;
	return StringMonomial(s,x[2],option);end),".");
    fi;
    l:=List(b.elm,v->M.AtomListSimple(v));
    if IsBound(option.GAP) then
      l:=Join(Concatenation(l),",");
    else
      l:=Join(List(l,x->IntListToString(x,"[]")),".");
    fi;
    if not IsBound(b.pd) or b.pd=0 then return l;fi;
    if IsBound(option.GAP) then return SPrint("[",l,"],",b.pd);
    else
      if l<>"" then l:=Concatenation(".",l);fi;
      if IsBound(option.Greedy) then return SPrint(M.stringDelta,"^",b.pd,l);
      else return SPrint(Join(List([1..b.pd],i->M.stringDelta),"."),l);
      fi;
    fi;
  end;
  if IsBound(option.Greedy) or not IsBound(b.pd) then b:=["",p(b)];
  else b:=List(AsFraction(b),p);
  fi;
  if IsBound(option.GAP) then res:=List(b,x->SPrint("B(",x,")"));
       res:=[SPrint(res[1],"^-1"),"*",res[2]];
  else res:=[SPrint("(",b[1],")^-1"),".",b[2]];
  fi;
  if b[1]<>"" then 
    if b[2]<>"" then return Concatenation(res);
    else return res[1];
    fi;
  elif b[2]<>"" or IsBound(option.GAP) then return res[3];
  else return ".";
  fi;
end;

GarsideEltOps.String:=b->GarsideEltOps.Format(b,rec());

#F Print(b) print b (using Format)

GarsideEltOps.Print:=function(b)Print(Format(b,CHEVIE.PrintGarside));end;

if not IsBound(CHEVIE.PrintGarside) then CHEVIE.PrintGarside:=rec();fi;

GarsideEltOps.\*:=function(a,b)local M,x,res;
  if IsList(a) then return List(a,x->x*b);
  elif IsList(b) then return List(b,x->a*x);
  fi;
  M:=a.monoid;
  if M<>b.monoid then 
    Error("not elements of the same garside monoid or group");
  fi;
  res:=M.Elt(ShallowCopy(a.elm));
  if IsBound(M.delta) then
    res.pd:=a.pd+b.pd;res.elm:=List(res.elm,x->M.DeltaAction(x,b.pd));
  fi;
  for x in b.elm do res.elm:=M.AddToNormal(res.elm,x); od;
  return GarsideEltOps.Normalize(res);
end;

GarsideEltOps.\/:=function(a,b)return a*b^-1;end;

GarsideEltOps.\^:=function(b,n)local M,p,res,i,k;
  if IsInt(n) then
    M:=b.monoid;
    if n<0 then
      if not IsBound(M.delta) then 
        Error("negative power in a monoid which is not Garside");
      fi;
      k:=Length(b.elm);
      b:=M.Elt(List([k,k-1..1],
         i->M.DeltaAction(M.RightComplementToDelta(b.elm[i]),-i-b.pd)),-b.pd-k);
      n:=-n;
    fi;
    p:=b;res:=M.Elt([]);
    while n>0 do
     if n mod 2 <> 0 then res:=res*p;fi;
     p:=p*p;
     n:=QuoInt(n,2);
    od;
    return res;
  else return n^-1*b*n;
  fi;
end;

CompleteGarsideRecord:=function(M)
  M.nrAtoms:=Length(M.atoms);
  if not IsBound(M.Elt) then
    M.Elt:=function(arg)local res;
       res:=rec(elm:=arg[1],operations:=GarsideEltOps,monoid:=M);
       if Length(arg)>1 then res.pd:=arg[2];
       elif IsBound(M.delta) then res.pd:=0;
       fi;
       return res;
    end;
  fi;
  if not IsBound(M.AtomListSimple) then
    M.AtomListSimple:=function(s)local i,res;
      res:=[]; i:=1;
      while i<=M.nrAtoms do
        if M.IsLeftDescending(s,i) then s:=M.atoms[i]^-1*s;Add(res,i);i:=1;
        else i:=i+1;
        fi;
      od;
      if s<>M.identity then Error("monoid seems not Garside");fi;
      return res;
    end;
  fi;
  if IsBound(M.delta) then
    if not IsBound(M.RightLcmSimples) then
      M.RightLcmSimples:=function(a,b)local x;
	x:=M.RightGcdSimples(M.RightComplementToDelta(a),
			     M.RightComplementToDelta(b));
	return [a*x[2],x[2],x[3]];
      end;
    fi;
    if not IsBound(M.LeftLcmSimples) then
      M.LeftLcmSimples:=function(a,b)local x;
	x:=M.LeftGcdSimples(M.LeftComplementToDelta(a),
	                    M.LeftComplementToDelta(b));
	return [x[2]*a,x[2],x[3]];
      end;
    fi;
  fi;
  if not IsBound(M.LeftGcdSimples) then
    M.LeftGcdSimples:=function(a,b)local i,s,x;x:=M.identity;
      repeat
        s:=M.identity;
        for i in [1..M.nrAtoms] do
	  if M.IsLeftDescending(a,i) and M.IsLeftDescending(b,i) then
	    s:=s*M.atoms[i];a:=M.atoms[i]^-1*a;b:=M.atoms[i]^-1*b;
          fi;
        od;
        x:=x*s;
      until s=M.identity;
      return [x,a,b];
    end;
  fi;
  if not IsBound(M.RightGcdSimples) then
    if IsBound(M.ReverseSimple) then
    M.RightGcdSimples:=function(a,b)return List(M.LeftGcdSimples(
      M.ReverseSimple(a),M.ReverseSimple(b)),M.ReverseSimple);end;
    else
    M.RightGcdSimples:=function(a,b)local i,x;
      i:=1;x:=M.identity;
      while i<=M.nrAtoms do
	if M.IsRightDescending(a,i) and M.IsRightDescending(b,i) then
          x:=M.atoms[i]*x;a:=a/M.atoms[i];b:=b/M.atoms[i];
	  i:=1;
	else i:=i+1;
	fi;
      od;
      return [x,a,b];
    end;
    fi;
  fi;
  if not IsBound(M.alpha2) then
    # Given x,v simples returns x', v' such that x' * v' = x * v and where
    #  x' = x * (longest subword z of v such that xz is simple)
    M.alpha2:=function(x,v)local i;
      if IsBound(M.delta) then
        i:=M.LeftGcdSimples(M.RightComplementToDelta(x),v){[1,3]};
        return [x*i[1],i[2]];
      else
	i:=1;
	repeat
	  if M.IsLeftDescending(v,i) and M.IsRightAscending(x,i)
	  then x:=x*M.atoms[i];v:=M.atoms[i]^-1*v;i:=1;
	  else i:=i+1;
	  fi;
	until i>M.nrAtoms;
	return [x,v];
      fi;
    end;
  fi;
  if not IsBound(M.AddToNormal) then
    # given a normal form v and a simple x return the normal form of v.x 
    # (discarding a final 1 but not an initial delta in the result)
    M.AddToNormal:=function(v,x)local i,l;
      l:=Length(v);if l=0 then return [x];fi;
      x:=M.alpha2(v[l],x);
      if x[2]<>M.identity then Add(v,x[2]);fi;
      v[l]:=x[1];
      for i in [l,l-1..2] do
        x:=M.alpha2(v[i-1],v[i]);
        if x[1]=v[i-1] then return v;fi;
        v{[i-1,i]}:=x;
      od;
      return v;
    end;
  fi;
end;

#############################################################################
##
#F  GarsideWords(M,l)  . . . . . . . . . . .  monoid elements of length l
##
##  M is a (locally) Garside monoid with an additive length function, 
##  GarsideWords(W,l) returns the set of elements of M of length l.

GarsideWords:=function(M,l)local i,bb,w,b,B,s,r,rr,RA,lb;
  if l=0 then return [M.Elt([])];fi;
  if not IsBound(M.words) then 
    M.words:=[Set(List(M.atoms,x->M.Elt([x])))];
  fi;
  RA:=w->Filtered([1..M.nrAtoms],i->M.IsRightAscending(w,i));
  if not IsBound(M.words[l]) then 
    M.words[l]:=[];
    for b in GarsideWords(M,l-1) do
      lb:=Length(b.elm);
      if lb=0 then r:=[];
      else
        r:=RA(b.elm[lb]);if lb=1 then rr:=[]; else rr:=RA(b.elm[lb-1]);fi;
        for s in r do
          w:=b.elm[lb]*M.atoms[s];
          if not ForAny(rr,i->M.IsLeftDescending(w,i)) then
            bb:=ShallowCopy(b);bb.elm:=ShallowCopy(bb.elm);
            bb.elm[Length(bb.elm)]:=w;
            AddSet(M.words[l],GarsideEltOps.Normalize(bb));
          fi;
        od;
      fi;
      for s in M.atoms{Difference([1..M.nrAtoms],r)} do
        bb:=ShallowCopy(b);bb.elm:=Concatenation(bb.elm,[s]);
        AddSet(M.words[l],GarsideEltOps.Normalize(bb));
      od;
    od;
  fi;
  return M.words[l]; 
end;

# all divisors of the simple s
SimpleLeftDivisors:=function(M,s)local res,i,rest,x,new;
  rest:=[[M.identity,s]];
  res:=[];
  while Length(rest)>0 do
    Add(res,rest);
    new:=[];
    for x in rest do
      for i in [1..M.nrAtoms] do
        if M.IsLeftDescending(x[2],i) then
          Add(new,[x[1]*M.atoms[i],M.atoms[i]^-1*x[2]]);
        fi;
      od;
    od;
    rest:=Set(new);
#   InfoChevie("new:",Length(new)," different:",Length(rest),"\n");
  od;
# return List(res,x->GarsideEltOps.Normalize(M.Elt([x[1]])));
  return List(res,x->List(x,y->y[1]));
end;

#############################################################################
##                                                                         ##
##  First example of Garside monoid: the Artin-Tits braid monoid           ##
##                                                                         ##
#############################################################################
#  Maybe the best illustration is the session log below
#  
#  gap> w:=Braid(CoxeterGroup("A",4))(1,2,3,4,3);
#  12343
#  gap> w*w;
#  12132432.43
#  gap> w^4;
#  w0^2.
#############################################################################

BraidOps:=OperationsRecord("BraidOps",GarsideEltOps);

BraidMonoid:=function(W)local M,ILD,l;
  if not IsCoxeterGroup(W) then Error("W should be a Coxeter Group");fi;
  if IsBound(W.braidMonoid) then  return W.braidMonoid;fi;
  ILD:=W.operations.IsLeftDescending; # this shaves some dispatching overhead 
  M:=rec(group:=W,
    operations:=rec(Print:=function(M)Print("BraidMonoid(",W,")");end),
    IsLeftDescending:=function(w,i)return ILD(W,w,i);end,
    IsLeftAscending:=function(w,i)return not ILD(W,w,i);end,
    IsRightAscending:=function(w,i)return not ILD(W,w^-1,i);end,
    IsRightDescending:=function(w,i)return ILD(W,w^-1,i);end,
    AtomListSimple:=v->W.operations.CoxeterWord(W,v),
    atoms:=W.reflections{W.generatingReflections},
    identity:=W.identity
  );
  if IsFinite(W) then 
    M.delta:=LongestCoxeterElement(W);
    l:=W.reflections{W.generatingReflections};
    M.orderDelta:=OrderPerm(PermListList(l,List(l,x->x^M.delta)));
    M.stringDelta:="w0";
    M.DeltaAction:=function(s,i)return s^(M.delta^(i mod M.orderDelta));end;
    M.RightComplementToDelta:=x->x^-1*M.delta;
    M.LeftComplementToDelta:=x->M.delta/x;
  fi;
  M.Elt:=function(arg)local res;
     res:=rec(elm:=arg[1],operations:=BraidOps,monoid:=M);
     if Length(arg)>1 then res.pd:=arg[2];
     elif IsBound(M.delta) then res.pd:=0;
     fi;
     return res;
  end;
  M.ReverseSimple:=x->x^-1;
  M.Reverse:=function(b)local res;
    if Length(b.elm)=0 then return b;fi;
    res:=Product(Reversed(b.elm),x->M.Elt([M.DeltaAction(x^-1,b.pd)]));
    res.pd:=b.pd;
    return res;
  end;
  CompleteGarsideRecord(M);
  if IsBound(W.operations.LeftGcdSimples) then
    M.LeftGcdSimples:=function(x,v)return W.operations.LeftGcdSimples(W,x,v);end;
  fi;
  W.braidMonoid:=M;
##  M.B: a function which can be called in the forms
#F     B(<element of W> [, pw0])  
#F  or B([s1,...,sn] [, pw0])  
#F  or B(s1,...sn)  
##
##  returns  an element  of the  Artin braid  monoid of  W: either  the
##  reduced braid  corresponding to an  element of the Coxeter  group W,
##  or  (non  necessarily  reduced)  braid  product  of  the  generators
##  corresponding to the  labels for Coxeter generators  s1,...,sn. As a
##  special case, if any  of the s_i is a negative  integer, it is taken
##  to represent the inverse of the corresponding generator.
##  If  pw0 (a  positive or  negative integer)  is given,  the resulting
##  element is multiplied by w0^pw0.
#############################################################################
  M.B:=function(arg)local res,s,r,pd;
    res:=M.Elt([]);
    r:=x->M.Elt([W.reflections[W.operations.ReflectionFromName(W,x)]]);
    if Length(arg)<>0 then
      if not IsWordFor(W,arg[1]) and not arg[1] in W.reflectionsLabels 
          and not (IsInt(arg[1]) and -arg[1] in W.reflectionsLabels)
      then
      # assume h is a group element
        if arg[1]<>W.identity then res.elm:=[arg[1]];fi;
        if Length(arg)=2 then res.pd:=arg[2];fi;
        return GarsideEltOps.Normalize(res);
      else
        pd:=0;
        if IsList(arg[1]) then 
          if Length(arg)=2 then pd:=arg[2];fi;
          arg:=arg[1];
        fi;
        for s in Reversed(arg) do # faster in reversed order (see \*)
          if IsInt(s) and s<0 then res:=r(-s)^-1*res; else res:=r(s)*res;fi;
        od;
        if IsBound(res.pd) then res.pd:=res.pd+pd;fi;
      fi;
    fi;
    return res;
  end;
  return M;
end;

#############################################################################
##
#F  Braid(W) . . . . . . . . . . . . . returns braid element making function
##
#############################################################################
Braid:=W->BraidMonoid(W).B;

#############################################################################
##
#F  EltBraid(w)  . . . . . . . . . . . . . . .  Image in the Coxeter group.
##
##  returns the image in w.monoid.group of w (works also for Dual braids)
##
EltBraid:=function(x)local res,M;
  M:=x.monoid;
  if IsBound(x.pd) then res:=M.delta^x.pd;
  else res:=M.identity;
  fi;
  if Length(x.elm)=0 then return res;
  else return res*Product(x.elm);
  fi;
end;

BraidOps.Frobenius:=function(W,x)
  x:=ShallowCopy(x);x.elm:=List(x.elm,y->Frobenius(W)(y));
  return x;
end;

# end of functions which work also for DualBraidOps

BraidOps.BrieskornNormalForm:=function(b)local W,res,v;
  W:=b.monoid.group;
  res:=[];
  if IsBound(b.pd) then
    if b.pd<0 then Error(b," should be in the Braid Monoid");fi;
    for v in [1..b.pd] do Add(res,W.reflectionsLabels);od;
  fi;
  for v in b.elm do Append(res,BrieskornNormalForm(W,v));od;
  return res;
end;
  
###########################################################################
##
#F GoodCoxeterWord( <W>, <w> ) . . . . . . . . . . . . check if w is a good 
## . . . . . . . . element in the braid group in the sense of [Geck-Michel]
##
## 'GoodElement' checks if w represents
## a good element in the braid group, i.e., if w^d (where d is the order of
## the image of w in W) is a product of longest elements in a 
## decreasing chain of parabolic subgroups of W.  If this is true, then the
## list of the corresponding subsets of the generators is returned.
## Otherwise, false is returned.

GoodCoxeterWord:=function(W,w)local res,i,bw,d;
  bw:=Braid(W)(w);
  d:=Order(W,EltBraid(bw));
  InfoChevie("#I ",bw,"^",d,"=");
  res:=BrieskornNormalForm(bw^d);
  if not ForAll([1..Length(res)-1],i->IsSubset(res[i],res[i+1])) then
         InfoChevie("#I ",bw," is not good\n");return false;
  fi;
  res:=Collected(res);
  Sort(res,function(x,y)return IsSubset(x[1],y[1]);end);
  for i in res do
     InfoChevie("w_",IntListToString(i[1],"{}"),"^",i[2],".");
  od;
  InfoChevie("\n");
  return res;
end;

#############################################################################
##
##  Functions for the dual braid monoid
##
#  Maybe the best illustration is the session log below
#  
#  gap> w:=DualBraid(CoxeterGroup("A",4))(1,2,3,4,3);
#  124.34
#  gap> w*w;
#  c.[5,6,10].34.3
#  gap> w^4;
#  c.c.c.c.c
#
#############################################################################

ReflectionLength:=function(W,w)
  if not IsBound(W.reflectionLengths) then
    W.reflectionLengths:=List(ReflectionEigenvalues(W),x->Number(x,y->y<>1));
  fi;
  return W.reflectionLengths[PositionClass(W,w)];
end;

# ReflectionWord(W,w[,list of reflections])
ReflectionWord:=function(arg)local W,w,ref,res,ref,l,r;
  W:=arg[1];w:=arg[2];res:=[];
  if Length(arg)=3 then ref:=arg[3];else ref:=Reflections(W);fi;
  l:=ReflectionLength(W,w);
  while l>0 do
    r:=PositionProperty(ref,x->ReflectionLength(W,x^-1*w)<l);
    if r=false then Error(w," is not a dual simple");fi;
    Add(res,r);
    w:=ref[r]^-1*w;
    l:=l-1;
  od;
  return res;
end;

if false then
PoincareDualSimples:=function(W)local d,M;
  d:=ReflectionDegrees(W);
  InfoChevie("#  predicted=",Product(d+Maximum(d))/Product(d),"\n");
  M:=DualBraidMonoid(W);
  d:=SimpleLeftDivisors(M,M.delta);
  return Sum([1..Length(d)],i->X(Rationals)^(i-1)*Length(d[i]));
end;
fi;

# BipartiteDecomposition a decomposition of the generating reflections
# of W in two sets L, R such that the subgroups W_L and W_R are abelian
BipartiteDecomposition:=function(W)local L,R,comm,r,rest;
  L:=[];R:=[];rest:=Set(W.generatingReflections);
  comm:=function(x,y)x:=W.reflections[x];y:=W.reflections[y];return x*y=y*x;end;
  while Length(rest)>0 do
    r:=PositionProperty(rest,x->ForAny(L,y->not comm(x,y)));
    if r<>false then 
      if ForAny(R,y->not comm(rest[r],y)) then
        Error(W," has no bipartite decomposition");
      fi;
      Add(R,rest[r]); SubtractSet(rest,[rest[r]]);
    else r:=PositionProperty(rest,x->ForAny(R,y->not comm(x,y)));
      if r<>false then Add(L,rest[r]); SubtractSet(rest,[rest[r]]);
      else Add(L,rest[1]); SubtractSet(rest,[rest[1]]);
      fi;
    fi;
  od;
  return [W.rootInclusion{L},W.rootInclusion{R}];
end;

DualBraidMonoid:=function(arg)local W,new,M,n,r;
  W:=arg[1];
  M:=rec(group:=W,
    operations:=rec(Print:=function(M)Print("DualBraidMonoid(",W,",",
      M.deltaword,")");end),
    IsLeftDescending:=function(simp,i)
      return ReflectionLength(W,M.atoms[i]^-1*simp)<ReflectionLength(W,simp);
    end,
    IsRightDescending:=function(simp,i)
      return ReflectionLength(W,simp/M.atoms[i])<ReflectionLength(W,simp);
    end,
    IsRightAscending:=function(simp,i) 
      return M.IsLeftDescending(simp^-1*M.delta,i);
    end,
    IsLeftAscending:=function(simp,i) 
      return M.IsRightDescending(M.delta*simp^-1,i);
    end,
    identity:=W.identity
  );
  if not IsFinite(W) then Error("W should be finite");fi;
  if Length(arg)=1 then
    if not IsCoxeterGroup(W) then Error(
  "For a non-Coxeter Group a decomposition of a coxeter element must be given");
    fi;
    M.deltaword:=Concatenation(BipartiteDecomposition(W));
    M.delta:=EltWord(W,M.deltaword);
  elif IsWordFor(W,arg[2]) then
    M.deltaword:=arg[2];M.delta:=EltWord(W,M.deltaword);
  else M.delta:=arg[2];M.deltaword:=CoxeterWord(W,M.delta);
  fi;
  if IsCoxeterGroup(W) then
    M.AtomListSimple:=v->ReflectionWord(W,v);
    M.atoms:=Reflections(W){[1..W.N]};
  else
    n:=ReflectionLength(W,M.delta);
    M.atoms:=Filtered(Set(Reflections(W)),r->ReflectionLength(W,M.delta/r)<n);
    SortParallel(List(M.atoms,x->Position(Reflections(W),x)),M.atoms);
    M.AtomListSimple:=v->ReflectionWord(W,v,M.atoms);
  fi;
  M.orderDelta:=Order(W,M.delta);
  M.stringDelta:="c";
  M.DeltaAction:=function(s,i)return s^(M.delta^(i mod M.orderDelta));end;
  M.RightComplementToDelta:=x->x^-1*M.delta;
  M.LeftComplementToDelta:=x->M.delta/x;
  M.Reverse:=function(b)local res;
    if Length(b.elm)=0 then return M.revMonoid.Elt([],b.pd);fi;
    res:=Product(Reversed(b.elm),
      x->M.revMonoid.Elt([M.revMonoid.DeltaAction(x^-1,b.pd)]));
    res.pd:=b.pd;
    return res;
  end;
##  M.B: a function which can be called in the forms
#F     B(<element of W> [, pd])  
#F  or B([s1,...,sn] [, pd])  
#F  or B(s1,...sn)  
#
#  and returns  an element  of the  Dual braid monoid  of W:  either the
#  reduced dual braid  corresponding to an element of  the Coxeter group
#  W,  or (non  necessarily reduced)  braid product  of the  reflections
#  corresponding to the  labels for reflections s1,...,sn.  As a special
#  case,  if any  of the  s_i  is a  negative  integer, it  is taken  to
#  represent the inverse of the corresponding atom. 
#  If  pd (a  positive  or  negative integer)  is  given, the  resulting
#  element is multiplied by c^pd.
  M.B:=function(arg)local res,s,p;
    res:=M.Elt([]);
    if Length(arg)<>0 then
      if not IsList(arg[1]) and not (IsInt(arg[1]) and 
        AbsInt(arg[1]) in [1..W.N]) then
      # assume h is a group element
        arg:=ReflectionWord(W,arg[1],M.atoms);
      elif Length(arg)>0 and IsList(arg[1]) then 
        if Length(arg)=2 then res.pd:=arg[2];fi;
        arg:=arg[1];
      fi;
      for s in Reversed(arg) do # faster in reversed order (see \*)
        if IsInt(s) and s<0 then p:=-1;s:=-s;else p:=1;fi;
        res:=M.Elt([M.atoms[s]])^p*res; 
      od;
    fi;
    return res;
  end;
  CompleteGarsideRecord(M);
  if Length(arg)=3 then M.revMonoid:=arg[3];
  else M.revMonoid:=DualBraidMonoid(W,Reversed(M.deltaword),M);
  fi;
  return M;
end;

#############################################################################
##
#F  DualBraid(W [, coxeter]) . . . . . . . returns Dual-Braid making function
##
#############################################################################
DualBraid:=function(arg)return ApplyFunc(DualBraidMonoid,arg).B;end;

#############################################################################
# Empirical reduction of the complexity of a generating set <bs> of
# a given subgroup of a Garside group
#############################################################################
ShrinkGarsideGeneratingSet:=function(bs)local b,a,ai,aib,bai,try,simplify,j,i,f;
  try:=function(p)local pos;
    p:=f(p);
#   Print("p=",p,"\n");
    pos:=Position(bs,p);
    if pos<>false and pos<>j then
      InfoChevie(" eliminated");
      if pos>j then bs:=bs{Concatenation([1..pos-1],[pos+1..Length(bs)])};
      else bs:=bs{Concatenation([1..j-1],[j+1..Length(bs)])};
      fi;
      return true;
    fi;
    if p[1]>bs[j][1] or (p[1]=bs[j][1] and p[2]<=bs[j][2]) 
    then return false;fi;
    InfoChevie(" reduced to <",p[1],".",p[2],"> ");
    bs[j]:=p;
    simplify:=true;
    return false;
  end;
  f:=function(b)local l;
    l:=List(AsFraction(b),x->Length(AsWord(x)));
    if l[1]>l[2] then return [l[1]+l[2],l[2],b^-1];
                 else return [l[1]+l[2],l[1],b];
    fi;
  end;
  bs:=List(bs,f);
  repeat
    simplify:=false;
    bs:=Set(bs);
    InfoChevie("#I total length ",Sum(bs,x->x[1]),
               "   maximal length ",bs[Length(bs)][1],"\n");
    j:=Length(bs);
    for j in [Length(bs),Length(bs)-1..2] do
      InfoChevie("#I <",j,":",bs[j][1],".",bs[j][2],">");
      i:=0;
      repeat i:=i+1;a:=bs[i][3];ai:=a^-1;b:=bs[j][3];aib:=ai*b;bai:=b*ai;
        InfoChevie(i,".\c");
#       Print("i=",i,"bs[i]=",bs[i]," bs[j]=",bs[j],"\n");
      until bs[i][1]>bs[j][1] or (bs[i][1]=bs[j][1] and bs[i][2]>=bs[j][2]) 
        or try(aib) or try(bai) or try(aib*a) or try(a*bai) or 
          try(a*b) or try(b*a);
      InfoChevie("\n");
    od;
  until not simplify;
  return List(bs,x->x[3]);
end;

#############################################################################
# returns a presentation of the Garside group corresponding to the monoid M 
# (by theorem 4.1 of Dehornoy-Paris)
Presentation:=function(M)local F,rels,i,j,v;
  F:=FreeGroup(Length(M.atoms));
  rels:=[];
  for i in [1..Length(M.atoms)] do
    for j in [1..i-1] do
      v:=M.LeftLcmSimples(M.atoms[i],M.atoms[j]);
      v:=List(v{[2,3]},M.AtomListSimple);
      Add(v[1],i);Add(v[2],j);
      Add(rels,Product(v[1],i->F.(i))*Product(v[2],i->F.(i))^-1);
    od;
  od;
  return PresentationFpGroup(F/rels);
end;
