#############################################################################
##
#A  gencox.g              CHEVIE library                          Jean Michel
##
##
#Y  Copyright (C) 1992 - 1999  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains functions for various ways of defining some
##  abstract Coxeter groups.
##

GenCoxOps:=OperationsRecord("GenCoxOps",MatGroupOps);
Inherit(GenCoxOps,AbsCoxOps);
# inherit from MatGroupOps and AbsCoxOps

GenCoxOps.Print:=function(W)
  Print("CoxeterGroupByCoxeterMatrix(");Print(CoxeterMatrix(W));Print(")");
end;

GenCoxOps.IsLeftDescending:=function(W,w,i)
  if  ForAll(w[i],IsInt) then return Sum(w[i])<0;
  else return evalf(Sum(w[i]))<evalf(0);
  fi;
end;

#############################################################################
##
#F  CoxeterGroupByCoxeterMatrix(<M>) . . . . Makes the Coxeter group defined
#F    by Coxeter matrix M. An infinite entry is represented by 0.
##
CoxeterGroupByCoxeterMatrix:=function(M)local tmp,W,i,j;
  tmp:=rec();
  tmp.nbGeneratingReflections:=Length(M);
  tmp.generatingReflections:=[1..tmp.nbGeneratingReflections];
  tmp.identity:=M^0;
  tmp.coxeterMat:=M;
  tmp.cartan:=CartanMatFromCoxeterMatrix(M);
  tmp.reflections:=List(tmp.generatingReflections,i->M^0);
  for i in tmp.generatingReflections do 
    for j in tmp.generatingReflections do 
      tmp.reflections[i][i][i]:=-1;
      if i<>j then
	if M[i][j]=0 then
	  tmp.reflections[i][j][i]:=2;
	else
	  tmp.reflections[i][j][i]:=E(2*M[i][j])+E(2*M[i][j])^-1;
	fi;
      fi;
    od;
  od;
  W:=ApplyFunc(Group,tmp.reflections);
  Inherit(W,tmp);
  AbsCoxOps.CompleteCoxeterGroupRecord(W);
  W.operations:=GenCoxOps;
  return W;
end;

#############################################################################
##
#F  CoxeterGroupByCartanMatrix(<M>) . . . . Makes the Coxeter group defined
#F    by Cartan matrix M
##
CoxeterGroupByCartanMatrix:=function(M)local tmp,W,k;
  if ForAll(CoxeterTypeFromCartanMat(M),x->x<>false) then 
     return CoxeterGroup(M);fi;
  tmp:=rec();
  tmp.nbGeneratingReflections:=Length(M);
  tmp.generatingReflections:=[1..tmp.nbGeneratingReflections];
  tmp.identity:=M^0;
  tmp.cartan:=M;
  tmp.reflections:=List(tmp.generatingReflections,i->M^0);
  M:=M^0-M;
  for k in [1..tmp.nbGeneratingReflections] do 
    tmp.reflections[k]{tmp.generatingReflections}[k]:=M[k]; 
  od;
  W:=ApplyFunc(Group,tmp.reflections);
  Inherit(W,tmp);
  W.operations:=GenCoxOps;
  AbsCoxOps.CompleteCoxeterGroupRecord(W);
  return W;
end;

#############################################################################
#
# Example I: Symmetric groups as Coxeter groups
#
CoxeterGroupSymmetricGroup:=function(n)local W;
  W:=SymmetricGroup(n);
  W.reflections:=List([2..n],i->(i-1,i));
  W.operations.IsLeftDescending:=function(W,w,i) return i^w>(i+1)^w;end;
  AbsCoxOps.CompleteCoxeterGroupRecord(W);
  W.operations.FirstLeftDescending:=function(W,w)local i;
    for i in [1..n-1] do if i^w>(i+1)^w then return i;fi;od;return false;end;
  W.operations.ReflectionDegrees:=W->[2..n];
  W.operations.Reflection:=function(W,i)local g,k;
    k:=0;g:=1;
    while n-g<i do i:=i-n+g;g:=g+1;od;
    return (i,i+g);
  end;
  W.operations.Reflections:=W->List([1..n*(n-1)/2],i->Reflection(W,i));
  W.operations.ReflectionCharValue:=
                         function(W,w)return Number([1..n],i->i=i^w)-1;end;
  W.reflectionLengths:=List(ConjugacyClasses(W),x->Sum(CycleLengths(
    Representative(x),[1..n])-1));
  W.operations.String:=W->SPrint("CoxeterGroupSymmetricGroup(",n,")");
  # next saves a factor 2 in Braid multiplications
  W.operations.LeftGcdSimples:=function(W,x,v)local res,i,s;res:=();
    repeat s:=();
      for i in [1..n-1] do
	if i^v>(i+1)^v and i^x>(i+1)^x then s:=s*(i,i+1);fi;
      od;
      res:=res*s;s:=s^-1;x:=s*x;v:=s*v;
    until s=();
    return [res,x,v];
  end;
  W.operations.LeftGcdSimples:=function(W,x,v)local res,i,s,j,l,k;res:=();
    repeat s:=();l:=1;
      for i in [1..n] do
	if i=n or i^v<(i+1)^v or i^x<(i+1)^x then 
          k:=i;
          while l>1 do s:=s*(k,k-l+1);l:=l-2;k:=k-1;od;
          l:=1;
        else l:=l+1;
        fi;
      od;
      res:=res*s;s:=s^-1;x:=s*x;v:=s*v;
    until s=();
    return [res,x,v];
  end;
  return W;
end;

#############################################################################
#
# Example II: HyperOctaedral groups as Coxeter groups
#
CoxeterGroupHyperoctaedralGroup:=function(n)local W,i,f;
  f:=[(n,n+1)];for i in [n,n-1..2] do Add(f,(i,i-1)(2*n+1-i,2*n+2-i));od;
  W:=Group(f,());
  W.reflections:=f;
  W.reflectionsLabels:=[1..n];
  W.operations.IsLeftDescending:=function(W,w,i) return (n+1-i)^w>(n+2-i)^w;end;
  AbsCoxOps.CompleteCoxeterGroupRecord(W);
  return W;
end;
