#############################################################################
##
#A  hecke.g                CHEVIE library           Meinolf Geck, Jean Michel 
##
#A  $Id: hecke.g,v 1.3 1997/03/21 15:24:50 werner Exp $
##
#Y  Copyright (C) 1992 - 1996  Lehrstuhl D fur Mathematik, RWTH Aachen, and
#Y  Universite Paris VII.
##
##  This file contains  GAP functions for working with Hecke algebras and
##  their character tables.
##
##############################################################################
##
## HeckeAlgebraOps: operations for Iwahori-Hecke algebras (for finite Coxeter groups)
##
HeckeAlgebraOps:=OperationsRecord("HeckeAlgebraOps",CoxeterHeckeAlgebraOps);

Inherit(HeckeAlgebraOps,CycHeckeOps); # for CharTable and SchurElements
# some operations of CycHecke ops, like HeckeCharValues are overwritten by
# better ones for the Coxeter case here
# some operations like HeckeCentralMonomials work in the general case
HeckeAlgebraOps.name:="HeckeAlgebraOps";

HeckeAlgebraOps.CoxeterGroup:=HeckeAlgebraOps.Group;
# for compatibility with previous versions of CHEVIE

HeckeAlgebraOps.AdjustCharTable:=function(tbl,param)local r,i;
  r:=[1..Length(tbl.classtext)];
  for i in r do tbl.irreducibles{r}[i]:=tbl.irreducibles{r}[i]*
         Product(-param{tbl.classtext[i]}[2]);
  od;
end;
  
#############################################################################
##
#F  HeckeClassPolynomials( <h> [, <reps>] )  . . . . . . the class polynomials
##
##  'HeckeClassPolynomials' returns the class polynomials of the element
##  <h>  of the Hecke algebra <H> with respect to representatives <reps>
##  of  minimal lengths in the conjugacy  classes of the Coxeter group W
##  of <H>. <h> is an element of <H> given in any of the standard bases.
##  If absent, <reps> is taken to be 'WordsClassRepresentatives(W)'

HeckeAlgebraOps.HeckeClassPolynomials:=function(arg)
  local reps,p,h,T,q,res,min,l,max,inds,i,o,orb,W,H;
  H:=Hecke(arg[1]);T:=Basis(H,"T");h:=T(arg[1]);W:=Group(H);
  if IsBound(arg[2]) then 
    reps:=arg[2];
    p:=MappingPermListList(List(reps,x->PositionClass(W,EltWord(W,x)),
        [1..Length(reps)]));
  else reps:=WordsClassRepresentatives(W);p:=();
  fi;

# if  there exists w conjugate by cyclic  shift to elm and an elementary
# reflection    s   such   that   l(sws)=l(elm)-2   then   orb   returns
# qT_sws+(q-1)T_sw else an error is signaled.

  orb:=function(elm)local orbit,w,y,s,yy,ind,ILD;
    ILD:=W.operations.IsLeftDescending;
    orbit:=[elm];
    for w in orbit do
      for s in [1..W.semisimpleRank] do
        q:=H.parameter[s];
	y:=W.reflections[s]*w;yy:=y*W.reflections[s];
	if ILD(W,w,s) then
	  if ILD(W,y^-1,s) then return T([yy,y],[-q[1]*q[2],q[1]+q[2]]);
	  elif not yy in orbit then Add(orbit,yy);
	  fi;
	elif ILD(W,y^-1,s) and not yy in orbit then Add(orbit,yy);
	fi; 
      od;
    od;
    Error("Geck-Pfeiffer theory");
  end;

  min:=[1..Length(reps)]*0*H.unit;
  while Length(h.elm)>0 do
   l:=List(h.elm,x->CoxeterLength(W,x));
   max:=Maximum(l);
   inds:=Filtered([1..Length(h.elm)],x->max=l[x]);
   max:=T(h.elm{inds},h.coeff{inds});
   h:=h-max;
   for i in [1..Length(max.elm)] do
     o:=PositionClass(W,max.elm[i])^p;
     if Length(reps[o])=CoxeterLength(W,max.elm[i]) then 
       min[o]:=min[o]+max.coeff[i];
     else
       o:=max.coeff[i]*orb(max.elm[i]);
       Append(h.elm,o.elm);Append(h.coeff,o.coeff);
     fi;
   od;
   HeckeEltOps.Normalize(h);
  od;
  return min;
end;

CoxeterGroupOpsHecke:=function(arg)local H;
  H:=ApplyFunc(AbsCoxOps.Hecke,arg);
  H.operations:=HeckeAlgebraOps;
  return H;
end;

HeckeAlgebraOps.LowestPowerGenericDegrees:=H->-List(SchurElements(H),
 function(x) if IsPolynomial(x) then return x.valuation;else return 0;fi;end);

############################################################################
##
#F  PoincarePolynomial( <H> ) . . . Poincare Polynomial of Hecke Algebra <H>
##
## returns  the Poincare Polynomial  of <H> (equal  to the Schur element
## associated  to the 'identity' character).  The point of separate data
## is  that the  Poincare polynomial  is often  more rational than other
## Schur  elements. It coincides with  the generic order for 1-parameter
## algebras.

PoincarePolynomial:=function(H)
  if not IsRec(H) or not IsBound(H.parameter) then
    Error("<H> should be a Hecke algebra");
  fi;
  return Product(ReflectionType(Group(H)),
     t->ReflTypeOps.PoincarePolynomial(t,H.parameter{t.indices}));
end;

############################################################################
##
#F  JInductionTable( <subgroup>, <group> ) . . . J-induction of characters
##  
##  This    function  works   like  'InductionTable'   but   computes  the
##  J-induction, as defined in [Lusztig-book, p.78].
##  
JInductionTable:=function(u,g)local it,i,j,au,ag;
  au:=LowestPowerGenericDegrees(u);
  ag:=LowestPowerGenericDegrees(g);
  if IsHeckeAlgebra(u) then u:=Group(u);fi;
  if IsHeckeAlgebra(g) then g:=Group(g);fi;
  it:=InductionTable(u,g);
  it.what:="JInductionTable";
  it.head:=function(t,option)
    if IsBound(option.TeX) then return SPrint("$J$-Induction from $",
      ReflectionName(t.u,option),"$ to $",ReflectionName(t.g,option),"$");
    else return SPrint("J-Induction from ",
      ReflectionName(t.u,option)," to ",ReflectionName(t.g,option));fi;end;
  for i in [1..Length(au)] do
    for j in [1..Length(ag)] do
      if ag[j]<>au[i] then it.scalar[j][i]:=0;fi;
    od;
  od;
  return it;
end;

############################################################################
##
#F  HeckeScalarProducts( <ti>, <char1>, <char2> )  . . . . . . . . . . . . .
#F   . . . . . . . . . . . . . . . . . .  scalar products between characters
##
##  'HeckeScalarProducts' specializes the parameters to  1  and  returns the
##  matrix of ordinary scalar products between the  specialized  characters.
##  This only works if the parameters are all equal to 1 or are polynomials
##  with value 1 at 1.
##
HeckeScalarProducts:=function(ti,char1,char2)local specialize;
  specialize:=function(p)
    if IsPolynomial(p) then
      return Value(p,1);
    else
      return p;
    fi;
  end;
  if ForAny(ti.parameter,x->specialize(-x[1]/x[2])<>1)
  then Error("parameters must be equal to 1 or have value 1 at 1\n");
  fi;
  char1:=List(char1,x->List(x,specialize));
  char2:=List(char2,x->List(x,specialize));
  return List(char1,i->List(char2,j->ScalarProduct(ti,i,j)));
end;

###########################################################################
##
#F HeckeCharValuesGood( <H>, <w> ) . . . character values of T_w^d for good w
##
## 'HeckeCharValuesGood'  computes the values of the irreducible characters
## of the  Iwahori-Hecke algebra <H> on T_w^d, the $d$-th power of the basis
## element corresponding to the good element <w> of the Coxeter group, where 
## $d$ is the order of <w>. The point is that the character table of the Hecke
## algebra is not used, and that all the eigenvalues of T_w^d are monomials
## in H.parameters, so this can be used to find the absolute value of the
## eigenvalues of T_w, a step towards computing the character table of <H>.
#
HeckeCharValuesGood:=function(H,w)local HI,HJ,J,eig,wd,o,W;
  W:=Group(H);
  wd:=GoodCoxeterWord(W,w);if wd=false then 
      Error("Conjugacy Class representatives of ",W," not good");
  fi;
  wd:=Reversed(wd);
  if Length(wd)=0 or wd[Length(wd)][1]<>[1..W.semisimpleRank] then
         Add(wd,[[1..W.semisimpleRank],0]);fi;
  HJ:=HeckeSubAlgebra(H,[]);eig:=[1];
  for J in wd do
    HI:=HeckeSubAlgebra(H,J[1]);
    eig:=InductionTable(Group(HJ),Group(HI)).scalar*eig;
    o:=List(HeckeCentralMonomials(HI),x->x^(J[2]/2));
    eig:=List([1..Length(eig)],i->eig[i]*o[i]);
    HJ:=HI;
  od;
  return eig;
end;
