#############################################################################
##
#A  heckeelt.g   CHEVIE library    Meinolf Geck, Andrew Mathas and Jean Michel 
##
#Y  Copyright (C) 1992 - 1999  Lehrstuhl D fur Mathematik, RWTH Aachen, and
#Y  Universite Paris VII.
##
##  This file contains  GAP functions for working with elements of an
##  Hecke algebra of a Coxeter group, in various bases.
##

HeckeEltOps:=OperationsRecord("HeckeEltOps");

#############################################################################
##
#F  HeckeEltOps.MakeRec(H,basis,elm,coeff)   . . . .  Internal service routine
#F   makes an Hecke element out of its fields
##
##  The code has been gathered in one place for cleaner reference; also if we
##  want to modify something this is the only place needing a change.
##
HeckeEltOps.MakeRec:=function(H,basis,elm,coeff)
  return rec(coeff:=coeff,elm:=elm,basis:=basis,
               operations:=H.operations.(basis),hecke:=H);
end;

##############################################################
##
#F  HeckeEltOps.Hecke(h)  returns Hecke algebra of h
##
HeckeEltOps.Hecke:=h->h.hecke;

#############################################################################
##
#F  HeckeEltOps.Normalize( <hecke element>)   . . . . Internal service routine
#F    Gathers coefficients for identical elements, and sorts the elements.
##
##   N.B. It happens that Normalize is ``generic'' but we don't care since
##    in every call we call explicitely HeckeEltOps.Normalize for efficiency.
##    We made it part of HeckeEltOps just to hide the name...
##
HeckeEltOps.Normalize:=function(t)local i,elm,coeff;
  SortParallel(t.elm,t.coeff);
  elm:=[];coeff:=[];
  for i in [1..Length(t.elm)] do
    if i=1 or t.elm[i]<>t.elm[i-1] then
	 Add(elm,t.elm[i]);Add(coeff,t.coeff[i]);
    else coeff[Length(coeff)]:=coeff[Length(coeff)]+t.coeff[i];
    fi;
  od;
  i:=List(coeff,x->x<>x*0);
  t.coeff:=ListBlist(coeff,i); t.elm:=ListBlist(elm,i);
  return t;
end;

#############################################################################
##
#F  HeckeEltOps.\+     . . . . . . . . . . . . . . Addition of Hecke elements
#F   If they are not of the same basis, both are converted to T.
##
HeckeEltOps.\+:=function(x,y)
  if x=0 then return y;
  elif y=0 then return x;fi; # to get Sum to work...
  if not IsIdentical(Hecke(x),Hecke(y)) then 
    Error("not elements of the same algebra");
  fi;
  if x.basis <> y.basis then 
    return Basis(Hecke(x),"T")(x)+Basis(Hecke(x),"T")(y);
  else
    x:=ShallowCopy(x);x.elm:=Concatenation(x.elm,y.elm);
    x.coeff:=Concatenation(x.coeff,y.coeff);
    return HeckeEltOps.Normalize(x);
  fi;
end;

#############################################################################
##
#F  HeckeEltOps.\+     . . . . . . . . . . . . . Subtraction of Hecke elements
##
HeckeEltOps.\-:=function(h1,h2) return h1+(-1)*h2; end;

#############################################################################
##
#F  HeckeEltOps.\*(x,y) . . . . . . . . . . Multiplication of Hecke elements
#F   If x is a scalar, just multiply the coefficients of y by x.
#F   If they are not of the same basis return T(x)*T(y).
#F   If they are of T basis, use quadratic relations.
#F   If they are both of Basis X, return X(T(x)*T(y)).
##
HeckeEltOps.\*:=function(x,y)local H,res,W,temp,i,xi,s,temp1,j,e,es,q;
  if IsList(x) then return List(x,z->z*y);
  elif IsList(y) then return List(y,z->x*z);
  fi;
  if IsRec(y) and IsBound(y.hecke) and IsBound(y.elm) then 
    H:=Hecke(y);
    if not IsRec(x) or not IsBound(x.hecke) or not IsBound(x.elm) then 
  # assume x is a scalar by which to multiply y
      y:=ShallowCopy(y);y.coeff:=y.coeff*(x*Hecke(y).unit);
      return HeckeEltOps.Normalize(y);
    fi;
  else # assume y is a scalar by which to multiply x
    x:=ShallowCopy(x);x.coeff:=x.coeff*(y*Hecke(x).unit);
    return HeckeEltOps.Normalize(x);
  fi;
  if not IsIdentical(H,Hecke(x)) then 
    Error("not elements of the same algebra");
  fi;
  if x.basis<>y.basis then 
     if x.operations.\* <> y.operations.\* then
          return x.operations.\*(x,y);
     else return Basis(H,"T")(x)*Basis(H,"T")(y);
     fi;
  elif x.basis="T" then 
    W:=Group(H);
    res:=HeckeEltOps.MakeRec(H,x.basis,[],[]);
    for i in [1..Length(x.elm)] do
      temp:=x.coeff[i]*y;
      xi:=x.elm[i]^-1;
      while xi<>W.identity do
	s:=FirstLeftDescending(W,xi);es:=W.reflections[s];xi:=es*xi;
	temp1:=HeckeEltOps.MakeRec(H,x.basis,[],[]);
	for j in [1..Length(temp.elm)] do
	  e:=temp.elm[j];
	  if IsLeftDescending(W,e,s) then
            q:=H.parameter[s];
	    Add(temp1.elm,e);Add(temp1.coeff,(q[1]+q[2])*temp.coeff[j]);
	    Add(temp1.elm,es*e);Add(temp1.coeff,-q[1]*q[2]*temp.coeff[j]);
	  else
	    Add(temp1.elm,es*e);Add(temp1.coeff,temp.coeff[j]);
	  fi;
	od;
	temp:=HeckeEltOps.Normalize(temp1);
      od;
      res:=res+temp;
    od;
    return res;
  else return Basis(H,x.basis)(Basis(H,"T")(x)*Basis(H,"T")(y));
  fi;
end;

#############################################################################
##
#F  HeckeEltOps.\^(x,y) . . . . . . . . . . Exponentiation for Hecke elements
#F   If y is an integer, return x^y using multiplication. y is allowed to be
#F    negative only for bases which have an .inverse method.
#F   Otherwise return the conjugate y^-1*x*y
##
HeckeEltOps.\^:=function(h,n) local i,p,H,W,T,q,e;
  H:=Hecke(h);W:=Group(H);
  if not IsInt(n) then return n^-1*h*n; # assume n is a Hecke element
  elif n<0 then
    if n=-1 then
      if IsBound(h.operations.inverse) then return h.operations.inverse(h);
      else Error(h," has no method for inverse");
      fi;
    else return (h^-1)^(-n);
    fi;
  else
   if n=0 then return Basis(H,h.basis)(Basis(H,"T")());fi; 
   # note that the above possibly expensive expression is called only for h^0
   p:=false;
   while n>0 do
    if n mod 2 <> 0 then 
      if p=false then p:=h;
      else p:=p*h;
      fi;
    fi;
    h:=h*h;
    n:=QuoInt(n,2);
   od;
   return p;
   fi;
end;

#############################################################################
##
#F  HeckeEltOps.\/(x,y) . . . . . . . . . . Division for Hecke elements
##
HeckeEltOps.\/:=function(a,b) return a*b^-1;end;

#############################################################################
##
#F AlphaInvolution(h)    
## The involution on Hecke Elements defined by T_w->T_{w^{-1}}
## (and same in other bases)
HeckeEltOps.AlphaInvolution:=h->HeckeEltOps.MakeRec(Hecke(h),h.basis,
                                                 List(h.elm,x->x^-1),h.coeff);

#############################################################################
##
#F  HeckeEltOps.Print(h) . . . . . . . . . . Print method for Hecke elements
##  .Print calls .String as is general policy in CHEVIE
##
HeckeEltOps.Print:=function(h)Print(String(h));end;

CHEVIE.PrintHecke:=rec();

#############################################################################
##
#F  HeckeEltOps.String(h) . . . . . . . . . . String method for Hecke elements
##
HeckeEltOps.String:=function(h)local i,hcoeff,helm,s,res;
  if h.elm=[] then return "0";fi;
  res:="";
  hcoeff:=ShallowCopy(h.coeff);
  helm:=List(h.elm,x->CoxeterWord(Group(Hecke(h)),x));
  SortParallel(helm,hcoeff,function(x,y) return Length(x)<Length(y) or
			      (Length(x)=Length(y) and x<y);end);
  for i in [1..Length(helm)] do
    s:=StringCoefficient(hcoeff[i],SPrint(h.basis,"(",Join(helm[i]),")"),
      CHEVIE.PrintHecke);
    if Position(s,'-')<>1 and i<>1 then Append(res,"+");fi;
    Append(res,s);
  od;
  return res;
end;

#############################################################################
##
#F  HeckeEltOps.Coefficient(h,e) . . . . . . . . . . . Coefficient of e in h
#F     e is an element of Group(h) or a word in the generators of Group(h)
##  Uses that Coefficient is generic.
##
HeckeEltOps.Coefficient:= function(T,elm)local p,W;
  W:=Group(Hecke(T));
  if IsWordFor(W,elm) then elm:=EltWord(W,elm);fi;
  p:=Position(T.elm,elm);
  if p=false then return 0*Hecke(T).unit;else return T.coeff[p];fi;
end;

#############################################################################
##
#F  HeckeCharValues( <T> [,<irreds>] )  . . character values of <irreds> on <T>
##
##  T is an element of an Iwahori-Hecke algebra H (expressed in any of the 5
##  bases) and <irreds> is a set of irreducible characters of H.
##  HeckeCharValues returns the values of irreds on T, using
##  HeckeClassPolynomials.
##   If absent, <irreds> is taken to be the set of irreducible characters
##  of <H>.

HeckeEltOps.HeckeCharValues:=function(arg)local irrs,H;
  if IsBound(arg[1].coset) then H:=arg[1].coset;
  else H:=Hecke(arg[1]);fi;
  if IsBound(arg[2]) then irrs:=arg[2]*H.unit^0;
  else irrs:=CharTable(H).irreducibles;fi;
  return irrs*HeckeClassPolynomials(arg[1]);
end;

#############################################################################
##
#F  HeckeEltOps.Frobenius(W,h) . . . . . . . . . . . Frobenius(W)(h)
#F     W is a CoxeterCoset, and h an element of an HeckeCoset of W or of
#F     an hecke algebra of the group of which W is a coset.
##
HeckeEltOps.Frobenius:=function(W,x)
  x:=ShallowCopy(x);x.elm:=List(x.elm,y->Frobenius(W)(y));
  return x;
end;

#############################################################################
##
#F CreateHeckeBasis(basis,basisops,algebraops) . . . . . . . Create a new
#F   basis of Hecke algebras which have algebraops as operations.
##
## basis, a string, is  the name of the new basis  and basisops should be
## a  record  containing the  operations  in  HeckeEltOps which  will  be
## overridden for that particular basis. basisops should contain at least
## methods .T  and .(basis)  to convert  to and from  the T-basis.  If it
## contains a  method init(H) this  is called when  using Basis(H,basis).
## Most  other methods  of HeckeEltOps  can  be overriden  if needed,  in
## particular MakeBasisElt called  by 'Basis' is overriden  by the Murphy
## basis code (see the package 'Specht').
##
CreateHeckeBasis:=function(arg)local basis,basisops,algebraops;
# accepts no 3rd argument for compatibility with previous versions of CHEVIE
  basis:=arg[1];basisops:=arg[2];
  if Length(arg)=2 then algebraops:=HeckeAlgebraOps;
  else algebraops:=arg[3];
  fi;
  if not IsBound(algebraops.Basis) then # first basis ever created for algebra
    algebraops.Basis:=function(H,basis)local basisops;
    # we use that Basis is generic
     if not IsBound(H.operations.(basis)) or 
        not IsBound(H.operations.(basis).T)  # test it is not a module basis
     then 
        Error("basis ",basis," unknown");
     fi;
     basisops:=H.operations.(basis);
     if IsBound(basisops.init) then basisops.init(H);fi;
     return function(arg) return basisops.MakeBasisElt(H,basis,arg);end;
    end;
  fi;

  if not IsRec(basisops) or not IsBound(basisops.T) 
                         or not IsBound(basisops.(basis)) then 
    Error("The operations record must contain methods .T and .",basis);
  fi;
  algebraops.(basis):=OperationsRecord(Concatenation("Hecke",basis,"Ops"),
                                       HeckeEltOps);
  Inherit(algebraops.(basis),basisops);
end;

#############################################################################
##
#F HeckeEltOps.MakeBasisElt(H,basis,arg)   . . after setting X:=Basis(H,"X")
#F   the hecke-element of the 'X' basis making function X will contain
#F    ``function(arg) return H.operations.X.MakeBasisElt(H,basis,arg);end;''
#F   If the default version given
#F   below of MakeBasisElt is used it will accept the forms (here W=Group(H)):
## 
##1.  X(<group element g>) Interpreted as X([g],[1])
##2.  X([s_1,...,s_n])     where s_i is the name of generator S_i of W:
##        Interpreted as X(Product_i S_i) if s_1..s_n reduced. Otherwise
##        basis must be T and returns Product_i T(S_i) otherwise. As a 
##        special convention, a negative s_i represents the inverse of the
##        corresponding generator.
##3.  X(s_1,...,s_n)       Interpreted as X([s_1,...,s_n])
##4.  X([elts],[coeffs])   Makes the element of basis "X" with elements [elts]
##                         and coefficients [coeffs]
##5.  X(h)                 In this form h is a hecke element and the function
##                  tries to convert h to basis "X". It first looks if h has a
##                  method .("X"), and if not converts h to T and then to X.
HeckeEltOps.MakeBasisElt:=function(H,basis,arg)
  local w,res,s,h,basisops,W,isrefl;
  basisops:=H.operations.(basis);
  if Length(arg)=1 and IsRec(arg[1]) and IsBound(arg[1].hecke) and 
     IsBound(arg[1].elm) then  # case 5. assume arg is some hecke element h
    h:=arg[1];
    if h.basis=basis then return h;
    elif IsBound(h.operations.(basis)) then return h.operations.(basis)(h);
    else return basisops.(basis)(h.operations.T(h));
    fi;
  fi;
  W:=Group(H);
  if arg=[] then # case 2.
    return HeckeEltOps.MakeRec(H,basis,[W.identity],[H.unit]);
  fi;
  if Length(arg)=2 and IsList(arg[1]) and IsList(arg[2]) then # case 4.
    return HeckeEltOps.MakeRec(H,basis,arg[1],arg[2]);
  fi;
  isrefl:=x->IsInt(x) and (x in W.reflectionsLabels or -x in W.reflectionsLabels);
  if Length(arg)=1 and not isrefl(arg[1]) then
    h:=arg[1];
    if not(IsList(h) and ForAll(h,isrefl)) then 
      # assume h is a group element --- case 1.
      return HeckeEltOps.MakeRec(H,basis,[h],[H.unit]);
    fi;
  else h:=arg; # case 3
  fi;
  # now case 3. Check [s_1,...,s_n] is reduced -- else assume basis is "T"
  if ForAll(h,x->x>0) then
    w:=EltWord(W,h);
    if CoxeterLength(W,w)=Length(h) then 
      return HeckeEltOps.MakeRec(H,basis,[w],[H.unit]);
    fi;
  fi;
  if basis<>"T" then
   Error("Construction with non-reduced words implemented only for basis T");
  fi;
  res:=HeckeEltOps.MakeRec(H,basis,[W.identity],[H.unit]);
  for s in h do 
    if s>0 then res:=res*HeckeEltOps.MakeRec(H,basis,
        [W.reflections[W.operations.ReflectionFromName(W,s)]],[H.unit]);
    else res:=res/HeckeEltOps.MakeRec(H,basis,
        [W.reflections[W.operations.ReflectionFromName(W,-s)]],[H.unit]);
    fi;
  od;
  return res;
end;

HeckeEltOps.Specialization:=function(t,H2,f)local res;
  res:=Basis(H2,t.basis)(t.elm,List(t.coeff,f));
  res.operations.Normalize(res);
  return res;
end;

# The function below is made member of CoxeterHeckeAlgebraOps just to hide it
# it computes T(w^-1)^-1
CoxeterHeckeAlgebraOps.getTinv:=function(H,w)local p,i,q,W;
  W:=Group(H);
  if not IsBound(H.elts) then H.elts:=[W.identity]; fi;
  if not IsBound(H.Tinv) then H.Tinv:=[Basis(H,"T")()];fi;
  p:=Position(H.elts,w);
  if p=false then Add(H.elts,w);p:=Length(H.elts);fi;
  if not IsBound(H.Tinv[p]) then
    i:=FirstLeftDescending(W,w);
    q:=H.parameter[i];
    H.Tinv[p]:=Basis(H,"T")([W.identity,W.reflections[i]],
	               [(q[1]+q[2])/(q[1]*q[2]),-1/(q[1]*q[2])])*
                          H.operations.getTinv(H,W.reflections[i]*w);
  fi;
  return H.Tinv[p];
end;

#############################################################################
##
## An exemple of CreateHeckeBasis: create the 'T' basis for
## CoxeterHeckeAlgebraOps (i.e. all Hecke algebras for Coxeter groups...)
##
CreateHeckeBasis("T",rec(T:=x->x,     # method to convert to T
 inverse:=function(h)
   if Length(h.elm)<>1 then Error("inverse implemented only for single T_w");fi;
   return h.coeff[1]^-1*Hecke(h).operations.getTinv(Hecke(h),h.elm[1]^-1);
 end),CoxeterHeckeAlgebraOps);

ReadChv("prg/kl"); # this will actually add new methods to basis T ...
