#############################################################################
##
#A  kl.g           CHEVIE library    Meinolf Geck, Jean Michel, Andrew Mathas
##
#A  $Id: kl.g,v 1.1 1997/01/21 13:46:32 gap Exp $
##
#Y  Copyright (C) 1992 - 2000  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains  GAP functions to  compute Kazhdan - Lusztig polynomials
##  and left cells, and  for  working with KL bases of Hecke algebras. 
##
#############################################################################

#next routine computes q_x^(1/2)
QXHalf:=function(H,x)local W;
  W:=Group(H);
  if H.equal then return H.rootParameter[1]^CoxeterLength(W,x);
  else return Product(H.rootParameter{List(CoxeterWord(W,x),
                          y->Position(W.reflectionsLabels,y))});
  fi;
end;

RootParameter:=function(H,i,msg)local W,j;
  if not IsBound(H.rootParameter[i]) then
    W:=Group(H);
    j:=W.rootRestriction[W.orbitRepresentative[i]];
    if i=j then
       H.rootParameter[i]:=GetRoot(-H.parameter[i][1]/H.parameter[i][2],2,msg);
    else H.rootParameter[i]:=RootParameter(H,j,msg);
    fi;
  fi;
  return H.rootParameter[i];
end;

# The next routine  checks that the Hecke  algebra H is one of  a kind for
# which we know how to compute the KL basis:
#  According to  Lusztig, they are  defined when all  rootParameters are
# (or can  be) bound and are  elements of some totally  ordered group G.
# Then coefficients of Hecke elements are  elements of the ring Z[G]. We
# need to be able to perform the following operations in this ring:
#
#   H.Bar: the involution \sum_g a_g g -> \sum_g a_g g^-1
#   H.PositivePart:      \sum_g a_g g -> \sum_{g>=1} a_g g
#   H.NegativePart:      \sum_g a_g g -> \sum_{g<=1} a_g g
#
# We  know  how  to  construct  automatically  such  functions  when  all
# rootParameters  are powers  (GAP monomials)  of the  same variable  q.
# Otherwise they need  to be bound already (Positive for equal parameters
# and Negative in the general case: there is special code faster than the
# general case when all parameters are equal).

PrepareForPolynomials:=function(H)local q;
# prepare to work with 1-variable polynomials

  for q in H.rootParameter do # need parameters positive monomials
    if not IsPolynomial(q) or Length(q.coefficients)<>1 or
      q.coefficients[1]<>1 or Degree(q)<=0 then 
      return false;
    fi;
  od;

  H.Bar:=p->Polynomial(p.baseRing,Reversed(p.coefficients),
              1-p.valuation-Length(p.coefficients));

  # returns terms of positive degree of polynomial p
  H.PositivePart:=function(p)local v;
    v:=Maximum(0,-p.valuation);if v=0 then return p;fi;
    return Polynomial(p.baseRing,p.coefficients{[1+v..Length(p.coefficients)]},
      p.valuation+v);
  end;

  # returns terms of negative degree of polynomial p
  H.NegativePart:=p->Polynomial(p.baseRing,p.coefficients{[1..Minimum(
           Length(p.coefficients),1-p.valuation)]},p.valuation);

  return true;
end;

# to prepare Hecke algebras to use Mvps for KL theory
PrepareForMvp:=function(H)local q;
# prepare to work with Mvp -- lexicographic order

  for q in H.rootParameter do # need parameters positive monomials
    if not IsMvp(q) or Length(q.coeff)<>1 or q.coeff[1]<>1 or 
      Length(q.elm[1].elm)=0 or q.elm[1].coeff[1]<0 then 
      return false;
    fi;
  od;

  H.Bar:=p->MvpOps.Normalize(List(p.elm,x->rec(elm:=x.elm,coeff:=-x.coeff)),
    p.coeff);
  
  H.PositivePart:=function(p)local v;
    v:=Filtered([1..Length(p.elm)],x->p.elm[x].coeff=[] or p.elm[x].coeff[1]>0);
    return MvpOps.Normalize(p.elm{v},p.coeff{v});
  end;

  H.NegativePart:=function(p)local v;
    v:=Filtered([1..Length(p.elm)],x->p.elm[x].coeff=[] or p.elm[x].coeff[1]<0);
    return MvpOps.Normalize(p.elm{v},p.coeff{v});
  end;

  return true;
end;

CoxeterHeckeAlgebraOps.InitKL:=function(H,msg)
  if IsBound(H.CpBasis) then return;fi; # assume already called
  if not IsCoxeterGroup(Group(H))then Error(msg,": only for Coxeter groups");fi;
  H.equal:=Length(Set(List([1..Length(H.parameter)],
                      i->RootParameter(H,i,msg))))<=1;
  # this takes care rootParameters are computed

  # As the C'(w) basis elements are computed they are stored in the list
  # CpBasis at the position in elts of w (which is added to elts if not there)
  if not IsBound(H.elts) then H.elts:=[Group(H).identity]; fi;
  H.CpBasis:=[Basis(H,"T")()];
  if not IsBound(H.Bar) and not PrepareForPolynomials(H) 
    and not PrepareForMvp(H) then
      Error(msg,": parameters should be positive monomials\n");
  fi;
end;

CoxeterHeckeAlgebraOps.InitD:=function(H,msg)
  CoxeterHeckeAlgebraOps.InitKL(H,msg);
  if not H.equal or not IsFinite(Group(H)) then 
    Error(msg,": only for equal parameters algebras of finite Coxeter groups");
  fi;
end;

##  Alt is The involution of H defined by v -> v^-1 and
##  T_w -> (-1)^{l(w)} QXHalf(w)^-2 T_w.
##  It swaps C_w and C'_w, and D_w and D'_w.
##  Essentially it corresponds to tensoring with the sign representation.

# The function  below is made  member of CoxeterHeckeAlgebraOps  just to
# hide it. It applies v->v^-1 to the coefficients of h. This is not Alt!
# One needs also to do some normalisation depending on the basis.
CoxeterHeckeAlgebraOps.Alt:=function(h,newbasis)local H;
  h:=ShallowCopy(h); H:=Hecke(h);
  h.basis:=newbasis;h.operations:=H.operations.(newbasis);
  h.coeff:=List(h.coeff,H.Bar);
  return h;
end;

##  Beta is the involution on H defined by v -> v^-1 and
##  T_w -> v^{-l(w_0)}T_{w_0w}.
## It swaps C'_w and D_{w_0w}, and C_w and D'_{w_0w}.

# The function  below is made  member of CoxeterHeckeAlgebraOps  just to
# hide it. It is not the full Beta!
CoxeterHeckeAlgebraOps.Beta := function(h,newbasis)local H;
  h:=ShallowCopy(h);H:=Hecke(h);
  h.basis:=newbasis;h.operations:=H.operations.(newbasis);
  h.elm:=LongestCoxeterElement(Group(H))*h.elm;
  h.coeff:=List(h.coeff,H.Bar);
  return h;
end;

# now add operations to basis "T"

# The complete Alt for T basis
CoxeterHeckeAlgebraOps.T.AltInvolution:= function(h)local W,q,i,H;
  H:=Hecke(h);
  W:=Group(H);
  H.operations.InitKL(H,"alt involution");
  h:=H.operations.Alt(h,"T");
  for i in [1..Length(h.coeff)] do
    h.coeff[i]:=QXHalf(H,h.elm[i])^-2*(-1)^CoxeterLength(W,h.elm[i])*h.coeff[i];
  od;
  return h;
end;

# The complete Beta for T basis
CoxeterHeckeAlgebraOps.T.BetaInvolution:=function(h)local H,W;
  H:=Hecke(h);
  H.operations.InitKL(H,"beta involution");
  h:=H.operations.Beta(h,"T");
  W:=Group(H);
  if not IsBound(W.isFinite) or not W.isFinite then 
    Error("BetaInvolution only for finite Coxeter groups");
  fi;
  h.coeff:=QXHalf(H,LongestCoxeterElement(W))^-1*h.coeff;
  return h;
end;

# The function below is made member of CoxeterHeckeAlgebraOps just to hide it
CoxeterHeckeAlgebraOps.getCp := function(H,w)local W,iw,i,qx,smaller,x,c;
  W:=Group(H);iw := Position(H.elts, w);
  if iw=false then Add(H.elts,w);iw:=Length(H.elts);fi;
  if IsBound(H.CpBasis[iw]) then return ShallowCopy(H.CpBasis[iw]);fi;
# InfoChevie("# computed ",Number([1..Length(H.CpBasis)],
# i->IsBound(H.CpBasis[i]))," C':");Ptime("");
  if H.equal then
    i:=FirstLeftDescending(W,w);
    qx:=QXHalf(H,W.reflections[i]);
    H.CpBasis[iw]:=Basis(H,"T")([W.identity,W.reflections[i]],[qx^-1,qx^-1])*
                      H.operations.getCp(H,W.reflections[i]*w);
	
    # new implementation JM and FD march 2000 replacing A.Mathas dec. 1994
    # We use the following formulae:
    # $$ C'_w=\sum_{y\le w}P_{y,w}(q^2)q^{-l(w)}T_y$$
    # and if sw<w then
    # $$ C'_s C'_{sw}=C'w+\sum_{y<sw}\mu(y,sw)C'y=\sum_{v\le w}\mu_v T_v$$
    # where
    # $$\mu_v=P_{v,w}(q^2)q^-l(w)+\sum_{v\le y\le sw}
    # \mu(y,sw)P_{v,y}(q^2)q^{-l(y)}$$
    #
    # It follows  that if  \mu_v is of  degree \ge -l(v)  then it  is of
    # degree  exactly  -l(v)  with corresponding  coefficient  equal  to
    # \mu(v,sw) (this happens exactly for y=v in the sum which occurs in
    # the formula for \mu_v).

    H.CpBasis[iw]:=H.CpBasis[iw]-Sum([1..Length(H.CpBasis[iw].elm)],
      function(i)local c,e;
	e:=H.CpBasis[iw].elm[i];if e=w then return 0;fi;
	c:=H.PositivePart(H.CpBasis[iw].coeff[i]*QXHalf(H,e));
	if c<>0 then return c*H.operations.getCp(H,e); else return 0; fi;
      end);
  else
    # we follow formula 2.2 in Lusztig's 'Left cells in Weyl groups'

    smaller:=Concatenation(Reversed(BruhatSmaller(W,w)));
    H.CpBasis[iw]:=QXHalf(H,w)^-1*Basis(H,"T")(w);
    H.CpBasis[iw].elm:=smaller;
    for i in [2..Length(smaller)] do
      x:=smaller[i];
      c:=CriticalPair(W,x,w);
      if x<>c[1] then 
	H.CpBasis[iw].coeff[i]:=H.CpBasis[iw].coeff[Position(smaller,c[1])];
      else
	qx:=QXHalf(H,x);
	H.CpBasis[iw].coeff[i]:=-H.NegativePart(qx^-1*Sum([1..i-1],
	 iy->H.Bar(Coefficient(H.operations.getTinv(H,smaller[iy]),x))
	   *H.CpBasis[iw].coeff[iy]))/qx;
      fi;
    od;
    H.CpBasis[iw]:=HeckeEltOps.Normalize(H.CpBasis[iw]);
  fi;
  return ShallowCopy(H.CpBasis[iw]);
end; 

# The function below is made member of CoxeterHeckeAlgebraOps just to hide it
CoxeterHeckeAlgebraOps.ToKL:=function(h,target,index)local x,res,lens,H,coeff;
  H:=Hecke(h);res:=Basis(H,target)([],[]); 

# To convert from "T", we use the fact that the transition matrix M from
# any  KL  bases to  the  standard  basis  is triangular  with  diagonal
# coefficient on  T_w equal  to QXHalf(w)^-1.  The transition  matrix is
# lower triangular for the C and  C' bases, and upper triangular for the
# D and D' bases which is what index() is for.

  h:=ShallowCopy(h); 
  while h.elm <> [] do
    lens:=List([1..Length(h.elm)],i->CoxeterLength(Group(H),h.elm[i]));
    x:=index(lens); lens:=Filtered([1..Length(h.elm)],i->lens[i]=x);
    coeff:=List(lens,i->h.coeff[i]*QXHalf(H,h.elm[i]));
    x:=Basis(H,target)(h.elm{lens},coeff);
    res:=res+x; h:=h-Basis(H,"T")(x);
  od;
  return res;
end;

CreateHeckeBasis("C",rec(
  init:=function(H)H.operations.InitKL(H,"C basis");end,
  BetaInvolution:=h->Hecke(h).operations.Beta(h,"D'"),
  AltInvolution:=h->Hecke(h).operations.Alt(h,"C'"),
  T:=function(h)local H;H:=Hecke(h);
  if Length(h.coeff)=0 then return Basis(H,"T")([],[]);
  else return h.coeff*List(h.elm,x->(-1)^CoxeterLength(Group(H),x)*
        H.operations.T.AltInvolution(H.operations.getCp(H,x)));fi;end,
  C:=h->Hecke(h).operations.ToKL(h,"C",Maximum)
  ),CoxeterHeckeAlgebraOps);

CreateHeckeBasis("C'",rec(
  init:=function(H)H.operations.InitKL(H,"C' basis");end,
  BetaInvolution:=h->Hecke(h).operations.Beta(h,"D"),
  AltInvolution:=h->Hecke(h).operations.Alt(h,"C"),
  T:=function(h)local H;H:=Hecke(h);
  if Length(h.coeff)=0 then return Basis(H,"T")([],[]);
  else return h.coeff*List(h.elm,x->H.operations.getCp(H,x));fi;end,
  ("C'"):=h->Hecke(h).operations.ToKL(h,"C'",Maximum)
  ),CoxeterHeckeAlgebraOps);

CreateHeckeBasis("D",rec(
  init:=function(H)H.operations.InitD(H,"D basis");end,
  BetaInvolution:=h->Hecke(h).operations.Beta(h,"C'"),
  AltInvolution:=h->Hecke(h).operations.Alt(h,"D'"),
  T:=function(h)local H;H:=Hecke(h);
  if Length(h.coeff)=0 then return Basis(H,"T")([],[]);
  else return h.coeff*List(h.elm,x->H.operations.T.BetaInvolution(
     H.operations.getCp(H,LongestCoxeterElement(Group(H))*x)));fi;end,
  D:=h->Hecke(h).operations.ToKL(h,"D",Minimum)
  ),CoxeterHeckeAlgebraOps);

CreateHeckeBasis("D'",rec(
  init:=function(H)H.operations.InitD(H,"D' basis");end,
  BetaInvolution:=h->Hecke(h).operations.Beta(h,"C"),
  AltInvolution:=h->Hecke(h).operations.Alt(h,"D"),
  T:=function(h)local H;H:=Hecke(h);
  if Length(h.coeff)=0 then return Basis(H,"T")([],[]);
  else return h.coeff*List(h.elm,x->(-1)^CoxeterLength(Group(H),x)*
      H.operations.T.AltInvolution(
        H.operations.T.BetaInvolution(H.operations.getCp(
	H,LongestCoxeterElement(Group(H))*x))));fi;end,
  ("D'"):=h->Hecke(h).operations.ToKL(h,"D'",Minimum)
  ),CoxeterHeckeAlgebraOps);

#############################################################################
##
#F  CriticalPair( <W>, <y>, <w> ) . . . . . . . . . . . the critical pair 
#F   . . . . . . . . . . . . . . associated with a given pair of elements 
##
## A pair of elements  $y$, $w$ in $W$ is called  critical if neither the
## left descent set  of $y$ is contained  in the left descent  set of $w$
## nor the  right descent  set of $y$  is contained in  that of  $w$. The
## function 'CriticalPair' returns, for a given pair <y>, <w> of elements
## in  a  Coxeter  group  <W>,  the pair  itself  if  this  is  critical.
## Otherwise,  <y> can  be multiplied  from the  left or  the right  by a
## generator $s$, and we can proceed with the pair consisting of this new
## element and <w>.  Repeating in this way, we finally  obtain a critical
## associated with the given pair.  
## The Length(z)-Length(y) is also returned where <z> is the new element
## replacing <y>.
## The  significance of  this  construction is  that the  Kahzdan-Lusztig
## polynomials corresponding to  the original pair and  to the associated
## critical pair are equal.
##

CriticalPair:=function(W,y,w)local cr,Rw,Lw,IL, sR,rI,pN;
  Lw:=List(LeftDescentSet(W,w),x->Position(W.reflectionsLabels,x));
  Rw:=List(RightDescentSet(W,w),x->Position(W.reflectionsLabels,x));

  IL:=W.operations.IsLeftDescending; # avoid dispatching overhead.
  
  cr:=function(y,ly) local s;
    for s in Lw do 
      if not IL(W,y,s) then return cr(W.reflections[s]*y,ly+1);fi;
    od;
    for s in Rw do 
      if not IL(W,y^-1,s) then return cr(y*W.reflections[s],ly+1);fi;
    od;
    return [y,w,ly];
  end;

  return cr(y,0);
end;


#############################################################################
##
#F  KazhdanLusztigPolynomial( <W>, <y>, <w> [, <ly>, <lw> ] )  . . . . . . .
##  . . . . . . . . . . . . . . . . . . . . . . . Kazhdan-Lusztig polynomial
##
##  'KazhdanLusztigPolynomial'   returns   the  Kazhdan - Lusztig  polynomial 
##  in an indeterminate over the rationals corresponding to the elments <y> 
##  and <w>  (given as permutations) in the Coxeter group <W>. The optional
##  variables <ly> and <lw>  contain the length of <y> and <w>, respectively.
##
KazhdanLusztigPolynomial:=function(arg)
  local W,y,w,ly,lw,pw,py,cr,s,v,lv,z,lz,m,tmp,2N,pol,wcr,otc1,ild;
  W:=arg[1]; y:=arg[2]; w:=arg[3];
  if Length(arg)=3 then
    ly:=CoxeterLength(W,y);
    lw:=CoxeterLength(W,w);
  else
    ly:=arg[4]; lw:=arg[5];
  fi;
  if not Bruhat(W,y,w,ly,lw) then
    return 0*X(Rationals);
  fi;
  cr:=CriticalPair(W,y,w);
  ly:=ly+cr[3];y:=cr[1];
  if lw-ly<=2 then
    return X(Rationals)^0;
  fi; 
  
  if not IsBound(W.criticalPairs) then
    W.criticalPairs:=[];
    W.klpol:=[];  
  fi;
  pw:=Position(CoxeterElements(W,lw),w);
  if not IsBound(W.criticalPairs[lw+1]) then
    W.criticalPairs[lw+1]:=[];
  fi;  
  wcr:=W.criticalPairs[lw+1];
  
  # encode y in some way to save memory (an element is uniquely determined
  # by its images of the simple roots)
# 2N:=2*W.parentN;
# tmp:=OnTuples(W.reflectionsLabels{W.generatingReflections},y);
# otc1:=tmp[1];
# for z in [2..Length(tmp)] do
#   otc1:=otc1*2N+tmp[z];
# od;

  # encode y to save memory # changed JM 2-7-2000 to work for any Coxeter group
  # W.criticalPairs also changed accordingly
  otc1:=Position(CoxeterElements(W,ly),y);

  if not IsBound(wcr[pw]) then  wcr[pw]:=[];fi;
  if not IsBound(wcr[pw][ly]) then  wcr[pw][ly]:=[[],[]];fi;
  py:=Position(wcr[pw][ly][1],otc1);
  if py<>false then
    return Rationals.operations.Polynomial(Rationals,W.klpol[wcr[pw][ly][2][py]],0);
  fi;
  
  s:=FirstLeftDescending(W,cr[2]);
  v:=W.reflections[s]*w;
  lv:=lw-1;
  pol:=KazhdanLusztigPolynomial(W,W.reflections[s]*y,v,ly-1,lv)
                     +X(Rationals)*KazhdanLusztigPolynomial(W,y,v,ly,lv);
  lz:=lv-1;
  ild:=W.operations.IsLeftDescending;
  while lw-lz<=2*Degree(pol) do
    for z in CoxeterElements(W,lz) do 
      if lw-lz<=2*Degree(pol) and pol.coefficients[(lw-lz)/2+1]>0 
          and ild(W,z,s) and Bruhat(W,y,z,ly,lz) then
        m:=KazhdanLusztigMue(W,z,v,lz,lv);
        if m<>0 then
          pol:=pol-m*X(Rationals)^((lv-lz+1)/2)*
                                   KazhdanLusztigPolynomial(W,y,z,ly,lz);
        fi;
      fi;
    od;
    lz:=lz-2;
  od;
  
  Add(wcr[pw][ly][1],otc1);
  py:=Position(W.klpol,pol.coefficients);
  if py=false then
    Add(W.klpol,pol.coefficients);
    py:=Length(W.klpol);
  fi;
  Add(wcr[pw][ly][2],py);
  return pol;
end;

#############################################################################
##
#F  KazhdanLusztigMue( <W>, <y>, <w> [, <ly>, <lw> ] )  . . . . . . . . . . .
#F  . . . . . . . . . the highest coefficient of a Kazhdan-Lusztig polynomial
##
##  'KazhdanLusztigMue'  returns the coefficient of highest possible degree
##  of the Kazhdan-Lusztig polynomial of the Coxeter group <W> corresponding 
##  to the elements <y> and <w>. The optional variables <ly> and <lw> contain 
##  the  length of <y> and <w>, respectively. That highest possible degree is 
##  $(lw- ly -1)/2$. If <y> and <w> are not related by the Bruhat order, the 
##  value 0 is returned.
##
KazhdanLusztigMue:=function(arg) local W,y,w,ly,lw,s,Lw,Rw,pol,ild;
  W:=arg[1]; y:=arg[2]; w:=arg[3];
  if Length(arg)>3 then ly:=arg[4]; lw:=arg[5];
  else ly:=CoxeterLength(W,y); lw:=CoxeterLength(W,w);
  fi;
  if ly=lw or not Bruhat(W,y,w,ly,lw) then return 0; fi;
  if lw=ly+1 then return 1; fi; 
  Lw:=List(LeftDescentSet(W,w),x->Position(W.reflectionsLabels,x));
  Rw:=List(RightDescentSet(W,w),x->Position(W.reflectionsLabels,x));

  ild:=W.operations.IsLeftDescending;
  if ForAny(Lw,s->not ild(W,y,s)) or
     ForAny(Rw,s->not ild(W,y^-1,s)) then return 0;
  fi;

  pol:=KazhdanLusztigPolynomial(W,y,w,ly,lw);
  if Degree(pol)=(lw-ly-1)/2 then return pol.coefficients[(lw-ly+1)/2];
  else return 0;
  fi;
end;


#############################################################################
##
#F  KazhdanLusztigCoefficient( <W>, <y>, <w>, [ <ly>, <lw> ], <k> ) . . . . .
#F  . . . . . . . . . . . . . . . coefficient of a Kazhdan-Lusztig polynomial
##
##  'KazhdanLusztigCoefficient'  returns the <k> th coefficient of the 
##  Kazhdan-Lusztig polynomial of the Coxeter group <W> corresponding to the 
##  elements <y> and <w>. The variables <ly> and <lw> contain the length
##  of <y> and <w>, respectively.  If <y> and <w> are not related by the 
##  Bruhat order, the  value 0 is returned.
##
KazhdanLusztigCoefficient:=function(arg) local W,y,w,ly,lw,k,s,Lw,Rw,pol;
  W:=arg[1]; y:=arg[2]; w:=arg[3];
  if Length(arg)>4 then ly:=arg[4]; lw:=arg[5]; k:=arg[6];
  else k:=arg[4]; ly:=CoxeterLength(W,y); lw:=CoxeterLength(W,w);
  fi;
  if k<0 or not IsInt(k) or ly>lw or (ly<lw and 2*k>lw-ly-1) 
                                         or not Bruhat(W,y,w,ly,lw) then
    return 0;
  fi;
  if k=0 then return 1; fi;
  pol:=KazhdanLusztigPolynomial(W,y,w,ly,lw);
  if k<=Degree(pol) then return pol.coefficients[k+1];
  else return 0;
  fi;
end;

#############################################################################
##
#F  KLMueMat( <W>, <list> )  . . . .  the matrix of leading coefficients of
#F  Kazhdan-Lusztig polynomials  of elements in a given list of reduced words
##
##  The elements in <list> must be ordered by increasing length. 
##
##  This should be fixed to return a square matrix.   Jean
##
KLMueMat:=function(W,c) 
  local m,i,j,k,ls,x,n,p,d,s,w0;
  w0:=LongestCoxeterElement(W); 
  n:=Length(c); 
  ls:=List(c,i->CoxeterLength(W,i));
  m:=List([1..n],i->[]);
  for k in [0..ls[n]-ls[1]] do
    for i in [1..n] do
      for j in [1..i] do
	d:=ls[i]-ls[j];
	if d=k then
	  if ls[i]+ls[j]>W.N  then
	     m[i][j]:=KazhdanLusztigMue(W,w0*c[i],w0*c[j],W.N-ls[i],W.N-ls[j]);
	  else 
             m[i][j]:=KazhdanLusztigMue(W,c[j],c[i],ls[j],ls[i]);
	  fi;
	fi;
      od;
    od;
  od;
  return m;
end;

DecomposedLeftCells := function ( W, c)
  local  Lleq, s1, s2, d, ce, i, j, m, rest, mue;
  Lleq := function ( W, mue, x, y )
    local  i;
    if x = y  then return false;
    elif x < y and mue[y][x] = 0  then return false;
    elif x > y and mue[x][y] = 0  then return false;
    else
      for i  in W.generatingReflections  do
        if IsLeftDescending(W,c[x],i) and not IsLeftDescending(W,c[y],i) then
          return true;
        fi;
      od;
      return false;
    fi;
  end;
  mue:=KLMueMat(W,c);
  ce := [  ];
  rest := [ 1 .. Length( c ) ];
  while rest <> [  ]  do
      s1 := [ rest[1] ];
      for j  in s1  do
          for i  in [ 1 .. Length( c ) ]  do
              if not i in s1 and Lleq( W, mue, j, i )  then Add( s1, i ); fi;
          od;
      od;
      Sort( s1 );
      s2 := [ rest[1] ];
      for j  in s2  do
          for i  in [ 1 .. Length( c ) ]  do
              if not i in s2 and Lleq( W, mue, i, j )  then Add( s2, i ); fi;
          od;
      od;
      Sort( s2 );
      d := Intersection( s1, s2 );
      Sort( d );
      m := [  ];
      for i  in [ 1 .. Length( d ) ]  do
          m[i] := [  ];
          for j  in [ 1 .. i ]  do m[i][j] := mue[d[i]][d[j]]; od;
      od;
      Add( ce, [ List( d, i->CoxeterWord(W,c[i])), m ] );
      SubtractSet( rest, d );
  od;
  return ce;
end;

#############################################################################
##
#F  LeftCells( <W> )  . . . . . . . . . . . . . . . . . . the left cells of W
##
##  'LeftCells' returns a list of pairs. The first  component  of  each  pair
##  consists of the reduced words in the Coxeter group <W> which lie in 
## one left cell C,  the  second  component  consists  of the corresponding 
##  matrix of  highest coefficients mue(y,w), where y,w are in C.
##
LeftCells := function ( arg )
  local  W, t, i, j, v, k, max, res, rw, x;
  W := arg[1];
  t := Combinations( W.reflectionsLabels{W.generatingReflections});
  if  Length( arg ) = 2  then max := arg[2];
  else max := Size(W);
  fi;
  rw := List([ 1 .. Length( t ) ],j->[  ]);
  for i  in Concatenation(List([0..W.N],i->CoxeterElements(W,i))) do
    Add( rw[Position( t, RightDescentSet( W, i ) )], i);
  od;
  res := [  ];
  for i  in rw  do
      if Length(i) <= max then
	  InfoChevie( "#I  R(w) = ", t[Position( rw, i )],
				   " :  Size = ", Length( i ) , "\c" );
	  x := DecomposedLeftCells( W, i);
	  if Length( x ) = 1  then
	      InfoChevie( ",  ", Length( x ), " new cell \n" );
	  else
	      InfoChevie( ",  ", Length( x ), " new cells \n" );
	  fi;
	  Append( res, x );
      fi;
  od;
  for v in res do
    for j in [1..Length(v[1])] do
      for k in [j+1..Length(v[1])] do v[2][j][k]:=v[2][k][j];
      od;
    od;
  od;
  return res;
end;

#############################################################################
##
#F  LeftCellRepresentation( <H> , <cell> ) . . . . . . . . . . representation
#F  of  the  Hecke  algebra  associated  to  a  left cell of  a  Coxeter group
##
##  'LeftCellRepresentation' returns a list of matrices giving the left  cell 
##  representation of the  Hecke algebra <H> associated to <cell>, given as
##  a pair with  first component  a  list of reduced  words which  form  a 
##  left cell, and  second component  the corresponding matrix of mue's.
##  Typically, <cell> is the result of the  function 'LeftCells'.
##
LeftCellRepresentation := function ( H, cell )
  local  s, i, j, rep, t, f1, f2, pc, c, mue, v, W;
  W:=Group(H);
  if Length(Set(List(W.generatingReflections,
                i->RootParameter(H,i,"Left Cell Representation"))))>1 then
  # this checks one can compute rootParameters
    Error("cell representations for unequal parameters not yet implemented");
  else 
    v := H.rootParameter[1];
  fi;
  c := cell[1];
  pc := List(c,i->EltWord(W,i));
  mue := cell[2];
  rep := [  ];
  for s  in W.generatingReflections  do
    f1 := List(pc,i->not IsLeftDescending(W,i,s));
    f2 := List(pc,i->Position( pc, W.reflections[s] * i ) );
    t := [  ];
    for i  in [ 1 .. Length( c ) ]  do
      t[i] := [  ];
      for j  in [ 1 .. Length( c ) ]  do
	if i = j  then
	  if f1[i]  then t[i][i] := v ^ 2;
	  else t[i][i] := -v^0;
	  fi;
	elif i > j  then
	  if f2[j] = i  then t[i][j] := v;
	  else t[i][j] := 0*v;
	  fi;
	else
	  if f1[j] and mue[j][i]<>0 and not f1[i] then t[i][j] := mue[j][i]*v;
	  else t[i][j] := 0*v;
	  fi;
	fi;
      od;
    od;
    Add( rep, t );
  od;
  return rep;
end;

########################################################################
#               Functions for W-graphs 
# (Jean Michel june/december 2003 from  code/data of Geck, Marin, Alvis,
# Naruse, Howlett,Yin)
# Chevie stores  some representations of simply-laced  Hecke algebras as
# $W$-graphs. These $W$-graphs are defined by a set of vertices $C$ with
# a function $I:C\to  Parts(S)$ and a map $\mu$ from  $C\times C$ to the
# field of definition  of $W$; it defines a representation  of the Hecke
# algebra with  single sqrtparameter $v$  on a space with  basis $e_y_{y
# \in C}$ by:
#
#  $$ T_s(e_y)=\cases{-e_y&                            if $s\in I(y)$\cr
#              v^2 e_y+\sum_{x\in C-\{y\}} v\mu(x,y)e_x&otherwise\cr}$$
#
# The W-graphs  are stored in a  compact format to save  space. They are
# represented  as a  pair. 
# -The first  element is a list  giving I(x); its elements  are either a
# set I(x), or an integer n  specifying to repeat the previous element n
# more times.
# -The second  element is a list  which specifies mu. We  first describe
# the mu-list for symmetric  W-graphs (when \mu(x,y)=\mu(y,x)). there is
# one element  of the mu-list  for each non-zero  value m taken  by \mu,
# which  consists  of  a  pair  whose  first  element  is  m  and  whose
# second  element is  a  list of  lists;  if  l is  one  of these  lists
# each  pair [l[1],l[i]]  represents an  edge (x=l[1],y=l[i])  such that
# \mu(x,y)=\my(y,x)=m. For non-symmetric W-graphs,  the first element of
# each  pair in  the  mu-list is  a  pair [m1,m2]  and  each edge  [x,y]
# obtained from the lists in the second element has to be interpreted as
# mu(x,y)=m1 and mu(y,x)=m2.

# The next function given a W-graph gr for some Hecke algebra of rank rk
# with sqrtparameter v constructs the rk matrices it specifies
WGraphToRepresentation:=function(rk,gr,v)local l,x,y,i,j,n,S,V,mu;
  V:=[];
  for S in gr[1] do 
    if  IsInt(S) then Append(V,List([1..S],i->V[Length(V)]));
    else Add(V,S);
    fi;
  od;
  n:=Length(V);
  S:=List([1..rk],i->IdentityMat(n)*v^2);
  for j in [1..n] do for i in V[j] do S[i][j][j]:=-v^0;od;od;
  for i in gr[2] do 
    if IsList(i[1]) then mu:=i[1];else mu:=[i[1],i[1]];fi;
    for l in i[2] do 
      x:=l[1];
      for y in l{[2..Length(l)]} do
        for j in Difference(V[y],V[x]) do S[j][y][x]:=mu[2]*v;od;
        for j in Difference(V[x],V[y]) do S[j][x][y]:=mu[1]*v;od;
      od;
    od;
  od;
  return S;
end;

# the next function returns the dual W-graph of gr (for an Hecke algebra of
# rank rk). A dual W-graph corresponds to a Curtis Dual representation.
DualWGraph:=function(rk,gr)
  return [List(gr[1],function(x)if IsInt(x) then return x;
                                else return Difference([1..rk],x);fi;end),
          List(gr[2],function(x)if IsList(x[1]) then return 
	  [-Reversed(x[1]),x[2]]; else return [-x[1],x[2]];fi;end)];
end;
