#############################################################################
##
#F  RelativeGroup(<W>,<J>)
#
#   W is  a Coxeter  group. If J  is not a  distinguished subset  of the
#   simple  roots (i.e.  stable  by w_0^K  F for  any  K containing  J),
#   returns false.
#   Else,  returns a  record  containing information  about the  Coxeter
#   group R=N_W(W_J)/W_J in its reflection representation on X(ZL_J/ZG).
#   (according to Lusztig's "Coxeter Orbits...", the images of the roots
#   of W in X(ZL_J/ZG) form a root system).
#
#     R is considered  both as a quotient with quotientmap  given by the
#     function R.MappingFromNormalizer, and as a subgroup of W generated
#     by  .parentmap= for i in R.relativeIndices, the elements w^0_J w^0_{J+i}
#     where .relativeIndices= indices of simple roots (Pi-J)
#
CoxeterGroupOpsRelativeGroup:=function(arg)local res,qr,I,indices,gens,r,W,J;
  W:=arg[1];J:=arg[2];
# Print("CoxeterGroupOps.RelativeGroup ",W,J," called \n");
  if not IsSubset(W.rootInclusion{W.generatingReflections},J) then
    Error("implemented only for standard parabolic subgroups");
  fi;
  indices:=Filtered(W.rootInclusion{W.generatingReflections},x->not x in J);
  # not Difference: we keep the order
  gens:=List(indices,
    i->LongestCoxeterElement(ReflectionSubgroup(W,Concatenation([i],J)))*
     LongestCoxeterElement(ReflectionSubgroup(W,J)));
  if ForAny(gens,g->Set(OnTuples(J,g))<>Set(J)) then 
    InfoChevie("*** Warning: ",J," is not distinguished in ",W,"\n");
    return false;
  fi;
  I:=W.rootRestriction{indices};
  qr:=i->W.roots[W.rootRestriction[i]]{I};
  if Length(indices)=0 then res:=CoxeterGroup();
  else res:=CoxeterGroup(List(indices,i->List(indices,j->
    ProportionalityCoefficient(qr(j)-qr(j^gens[Position(indices,i)]),qr(i)))));
  fi;
  res.relativeIndices:=indices;
  res.parentmap:=gens;
  res.MappingFromNormalizer:= # maps elts of N_W(W_J) to elts of res
    function(w)local c,r; c:=();
    while w<>() do
      r:=PositionProperty(indices,x->x^w>W.parentN);
      w:=gens[r]*w;
      c:=c*res.generators[r];
    od;
    return c;
    end;
  return res;
end;

CoxeterCosetOpsRelativeCoset:=function(WF,J)local res;
# Print("CoxeterCosetOps.RelativeCoset ",WF,J," called \n");
  res:=RelativeGroup(Group(WF),J);
  return CoxeterCoset(res,PermListList(res.parentmap,
     List(res.parentmap,x->x^WF.phi)));
end;

SpetsOpsRelativeCoset:=function(arg)local WF,W,R,res,t;
# Print("SpetsOpsRelativeCoset ",arg," called \n");
  WF:=arg[1];W:=Group(WF);
  R:=ApplyFunc(RelativeGroup,Concatenation([W],arg{[2..Length(arg)]}));
  res:=Spets(R,GetRelativeAction(W,ReflectionSubgroup(W,arg[2]),WF.phi));
  if arg[2]=[] then 
     Group(res).MappingFromNormalizer:=R.MappingFromNormalizer;
     return res;
# else Group(res).MappingFromNormalizer:=function(x)Error("MappingFromNormalizer failed");return false;end;
  fi;
  return res;
end;

TwistingElements:=function(WF,J)local L,e,W,W_L,H,h,N,WF_L,gens,i;
  W:=Group(WF);
  if J=[] then return List(ConjugacyClasses(W),Representative);fi;
  if IsCoxeterGroup(W) then
    h:=RepresentativeOperation(W,OnSets(Set(J),WF.phi),Set(J),OnSets);
    if h=false then Print("!!!! warning: no subspets for ",J,"\n");return [];fi;
    W_L:=Stabilizer(W,Set(J),OnSets);
    e:=List(ConjugacyClasses(Group(Concatenation(W_L.generators,
       [WF.phi*h]),W_L.identity)),Representative);
    e:=Filtered(e,x->WF.phi*h*x^-1 in W_L);
    return e*WF.phi^-1;
  fi;
  L:=ReflectionSubgroup(W,J); N:=Normalizer(W,L); W_L:=FactorGroup(N,L);
  if WF.phi<>() then Error("not implemented for twisted parent Spets");fi;
  if Size(W_L)>=10 then
    H:=Group(List(W_L.generators,x->GetRelativeAction(W,L,
         Representative(x.element))),IdentityMat(W.rank-L.semisimpleRank));
    if Size(H)=Size(W_L) then 
      h:=GroupHomomorphismByImages(H,W_L,H.generators,W_L.generators);
      h.isMapping:=true;
      e:=List(ConjugacyClasses(H),Representative);
      e:=List(e,x->Representative(Image(h,x).element));
      return e;
    fi;
  fi;
  return List(ConjugacyClasses(W_L),x->Representative(Representative(x).element));
end;

Twistings:=function(WF,J)
  if not IsSpets(WF) then WF:=Spets(WF);fi;
  return Filtered(List(TwistingElements(WF,J),
       x->SubSpets(WF,J,x)),x->not IsIdentical(x,false));
end;

# find series h of levi HF in WF (sers may be .harish or .almostHarish)
# return a record with 2 fields: ser= found series
#   op: an element of W which conjugates h.levi to ser.levi
#       (never needed for Coxeter groups and standard levis H)
FindSeriesInParent:=function(h,HF,WF,sers)local n,p,W,y;
  if IsSpets(WF) then W:=Group(WF); else W:=WF;fi;
  n:=Filtered(Split(h.cuspidalName),x->x<>"");Sort(n);
  for y in sers do
    if h.eigenvalue=y.eigenvalue and Length(h.levi)=Length(y.levi) then
      p:=Filtered(Split(y.cuspidalName),x->x<>"");Sort(p);
      if p=n then #JM: I should replace by more accurate code
        p:=RepresentativeOperation(W,
	  Set(Reflections(ReflectionSubgroup(W,h.levi))),
          Set(Reflections(ReflectionSubgroup(W,y.levi))),OnSets);
        if p<>false then  return rec(ser:=y,op:=p);fi;
      fi;
    fi;
  od;
  Error("series not found");
end;

# given a series h in WF, din the number of h.cuspidal in HF
FindCuspidalInLevi:=function(h,HF,WF)local strip,cusp;
  strip:=function(n)
    while Length(n)>0 and n[1]=',' do n:=n{[2..Length(n)]};od;
    while Length(n)>0 and n[Length(n)]=',' do n:=n{[1..Length(n)-1]};od;
    return n;
  end;
  cusp:=Position(List(UnipotentCharacters(HF).charNames,strip),
     strip(h.cuspidalName));
  if cusp=false then
    Error("cuspidal ",h.cuspidalName," not found in ",HF,"\n");
  fi;
  return cusp;
end;

# l is a list of linear forms in Mvps x1,..xn. The function returns
# values roots of unity for xi so that all linear forms take integral values.
# l is input as a list of vectors of length n (vector=coeffs of linear form)
FindIntSol:=function(l)local apply,vals,try,norm,stuck,simple,v;
  l:=l*List([1..Length(l[1])],
     function(i)if i=1 then return Mvp(1); else return Mvp(SPrint("x",i));
     fi;end);

# Print("FindIntSol",l,"\n");
  apply:=function(vals,v,val)local i,f,digits;
    digits:="0123456789";
    i:=0; for f in v{[2..Length(v)]} do i:=10*i+Position(digits,f)-1;od;
    if IsBound(vals[i]) then Error();fi;
    vals[i]:=val;
    for i in Filtered([1..Length(vals)],j->IsBound(vals[j]) and IsMvp(vals[j]))
    do vals[i]:=Value(vals[i],[v,val]);
       f:=ScalMvp(vals[i]);if f<>false then vals[i]:=f;fi;
    od;
    l:=List(l,x->Value(x,[v,val]));
#   Print(v,"->",val," vals=",vals,"\n");
    return vals;
  end;
      
  norm:=function(x)
    if IsRat(x) then if x<0 then return -x;else return x;fi;fi;
    return GetRoot(evalf(x*ComplexConjugate(x)),2);
  end;

  simple:=function(p)
    if Length(p.coeff)>2 or 
       (Length(p.coeff)=2 and Length(p.elm[1].elm)>0) then return false;fi;
    if Length(p.coeff)=1 then
      if Length(p.elm[1].elm)=0 then return p.coeff;fi;
      return [0,p.coeff[1],p.elm[1].elm[1]];
    fi;
    return [p.coeff[1],p.coeff[2],p.elm[2].elm[1]];
  end;

  try:=function()local p,v,N,i;
    for i in [1..Length(l)] do 
      p:=l[i];
      if Length(p.coeff)>0 and IsRat(p.coeff[1]) then 
	if Length(p.elm[1].elm)=0 then
	  p.coeff[1]:=p.coeff[1]-Int(p.coeff[1]);
	fi;
	if p.coeff[1]<0 then  p:=-p;fi;
	l[i]:=MvpOps.Normalize(p);
      fi;
    od;
    l:=Filtered(Set(l),x->not IsInt(ScalMvp(x)));
    if Length(l)=0 then return true;fi;
    for p in l do
      v:=simple(p);
      if v<>false then
	if Length(v)=1 then
	  InfoChevie("#I WARNING! FindIntSol failed:",p,"\n");return false;
	else
	  N:=NofCyc(v{[1,2]});if N mod 2<>0 then N:=2*N;fi;
	  i:=Filtered([0..N-1],j->IsInt(v[1]+v[2]*E(N)^j));
	  if Length(i)=1 then vals:=apply(vals,v[3],E(N)^i[1]);return true;
	  elif Length(i)=0 then
	    InfoChevie("#I WARNING! FindIntSol failed:",p,"\n");return false;
	  fi;
	fi;
      fi;
    od;
    for p in l do
      if Sum(p.coeff,norm)<15/16 then # the linear form must evaluate to 0
        i:=PositionProperty(p.elm,x->Length(x.elm)>0);
        v:=p.elm[i].elm[1];
        vals:=apply(vals,v,Mvp(v)-p/p.coeff[i]);return true;
      fi;
    od;
    stuck:=true;return false;
  end;

  vals:=[1];stuck:=false;
  while try() do if Length(l)=0 then return vals;fi;od;
  if stuck then
    if Length(l)=2 and ForAll(l,x->ForAll(2*x.coeff,IsInt)) and 
      ForAll(l,p->simple(p)<>false) then
      v:=Set(List(l,x->simple(x)[3]));
      if Length(v)=1 then return List([-1,1],x->apply(ShallowCopy(vals),v[1],x)); fi;
    fi;
    InfoChevie("#I WARNING! stuck in FindIntSol:",l,"\n");
  fi;
  return false;
end;


LusztigInductionPieces:=function(res)
  local LF,WF,W,L,h,uL,uW,hw,p,WFGL,LFGL,w,WGL,res,Jb,rh,L,ser,piece,i,j;
  WF:=res.g;LF:=res.u;
  uL:=UnipotentCharacters(LF);L:=Group(LF);
  uW:=UnipotentCharacters(WF);W:=Group(WF);
  if uL=false or uW=false then return false;fi;
  res.gNames:=function(t,option)return CharNames(uW,option);end;
  res.uNames:=function(t,option)return CharNames(uL,option);end;
  res.pieces:=[];
  hw:=uW.almostHarishChandra;
  for h in uL.almostHarishChandra do
    p:=FindSeriesInParent(h,LF,WF,hw);ser:=p.ser;
    L:=ReflectionSubgroup(W,ser.levi); # L^op contained in LF
    if IsBound(WF.isCoxeterCoset) then
      WFGL:=RelativeCoset(WF,h.levi);
      LFGL:=RelativeCoset(LF,h.levi);
      LFGL:=CoxeterSubCoset(WFGL,
	List(Group(LFGL).relativeIndices,
	  x->Position(Group(WFGL).relativeIndices,x)),
	    Group(WFGL).MappingFromNormalizer(LF.phi*WF.phi^-1));
    else
      Jb:=Concatenation(List(ser.relativeType,
         function(x)if IsBound(x.orbit) then return
          Concatenation(List(x.orbit,y->y.indices));
          else return x.indices;fi;end));
      WFGL:=RelativeCoset(WF,ser.levi,Jb);
      if WFGL=false then return false;fi;
      WGL:=Group(WFGL);
      if true then
        rh:=OnTuples(Concatenation(List(h.relativeType,
         function(x)if IsBound(x.orbit) then return
          Concatenation(List(x.orbit,y->y.indices));
          else return x.indices;fi;end)),p.op);
      else
	if p.op=() then rh:=Filtered(Jb,i->i in L.rootInclusion);
	else rh:=ReflectionSubgroup(W,
		OnTuples(L.rootInclusion{L.generatingReflections},p.op));
	  if not IsSubset(rh.generators,L.generators) then Error();fi;
	  rh:=Filtered(rh.rootInclusion{rh.generatingReflections},
	      i->not W.reflections[i] in W.reflections{ser.levi});
	fi;
      fi;
      w:=WGL.MappingFromNormalizer((LF.phi^p.op)*WF.phi^-1);
      if w=false then Error("Could not compute MappingFromNormalizer\n");fi;
      rh:=List(rh,function(x)local r;
	 r:=GetRelativeRoot(W,L,x);
	 return WGL.rootInclusion[Position(Reflections(WGL),
	    PermMatX(WGL,Reflection(r.root,r.coroot)))];end);
      LFGL:=SubSpets(WFGL,rh,w);
      ReflectionName(LFGL); ReflectionName(WFGL); # to shut up CharTable
    fi;
    piece:=rec(wnum:=ser.charNumbers,
	       hnum:=h.charNumbers,
	       uNames:=function(t,option)
	         return AlmostCharNames(uL,option){t.hnum};end,
	       gNames:=function(t,option)
	         return AlmostCharNames(uW,option){t.wnum};end,
 	       scalar:=ComplexConjugate(InductionTable(LFGL,WFGL).scalar),
#              scalar:=InductionTable(LFGL,WFGL).scalar,
	       u:=LFGL,
	       g:=WFGL,
	       what:="piece",
	       operations:=InductionTableOps);
    if h.levi=[] then piece.head:=function(t,option)
      if IsBound(option.TeX) then
       return SPrint("from $",ReflectionName(LF,option),
                     "$ to $",ReflectionName(WF,option),"$");
      else return SPrint("from ",ReflectionName(LF,option),
                     " to ",ReflectionName(WF,option));
      fi;end;
    else piece.head:=function(t,option) 
      if IsBound(option.TeX) then return SPrint(
      "from $W_{",ReflectionName(LF,option),"}(L_{",Join(h.levi),"},",
       h.TeXCuspidalName,")=",ReflectionName(t.u,option),
      "$\nto $W_{",ReflectionName(WF,option),"}(L_{",Join(ser.levi),"},",
       ser.TeXCuspidalName,")=",ReflectionName(t.g,option),"$");
      else return SPrint(
      "from W_",ReflectionName(LF,option),"(L[",Join(h.levi),"],",
       h.cuspidalName,")=",ReflectionName(t.u,option),
      "\nto   W_",ReflectionName(WF,option),"(L[",Join(ser.levi),"],",
       ser.cuspidalName,")=",ReflectionName(t.g,option));
      fi;end;
    fi;
    Add(res.pieces,piece);
  od;
  return res;
end;

# arguments (LF,WF[,force])
# if there is a third argument then return res (with the pieces)
# even in case of failure
LusztigInductionTable:=function(arg)
  local LF,WF,res,uW,uL,hh,scalars,fL,fWinv,map,maps,ret,p;
  LF:=arg[1];WF:=arg[2];
  if not IsSpets(WF) then WF:=Spets(WF);fi;
  if not IsSpets(LF) then LF:=Spets(LF);fi;
# if not IsBound(WF.isCoxeterCoset) and WF.phi<>() then 
#   CHEVIE.Diagnose("Non-trivial parent Spets not implemented");
#   return false;
# fi;
  res:=rec(u:=LF, g:=WF, what:="LusztigInductionTable",
      head:=function(t,option)
        if IsBound(option.TeX) then 
	  return SPrint("Lusztig Induction from $",ReflectionName(t.u,option),
	  "$ to $",ReflectionName(t.g,option),"$");
	else return SPrint("Lusztig Induction from ",
	  ReflectionName(t.u,option)," to ",ReflectionName(t.g,option));fi;end,
	  operations:=InductionTableOps);
  uW:=UnipotentCharacters(WF);
  if IsBound(CHEVIE.Cache.LusztigInduction) then
    if not IsBound(uW.inductionmaps) then uW.inductionmaps:=rec();fi;
    res.name:=ConcatenationString("Lusztig[",
   IntListToString(Group(LF).rootInclusion{[1..Group(LF).semisimpleRank]}),"]");
    if LF.phi<>WF.phi then
      if IsBound(WF.isCoxeterCoset) then
	Append(res.name,IntListToString(CoxeterWord(WF,LF.phi)));
      else
	Append(res.name,String(LF.phi*WF.phi^-1));
      fi;
    fi;
    if IsBound(uW.inductionmaps.(res.name)) then
      res.scalar:=uW.inductionmaps.(res.name);
      return res;
    fi;
  fi;
  res:=LusztigInductionPieces(res);if res=false then return false;fi;
  uL:=UnipotentCharacters(LF);
  fL:=UnipotentCharactersOps.Fourier(uL);
  hh:=uL.almostHarishChandra;
  fWinv:=UnipotentCharactersOps.FourierInverse(uW);
  maps:=List([1..Length(hh)],function(i)local piece;
    map:=List(uW.charNames,y->List(uL.charNames,x->0));
    piece:=res.pieces[i];
    map{piece.wnum}{piece.hnum}:=piece.scalar;
    return fWinv*map*fL;
  end);
  ret:=function(map)
    if map=false then 
       InfoChevie("#I WARNING: failed\n");
       if Length(arg)=2 then return false;
       else return res;
       fi;
    fi;
    res.scalar:=map;
    if IsBound(CHEVIE.Cache.LusztigInduction) then
      uW.inductionmaps.(res.name):=res.scalar;
    fi;
    return res;
  end;
  map:=Sum(maps);
  if ForAll(map,v->ForAll(v,IsInt)) then return ret(map);fi;
  if LF.phi=WF.phi then Error("non-positive RLG for untwisted L");fi;
  scalars:=FindIntSol(Set(TransposedMat(List(maps,Concatenation))));
  if scalars=false then return ret(scalars);fi;
  if ForAny(scalars,IsList) then
    CHEVIE.Diagnose("#I WARNING: ambiguity in scalars: ",scalars,"\n");
    scalars:=scalars[1];
  fi;
  if ForAny(scalars,x->x<>1) then
    if ForAll(scalars,IsInt) then p:="#I signs are ";
    else p:="#I scalars are ";
    fi;
#   CHEVIE.Diagnose(p,Join(List([1..Length(scalars)],function(i)local n;
#     n:=hh[i].cuspidalName;if n="" then n:="Id";fi;
#     return SPrint(n,"->",scalars[i]);end),";"),"\n");
  fi;
  return ret(Sum([1..Length(scalars)],i->maps[i]*scalars[i]));
end;

HCInductionTable:=function(HF,WF)
  local W,H,h,uh,uw,name,hw,hh,p,Wi,Hi,rh,Jb,t,N,pieces,piece,L;
  if not IsSpets(WF) then WF:=Spets(WF); fi;
  uw:=UnipotentCharacters(WF); W:=Group(WF);
  if not IsSpets(HF) then HF:=Spets(HF); fi;
  uh:=UnipotentCharacters(HF); H:=Group(HF);
  if not IsBound(uw.inductionmaps) then
    uw.inductionmaps:=rec();
  fi;
# Print("HF=",HF," WF=",WF,"\n");
  name:=ConcatenationString("HC[",
     IntListToString(H.rootInclusion{H.generatingReflections}),"]");
  pieces:=[];
  if not IsBound(uw.inductionmaps.(name)) then
    uw.inductionmaps.(name):=List(uw.charNames,y->List(uh.charNames,z->0));
    hw:=uw.harishChandra; hh:=uh.harishChandra;
    for h in hh do
      p:=FindSeriesInParent(h,HF,WF,hw);
      for t in p.ser.relativeType do t.operations:=ReflTypeOps;od;
      if IsCoxeterCoset(WF) then
	Wi:=ApplyFunc(ReflectionGroup,p.ser.relativeType);
#	InfoChevie("W_",ReflectionName(WF),"(",IntListToString(p.ser.levi),")=",
#	   ReflectionName(Wi),"\n");
	     Jb:=Concatenation(List(p.ser.relativeType,x->x.indices));
	if p.op<>() then
	  rh:=RelativeGroup(H,h.levi).generators;
	  rh:=Filtered(Arrangements([1..Length(Wi.generators)],Length(rh)),
	    a->ForAll([1..Length(a)],i->OrderPerm(Wi.generators[a[i]])=
						  OrderPerm(rh[i])));
	  if Length(rh)>1 then
	      InfoChevie("# WARNING ***: embedding ambiguous:",rh,"\n");
	  fi;
	  Hi:=ReflectionSubgroup(Wi,rh[1]);
	else Hi:=ReflectionSubgroup(Wi,
	       List(Concatenation(List(h.relativeType,x->x.indices)),
		 x->PositionProperty(Jb,y->x in Cycle(WF.phi,y))));
	fi;
      else
	Jb:=Concatenation(List(p.ser.relativeType,x->x.indices));
	  L:=ReflectionSubgroup(W,p.ser.levi);
	  Wi:=RelativeGroup(W,p.ser.levi,Jb);
	  if p.op=() then rh:=Filtered(Jb,i->Reflection(W,i) in H);
	  else rh:=ReflectionSubgroup(W,
		  OnTuples(H.rootInclusion{H.generatingReflections},p.op));
	    if not IsSubset(rh.generators,L.generators) then Error();fi;
	    rh:=Filtered(rh.rootInclusion{rh.generatingReflections},
		i->not W.reflections[i] in W.reflections{p.ser.levi});
	  fi;
	  rh:=List(rh,function(x)local r;
	     r:=GetRelativeRoot(W,L,x);
	     return Position(Reflections(Wi),
		PermMatX(Wi,Reflection(r.root,r.coroot)));end);
	  Hi:=ReflectionSubgroup(Wi,rh);
      fi;
 #    ReflectionName(Wi);ReflectionName(Hi); # to shut up CharTable
      piece:=rec(uNames:=function(t,option)return CharNames(uh,option){h.charNumbers};end,
		 gNames:=function(t,option)return CharNames(uw,option){p.ser.charNumbers};end,
		 scalar:=InductionTable(Hi,Wi).scalar,
		 u:=Hi,
		 g:=Wi,
		 what:="piece",
		 operations:=InductionTableOps);
      Add(pieces,piece);
      uw.inductionmaps.(name){p.ser.charNumbers}{h.charNumbers}:=piece.scalar;
    od;
  fi;
  return rec(u:=HF, g:=WF, what:="HCInductionTable",
    uNames:=function(t,option)return CharNames(uh,option);end,
    gNames:=function(t,option)return CharNames(uw,option);end,
    head:=function(t,option)
      if IsBound(option.TeX) then return SPrint("Harish-Chandra Induction from $",
       ReflectionName(t.u,option),"$ to $",ReflectionName(t.g,option),"$");
      else return SPrint("Harish-Chandra Induction from ",
       ReflectionName(t.u,option)," to ",ReflectionName(t.g,option));fi;
    end,
    operations:=InductionTableOps,
    scalar:=uw.inductionmaps.(name),
    pieces:=pieces
  );
end;
