#############################################################################
##
#A  uc.g      Unipotent Characters    Frank Luebeck and Jean Michel
##
#Y  Copyright 1996-1998, Rwth Aachen and Univ. Paris VII
##  
##  This file contains basic functions for unipotent characters of
##  Coxeter groups, Coxeter cosets and Spets.
##  

UnipotentCharactersOps:=OperationsRecord("UnipotentCharOps");

UnipotentCharactersOps.Print:=function(r)
  Print("UnipotentCharacters( ", ReflectionName(r.group), " )");
end;

# Fourier times the vector of unip. degrees is the vector of fake degrees
UnipotentCharactersOps.Fourier:=function(uc)local f;
  if not IsBound(uc.fourier) then
    uc.fourier:=IdentityMat(Length(uc.charNames));
    for f in uc.families do 
      uc.fourier{f.charNumbers}{f.charNumbers}:=f.fourierMat;
    od;
  fi;
  return uc.fourier;
end;

# FourierInverse times the vector of fake degrees is the vector of unip degrees
UnipotentCharactersOps.FourierInverse:=function(uc)local f;
  if not IsBound(uc.fourierinverse) then
    uc.fourierinverse:=IdentityMat(Length(uc.charNames));
    for f in uc.families do 
      uc.fourierinverse{f.charNumbers}{f.charNumbers}:=
	  TransposedMat(ComplexConjugate(f.fourierMat));
    od;
  fi;
  return uc.fourierinverse;
end;

UnipotentCharactersOps.Eigenvalues:=function(uc,s)local f;
  if not IsBound(uc.eigenvalues) then
    uc.eigenvalues:=[];
    for f in uc.families do uc.eigenvalues{f.charNumbers}:=f.eigenvalues;od;
  fi;
  return uc.eigenvalues{s};
end;

UnipotentCharactersOps.ParamsAndNames:=function(sers) # a field to hide its name
  local res, ser, t, nm, params;
  res:=rec(charParams:=[],charNames:=[],TeXCharNames:=[]);
  for ser in sers do
    t:=ser.relativeType;
    nm:=CHEVIE.Load("CharName",t).fct;
    params:=ReflTypeOps.ChevieCharInfo(t).charparams;
    res.charParams{ser.charNumbers}:=List(params, x->[ser.cuspidalName,x]);
    res.charNames{ser.charNumbers}:=List(params, function(x)local name;
      name:=Copy(ser.cuspidalName);
      if IsBound(t.orbit) then t.rank:=t.orbit[1].rank;fi;
      if Length(name)>0 and t.rank>0 then Add(name,':');fi;
      if t.rank>0 then  Append(name,nm(x,rec()));fi;
      return String(name);
      end);
    res.TeXCharNames{ser.charNumbers}:=List(params, function(x)local name;
      if not IsBound(ser.TeXCuspidalName) 
      then ser.TeXCuspidalName:=ser.cuspidalName;
      fi;
      name:=Copy(ser.TeXCuspidalName);
      if Length(name)>0 and t.rank>0 then Add(name,':');fi;
      if t.rank>0 then  Append(name,nm(x,rec(TeX:=1)));fi;
      return String(name);
      end);
  od;
  return res;
end;

FixRelativeType:=function(t)
# fix illegal relativeTypes B1 and C2 which appear in HC or almost HC
# series of classical groups
  if t.relativeType.series="B" then
    if t.relativeType.rank=1 then
      t.relativeType.series:="A";
      t.charNumbers{[1,2]}:=t.charNumbers{[2,1]}; # map B1->A1
    elif t.relativeType.rank=2 and IsBound(t.relativeType.cartanType) and
      t.relativeType.cartanType=1 then
      t.relativeType.cartanType:=2;
      t.relativeType.indices:=Reversed(t.relativeType.indices);
      t.charNumbers{[1,5]}:=t.charNumbers{[5,1]}; # map C2->B2
      if IsBound(t.parameterExponents) then
	t.parameterExponents:=Reversed(t.parameterExponents);
      fi;
    fi;
  fi;
end;

#############################################################################
##
#F  UnipotentCharacters( <rdr> ) . . . . . information about the
#F  unipotent characters of G(q) given by its complete root datum.
##  Works for Spets.
##  
#   The fields are as follows:
#
#   harishChandra: the Harish-Chandra series
#    each is a record with fields
#     levi: The indices in the diagram of W of the Levi L^F where is the
#           cuspidal (Given in terms of W's parent)
#     cuspidalName: The name of that cuspidal
#     eigenvalue:   Its Frobenius Eigenvalue
#     relativeType: A record describing  W_G^F(L). In addition to series
#        and  rank, indices  is the  index in  W's parent  of reflexions
#        whose F-orbits generate W_G^F(L)
#     parameterExponents: In the same order as relativeType.indices, the
#        exponents of  the powers of q  which are the parameters  of the
#        relative Hecke algebra.
#     charNumbers: the  numbers of the  characters in the HC  series, in
#        the same order as the characters of W_G^F(L)
#
#   almostHarishChandra: the almost Harish-Chandra series
#    each is a record with fields
#     levi: The indices in the diagram of W of the Levi L where is the
#           cuspidal character sheaf (Given in terms of W's parent)
#     cuspidalName: The name of that cuspidal
#     eigenvalue:   Its Frobenius Eigenvalue
#     relativeType:  A  record  describing (W_G(L),F).  In  addition  to
#        series  and  rank,  indices  is  the index  in  W's  parent  of
#        reflexions which  generate W_G(L), and the  action of Frobenius
#        is described.
#     charNumbers:  the  numbers of  the  characters  in the  almost  HC
#        series, in the same order as the characters of W_G(L)
#
#   The argument must be the reflctiontype  of a Spets

ReflTypeOps.UnipotentCharacters:=function(t) local uc,a,s,i;
  if t.orbit[1].series="B" and IsBound(t.orbit[1].cartanType) 
     and t.orbit[1].cartanType=1 then 
       uc:=CHEVIE.Data("UnipotentCharacters",t,t.orbit[1].cartanType);
  else uc:=CHEVIE.Data("UnipotentCharacters",t);
  fi;
  if uc=false then return uc;fi;
  Inherit(uc,UnipotentCharactersOps.ParamsAndNames(uc.harishChandra));
  if not IsBound(uc.charSymbols) then
    uc.charSymbols:=uc.charParams;
  fi;
  if false then
  # JM experimental code 4-2007: adjust principal series for scalar-twisted
  # should work only if order of scalar prime to ZW
  if IsBound(t.scalar) and t.scalar[1]<>1 then
    if Length(t.orbit)>1 then Error("pas envisage");fi;
    uc.families:=Copy(uc.families);
    for s in uc.families do
      a:=List(s.charNumbers,i->t.scalar[1]^(-uc.a[i]-uc.A[i]));
      s.eigenvalues:=Zip(s.eigenvalues,a,function(x,y)return x*y;end);
      s.fourierMat:=TransposedMat(Zip(TransposedMat(s.fourierMat),a,
            function(x,y)return x*y;end));
    od;
  fi;
  fi;
  # adjust things for descent of scalars
  # we would like to adjust indices so they fit with those stored in t
  # but we cannot when indices mention non-generating reflections!
  a:=Length(t.orbit);
  if IsBound(uc.a) then uc.a:=a*uc.a;fi;
  if IsBound(uc.A) then uc.A:=a*uc.A;fi;
  for s in uc.harishChandra do
    s.parameterExponents:=a*s.parameterExponents;
    s.eigenvalue:=s.eigenvalue^a;
    s.cuspidalName:=Join(List([1..a],i->s.cuspidalName));
    s.TeXCuspidalName:=Join(List([1..a],i->s.TeXCuspidalName),"\\otimes");
  od;
  if not IsBound(uc.almostHarishChandra) then
    uc.almostHarishChandra:=List(uc.harishChandra, function(s)local res,a;
      res:=rec();
      Inherit(res,s,["levi", "cuspidalName", "TeXCuspidalName", "eigenvalue", 
        "charNumbers"]);
      res.relativeType:=rec(orbit:=[ShallowCopy(s.relativeType)],twist:=());
      if t.twist<>() then
	a:=t.orbit[1].indices{s.relativeType.indices};
	res.relativeType.twist:=MappingPermListList(a,OnTuples(a,t.twist));
      fi;
      return res;
      end);
  else
    for s in uc.almostHarishChandra do
      if not IsBound(s.relativeType.orbit) then 
	s.relativeType:=rec(orbit:=[s.relativeType],twist:=());
      fi;
    od;
  fi;
  if not IsBound(uc.almostCharSymbols) then
    uc.almostCharSymbols:=uc.charSymbols;
  fi;
  a:=UnipotentCharactersOps.ParamsAndNames(uc.almostHarishChandra);
  for s in [["almostCharParams","charParams"],
    ["almostCharNames","charNames"],["TeXAlmostCharNames","TeXCharNames"]]
  do uc.(s[1]):=a.(s[2]);od;
  return uc;
end;

HasTypeOps.UnipotentCharacters:=function(WF) 
  local CartesianSeries, type, simp, t, H, s, a, r, f, tmp, uc, i, res, W;

  CartesianSeries:=function(sers)local res;
    res:=rec();
    res.levi:=Concatenation(List(sers,x->x.levi));
    res.relativeType:=Filtered(List(sers,x->x.relativeType),x->x.rank<>0);
    if IsBound(sers[1].eigenvalue) then
      res.eigenvalue:=Product(List(sers,x->x.eigenvalue));
    fi;
    if ForAll(sers,x->IsBound(x.parameterExponents)) then
      res.parameterExponents:=Concatenation(List(sers,x->x.parameterExponents));
    fi;
    res.charNumbers:=Cartesian(List(sers,x->x.charNumbers));
    res.cuspidalName:=Join(List(sers,x->x.cuspidalName));
    res.TeXCuspidalName:=Join(List(sers,x->x.TeXCuspidalName),"\\otimes");
    return res;
  end;

  type:=ReflectionType(WF);
  if Length(type)=0 then
  # UnipotentCharacters(CoxeterGroup());
    return rec( 
      harishChandra:=[
	rec(relativeType:=[rec(series:="A",indices:=[],rank:=0)], 
	    levi:=[], parameterExponents:=[],
	    cuspidalName:="", eigenvalue:=1, charNumbers :=[ 1 ])],
      families := [Family("C1",[1])],
      charParams := [ [ "", [ 1 ] ] ],
      charNames := [ "" ],
      TeXCharNames:= [""],
      charSymbols := [ [ "", [ 1 ] ] ],
      almostHarishChandra:=[
        rec(relativeType:=[rec(series:="A",indices:=[],rank:=0)], 
            levi:=[], 
            cuspidalName := "", eigenvalue:=1, charNumbers := [1])],
      almostCharSymbols := [ [ "", [ 1 ] ] ],
      almostCharParams := [ [ "", [ 1 ] ] ],
      almostCharNames := [ "" ],
      a := [ 0 ],
      A := [ 0 ],
      group:=WF,
      operations:=UnipotentCharactersOps);
  fi;

  simp:=[]; W:=Group(WF);
  for t in ReflectionType(WF) do
# adjust indices of Levis, almostLevis, relativetypes so they fit with
# the parent of WF
    uc:=ReflTypeOps.UnipotentCharacters(t);
    if uc=false then return false;fi;
    H:=List(t.orbit,x->ReflectionSubgroup(W,W.rootInclusion{x.indices}));
    for s in uc.harishChandra do
      s.levi:=Concatenation(List(H,x->x.rootInclusion{s.levi}));
      s.relativeType.indices:=H[1].rootInclusion{s.relativeType.indices};
    od;
    for s in uc.almostHarishChandra do
      s.levi:=Concatenation(List(H,x->x.rootInclusion{s.levi}));
      s.relativeType.orbit:=Concatenation(List(H,x->
        List(s.relativeType.orbit,function(r)
	  r:=ShallowCopy(r);
	  r.indices:=x.rootInclusion{r.indices};
	  return r;end)));
      s.relativeType.twist:=s.relativeType.twist^MappingPermListList(
         [1..Length(H[1].roots)],H[1].rootInclusion);
    od;

    Add(simp,uc);
  od;

  # "Kronecker product" of records in simp:
  r:=simp[1];f:=RecFields(r);res:=rec();
  for a in f do
    if Length(simp)=1 then res.(a):=List(r.(a),x->[x]);
    elif ForAll(simp,x->IsBound(x.(a))) then
      res.(a):=Cartesian(List(simp,x->x.(a)));
    fi;
  od;
  
  for a in ["charNames","TeXCharNames","almostCharNames","TeXAlmostCharNames"]
  do res.(a):=List(res.(a),Join);od;
  
  for a in [ "harishChandra", "almostHarishChandra" ] do
    res.(a):=List(res.(a),CartesianSeries);
  od;

  res.families:=List(res.families,x->ApplyFunc(CartesianFamilies,x));
  
  for a in ["a", "A"] do 
    if IsBound(res.(a)) then res.(a):=List(res.(a),Sum);fi;
  od;

  # finally the new 'charNumbers' lists
  tmp:=Cartesian(List(simp,a->[1..Length(a.charNames)]));
  for a in [ "harishChandra", "almostHarishChandra", "families"] do
    for s in res.(a) do
      s.charNumbers:=List(s.charNumbers,y->Position(tmp,y));
    od;
  od;

  # trying to save some memory
  for a in ["charSymbols","almostCharParams"] do
    if res.(a)=res.charParams then res.(a):=res.charParams; fi;
  od;
  if res.almostCharSymbols=res.almostCharParams then
    res.almostCharSymbols:=res.almostCharParams;
  fi;
  if res.almostCharNames=res.charNames then
    res.almostCharNames:=res.charNames;
  fi;
  if Length(res.almostHarishChandra)=Length(res.harishChandra) then
  for i in [1..Length(res.almostHarishChandra)] do
    for a in [ "levi", "cuspidalName", "charNumbers"] do
      if res.almostHarishChandra[i].(a)=res.harishChandra[i].(a) then
	res.almostHarishChandra[i].(a):=res.harishChandra[i].(a);
      fi;
    od;
  od;
  fi;
  
  res.group:=WF;
  res.operations:=UnipotentCharactersOps;
  return res;
end;

UnipotentCharactersOps.CharNames:=function(uc,option)
  if IsBound(option.TeX) then return uc.TeXCharNames;
  else return uc.charNames;
  fi;
end;

UnipotentCharactersOps.AlmostCharNames:=function(uc,option)
  if IsBound(option.TeX) then return uc.TeXAlmostCharNames;
  else return uc.almostCharNames;
  fi;
end;

UnipotentCharacters:=function(W)
  if IsSpets(W) then return 
    ChevieAttributeDispatcher("UnipotentCharacters","unipotentCharacters",[W]);
  else return UnipotentCharacters(Spets(W));
  fi;
end;

UnipotentDegrees:=function(W,q)local uc,v;
  uc:=UnipotentCharacters(W);v:=X(Cyclotomics);
  if not IsBound(uc.degrees) then 
    uc.degrees:=UnipotentCharactersOps.FourierInverse(uc)*FakeDegrees(uc,v);
  fi;
  if IsIdentical(q,v) then return uc.degrees;
  else return List(uc.degrees,x->Value(x,q));
  fi;
end;

CycPolUnipotentDegrees:=function(W)local uc;
  uc:=UnipotentCharacters(W);
  if not IsBound(uc.CycPolDegrees) then 
    uc.CycPolDegrees:=List(UnipotentDegrees(W,X(Cyclotomics)),CycPol);
  fi;
  return uc.CycPolDegrees;
end;

UnipotentCharactersOps.FakeDegrees:=function(uc,q)local fd;
  fd:=[1..Length(uc.charNames)]*0*q;
  fd{uc.almostHarishChandra[1].charNumbers}:=FakeDegrees(uc.group,q);
  return fd;
end;

UnipotentCharactersOps.Format:=function(uc,option)local title,e,s,uc,ud,fd,
      name,i,line,W,q,option,m,labels,res,v,l,n,f,TeX,LaTeX,makeline,head,
      delta1s,names;
   TeX:=IsBound(option.TeX);
   LaTeX:=IsBound(option.LaTeX);
   W:=uc.group;
   name:=ReflectionName(W,option);
   if IsBound(option.items) then title:=option.items;
   elif uc.charSymbols<>uc.charParams then 
     title:=["n0","Name","Symbol","Degree","FakeDegree","Eigenvalue","Family"];
   else 
     title:=["n0","Name","Degree","FakeDegree","Eigenvalue","Family"];
   fi;
   if "Family" in title then Add(title,"Delta1");fi;
   if IsBound(option.chars) then v:=option.chars;
   else v:=[1..Length(uc.charNames)];
   fi;
   if LaTeX then
     res:=SPrint("\\begin{center}Unipotent characters for $",name,
                 "$\\end{center}\n");
   elif TeX then
     res:=SPrint("\\centerline{Unipotent characters for $",name,"$}\n");
   else
     res:=SPrint("Unipotent characters for ",name,"\n");
   fi;
   head:=[];
   for n in title do
     if n="Degree" then ud:=CycPolUnipotentDegrees(W);
       if TeX then n:="Deg($\\gamma$)";fi;
     elif n="FakeDegree" then 
       q:=X(Cyclotomics);q.name:="q";fd:=FakeDegrees(uc,q);
     elif n="Eigenvalue" then e:=Eigenvalues(uc,[1..Length(uc.charNames)]);
       if TeX then n:="Fr($\\gamma$)";fi;
     elif n="Name" and TeX then n:="$\\gamma$";
     elif n="Family" then
      n:="Label";
      labels:=[];
      for f in uc.families do
        if TeX then labels{f.charNumbers}:=f.TeXCharLabels;
        else labels{f.charNumbers}:=f.charLabels;fi;
      od;
     elif n="Delta1" then
       if TeX then n:="$\\Delta$";fi;
       delta1s:=[];
       for f in uc.families do
         if IsBound(f.delta1) then delta1s{f.charNumbers}:=f.delta1;
	 else  delta1s{f.charNumbers}:=f.charNumbers*0+1;
	 fi;
       od;
     fi;
     if LaTeX then Add(head,SPrint("\\mbox{",n,"}"));
     elif TeX then Add(head,SPrint("\\hbox{",n,"}"));
     else Add(head,n);
     fi;
   od;
   if TeX then names:=uc.TeXCharNames; else names:=uc.charNames;fi;
   names:=ShallowCopy(names);
   makeline:=function(i)local line,n,o;
     line:=[];
     for n in title do
       if n="n0" then o:=i;
       elif n="Name" then o:=names[i];
       elif n="Symbol" then o:=Join(List(uc.charSymbols[i],StringSymbol));
       elif n="Degree" then o:=ud[i];
       elif n="FakeDegree" then o:=CycPol(fd[i]);
       elif n="Eigenvalue" then o:=e[i];
       elif n="Family" then o:=labels[i];
       elif n="Delta1" then o:=delta1s[i];
       else Error("field ",n);
       fi;
       Add(line,Format(o,option));
     od;
     return line;
   end;
   if IsBound(option.byFamily) then 
      if LaTeX then 
        Append(res,"\\medskip\\begin{center}{Trivial families}\\end{center}\n");
      elif TeX then Append(res,"\\medskip\\centerline{Trivial families}\n");
      else Append(res,"Trivial families\n");
      fi;
      m:=head{[1..Length(head)-2]};Add(m,"\\mbox{Family n$^0$}");
      m:=[m];
      for f in Filtered(uc.families,x->Length(x.charNumbers)=1) do
        s:=makeline(f.charNumbers[1]);s:=s{[1..Length(s)-2]};
        Add(s,Position(uc.families,f));Add(m,s);
      od;
      Append(res,FormatTable(m,option));
      for f in Filtered(uc.families,x->Length(x.charNumbers)>1) do
	if TeX then n:=f.TeXName; else n:=f.name; fi;
        if LaTeX then Append(res,
          SPrint("\\medskip\\begin{center}{Family $n^0.",
	    Position(uc.families,f),"$. of type $",n,"$}\\end{center}\n"));
	  if IsBound(f.comment) then Append(res,
	    SPrint("\\begin{center}{\\it ",f.comment,"}\\end{center}\n"));fi;
        elif TeX then
          Append(res,SPrint("\\medskip\\centerline{Family $",n,"$}\n"));
        else Append(res,SPrint("Family ",n,"\n"));
        fi;
	s:=f.charNumbers[f.special];
	if TeX then names[s]:=Concatenation("*\\hfill ",names[s]);
	else  names[s]:=Concatenation("*",names[s]);
	fi;
	if IsBound(f.cospecial) and f.cospecial<>f.special then
	  s:=f.charNumbers[f.cospecial];
	  if TeX then names[s]:=Concatenation("\\#\\hfill ",names[s]);
	  else names[s]:=Concatenation("#",names[s]);
	  fi;
        fi;
	if IsBound(f.delta1) and ForAny(f.delta1,x->x<>1) then
	  m:=[head];
	  Append(m,List(f.charNumbers,makeline));
	else
	  m:=[head{[1..Length(head)-1]}];
	  Append(m,List(List(f.charNumbers,makeline),x->x{[1..Length(x)-1]}));
	fi;
        Append(res,FormatTable(m,option));
      od;
   else m:=[head];Append(m,List(v,makeline));
        Append(res,FormatTable(m,option));
   fi;
   return res;
end;

UnipotentCharactersOps.Display:=function(uc,option)
  Print(Format(uc,option));
end;

UnipotentCharactersOps.LowestPowerGenericDegrees:=function(uc)local uc,ud;
  ud:=CycPolUnipotentDegrees(Group(uc));
  return List(ud{uc.harishChandra[1].charNumbers},x->x.valuation);
end;

UnipotentCharactersOps.HighestPowerGenericDegrees:=function(uc)local uc,ud;
  ud:=CycPolUnipotentDegrees(Group(uc));
  return List(ud{uc.harishChandra[1].charNumbers},Degree);
end;

if not IsBound(CHEVIE.families) then ReadChv("unip/families");fi;
