# series.g J.Michel 4/2003
# This file contains code to work with d-Harish-Chandra series of Spetses
#------------------------- Utility functions ---------------------------

# DRank(W d) dimension of the maximal E(d)-eigenspace of an element 
# of [Spets] W 
DRank:=function(W,d)local deg;
  if IsSpets(W) then deg:=ReflectionDegrees(Group(W));
       return Number([1..Length(deg)],x->E(d)^deg[x]/PhiFactors(W)[x]=1);
  else return Number(ReflectionDegrees(W),x->x mod d=0);
  fi;
end;

# DRank  for the part where W acts non-trivially
SemiSimpleDRank:=function(W,d)local deg;
  if IsSpets(W) then deg:=ReflectionDegrees(Group(W));
       return Number([1..Length(deg)],x->deg[x]<>1 and E(d)^deg[x]/PhiFactors(W)[x]=1);
  else return Number(ReflectionDegrees(W),x->x<>1 and x mod d=0);
  fi;
end;

# CuspidalUnipotentCharacters(WF[,d]) indices of the unipotent characters
# of the Spets WF which are d-cuspidal (d=1 if not specified)
CuspidalUnipotentCharacters:=function(arg)local WF,d,ad,ud;
  WF:=arg[1];
  if Size(WF)=1 then return [1];fi;
  if Length(arg)>1 then d:=arg[2];else d:=1;fi;
  ad:=SemiSimpleDRank(WF,d);
# if ad=0 then Error(d," should divide one of the degrees");fi;
  ud:=CycPolUnipotentDegrees(WF);
  return Filtered([1..Length(ud)],i->ad=Valuation(ud[i],d));
end;

# all orders of elements of W
AllOrders:=function(W)local dg;
  if IsSpets(W) then
    dg:=ReflectionDegrees(Group(W));
    return Set(Concatenation(List([1..Length(dg)],
      function(i)local q;
        q:=Denominator(AsRootOfUnity(PhiFactors(W)[i]));
	return Filtered(DivisorsInt(dg[i]*q),d->E(d)^dg[i]=PhiFactors(W)[i]);
      end)));
  else return Set(Concatenation(List(ReflectionDegrees(W),DivisorsInt)));
  fi;
end;

# Permutation of the conjugacy classes induced by an automorphism of W
PermutationOnClasses:=function(aut,W)
  return PermList(List(ConjugacyClasses(W),
     c->PositionClass(W,Representative(c)^aut)));
end;

# Permutation of the characters induced by an automorphism of W
# PermutationOnCharacters(aut,W [,charlist])
PermutationOnCharacters:=function(arg)local aut,W,p,ct;
  aut:=arg[1];W:=arg[2];
  p:=PermutationOnClasses(aut,W);
  ct:=CharTable(W).irreducibles;
  if Length(arg)=3 then ct:=ct{arg[3]};fi;
  return PermListList(ct,List(ct,r->Permuted(r,p)));
end;

# Permutation of the unipotent characters induced by an automorphism of W
# PermutationOnUnipotents(aut,W [,uniplist])
PermutationOnUnipotents:=function(arg)local aut,W,p,t,uc;
  aut:=arg[1];W:=arg[2];
  p:=PermutationOnClasses(aut,W);
  t:=TransposedMat(RwTable(W));
  if Length(arg)=3 then t:=t{arg[3]};fi;
  if Length(Set(t))<Length(t) then 
    uc:=UnipotentCharacters(W);
    t:=List(arg[3],x->Position(uc.harishChandra[1].charNumbers,x));
    if ForAll(t,x->x<>false) then return PermutationOnCharacters(aut,W,t);
    else Error("Rw cannot disambiguate\n");
    fi;
  fi;
  return PermListList(t,List(t,r->Permuted(r,p)));
end;

# s is a Set of tuples. Return E_1,...,E_n if
# s=List(Cartesian(E_1,...,E_n),Concatenation)
FactorsSet:=function(s)local i,j,s1,s2,r,a;
  if s=[[]] then return [];fi;
  for i in s[1] do
    s1:=Filtered(s,y->i in y);
    if Length(s1)=Length(s)/2 then
      r:=Difference(s[1],[i]);
      s2:=Filtered(s,y->not i in y);
      a:=PositionProperty(s2,y->IsSubset(y,r));
      if a<>false then j:=Difference(s2[a],r)[1];
	s1:=Set(List(s1,x->Difference(x,[i])));
	s2:=Set(List(s2,x->Difference(x,[j])));
	if s1=s2 then return Concatenation([[i,j]],FactorsSet(s1));fi;
      fi;
    fi;
  od;
  return [s];
end;

# Finds an element w of W which conjugates reflection subgroup H into
# a standard parabolic W_l if any exists; returns  [w,l]
StandardParabolic:=function(W,H)local wr,I,i,w,l,s,hr,w1,int,W1;
  hr:=H.rootInclusion{H.generatingReflections};
  if Length(hr)=0 then return [(),hr];fi;
  w1:=RepresentativeOperation(W,hr[1],
    W.orbitRepresentative[W.rootRestriction[hr[1]]]);
  if Length(hr)=1 then return [w1,[hr[1]^w1]];fi;
  wr:=W.rootInclusion{W.generatingReflections};
  if IsSubset(wr,hr) then return [(),hr];fi;
  H:=ReflectionSubgroup(W,OnTuples(H.rootInclusion{H.generatingReflections},w1));
  hr:=H.rootInclusion{H.generatingReflections};
  if IsSubset(wr,hr) then return [w1,hr];fi;
  s:=Size(H);
  I:=Combinations(wr,Length(hr));
# Print("\nfor ",ReflectionName(H),":|I|=",Length(I),"\c");
  I:=Filtered(I,l->Size(PermRootOps.ReflectionSubgroup(W,l))=s);
# Print(" correct size=>",Length(I)," left\n");
  for l in I do
    Print("trying ",l,"\c");
    if IsCoxeterGroup(W) then w:=RepresentativeOperation(W,hr,l,OnSets);
    else  w:=RepresentativeOperation(W,Set(Reflections(H)),
            Set(Reflections(ReflectionSubgroup(W,l))),OnSets);
    fi;
    if w<>false then 
      Print("found ",[w1*w,l]," in ",Stime(),"\n");
      return [w1*w,l];fi;
  od;
  Print("# ",hr," not conjugate to a standard parabolic\n");
  return [w1,hr];
end;

# find subsets of the set of vectors l of sum the vector S.
# in addition, lv is a vector of same length as l and the chosen subsets
# l{P} should statisfy lv{P}=v.
SubsetsSum:=function(S,l,v,lv)local c,inner,found,makell,sievev;
  sievev:=function(good,v)local i,p;
    for i in good do 
      p:=Position(v,lv[i]); if p=false then return [];fi;
      v:=v{Filtered([1..Length(v)],i->i<>p)};
    od;
    return v;
  end;
  makell:=function(nonsolved,t)local cands,i,j;
    cands:=List(nonsolved,x->[]);
    for j in t do for i in [1..Length(nonsolved)] do 
      if l[j][nonsolved[i]]<>0 then Add(cands[i],j);fi;
    od;od;
    return List([1..Length(nonsolved)],i->rec(pos:=nonsolved[i],cand:=cands[i]));
  end;
  c:=0;found:=0;
  # s= l{s} is known to be in subset at this stage
  # t= such that the problem is reduced to find subsets of l{t}
  # nonsolved= nonsolved colums of sum
  # v= remaining v to match
  inner:=function(S,s,t,nonsolved,v,factor)
    local ll,bad,good,p,sols,res,i,pb,sol,f,foo,llc,solved;
  # Print("#solved=",Length(l[1])-Length(nonsolved)," ",Join(s),"=>",Join(t),
  #       " v=",Join(List(Collected(v),x->Join(x,":"))," "),"\n");
    c:=c+1;
    if c mod 500=0 then
      Print(factor,":cols: ",Length(nonsolved)-1," set: ",Length(t),
            " found: ",found,"\n");
    fi;
    t:=Filtered(t,i->lv[i] in v);
    if Length(t)=0 then 
      if S=S*0 then 
        found:=found+1;
        return [[]];
      else 
      # Print(".\c");
        return [];
      fi;
    fi;
    ll:=makell(nonsolved,t);
    llc:=Filtered(ll,x->Length(x.cand)<=10);
    if Length(llc)=0 then llc:=[ll[1]];fi;
    SortParallel(List(llc,x->Length(x.cand)),llc);
    solved:=[];good:=[];bad:=[];
    for p in llc do
      p.sols:=Filtered(Combinations(p.cand),e->Sum(l{e}[p.pos])=S[p.pos]);
      if Length(p.sols)=0 then 
        # Print(",\c");
	return [];
      elif Length(p.sols)=1 then Add(solved,p.pos);
      fi;
      good:=Union(good,Intersection(p.sols));
      bad:=Union(bad,Difference(p.cand,Union(p.sols)));
    od;
    nonsolved:=Difference(nonsolved,solved);
    if Length(good)+Length(bad)>0 then
      # Print("*\c");
      pb:=inner(S-Sum(l{good}),Union(s,good),Difference(t,Union(good,bad)),
        nonsolved,sievev(good,v),factor);
      return List(pb,r->Concatenation(good,r));
    else
      res:=[];
      p:=Minimum(List(llc,x->Length(x.sols)));p:=First(llc,x->Length(x.sols)=p);
      f:=Length(p.sols);
      nonsolved:=Difference(nonsolved,[p.pos]);
      Print(factor,":cols: ",Length(nonsolved)," set: ",Length(t),
	 " in comb(",Length(p.cand),")=>",Length(p.sols),"\n");
#     if Length(p.sols)>15 then Error();fi;
      for sol in p.sols do
	good:=sol;
	bad:=Difference(p.cand,sol);
	if Intersection(good,bad)=[] then 
         pb:=inner(S-Sum(l{good}),Union(s,good),Difference(t,Union(good,bad)),
	   nonsolved,sievev(good,v),SPrint(factor,":",f));
	 Append(res,List(pb,r->Concatenation(good,r)));
	fi;
	f:=f-1;
      od;
      return res;
    fi;
  end;
  return inner(S,[],[1..Length(l)],[1..Length(S)],v,"");
end;
  
# given  e rationals  m and  e schur  elements sch  that we  know to  be
# of  the form  \prod_{j\ne  k}(1-E(e)^{s(k)-s(j)}q^{m_k-m_j}) for  some
# permutation s of  [1..e], find all permutations s  compatible with the
# data.  Each  perm  is  represented  modulo  a  translation  (can't  be
# detected). The result is a pair [v,G] where v is a vector of length
# nb. distinct sch and for each sch the corresponding
# v[i] holds the corresponding set of s(k).
# G is the inverse, ie. for k in [1..e] the entry G[k] is i such that 
# sch[k] corresponds to v[i].
FitParameter:=function(sch,m)
  local e,res,a,v,p,q,den,poss,pp,G,para,schurelm,positive,bad,good;
  positive:=p->ForAll(p.vcyc,x->x[2]>0); # tells if a CycPol is really a pol
  poss:=function(i)local v,k,avail,good,p,term;
  # for each element of Collected(m) tries to find a minimal
  # corresponding possible set of s(i)-s(k) for k such that m_k=m
    v:=sch[i];
    avail:=[1..Length(a)];p:=List(avail,k->[0..e-1]);good:=[];
    term:=function(j,k)return CycPol(1-E(e)^j*q^(den*(m[i]-a[k][1])));end;
    while true do
      for k in avail do 
        p[k]:=Filtered(Difference(p[k],Concatenation(p{good})),
                                   j->m[i]=a[k][1] or positive(v/term(j,k)));
      od;
      good:=Filtered(avail,i->Length(p[i])=a[i][2]);
      if Length(good)=0 then return p;fi;
      avail:=Difference(avail,good);
#     Print("good=",good,"\n");
      v:=v/Product(Filtered(good,k->m[i]<>a[k][1]),k->Product(p[k],j->term(j,k)));
    od;
  end;
  e:=Length(m); a:=Collected(m);q:=X(Cyclotomics);
  den:=Lcm(List(m,Denominator)); 
  sch:=List(sch,x->CycPolOps.DescentOfScalars(x,den));
  pp:=List([1..e],poss);
  p:=PositionProperty([1..e],i->List(a,x->x[2])=List(pp[i],Length));
  if p=false then return false;fi;
  v:=List([0..e-1],i->List(i+pp[p],x->Set(List(x,y->y mod e))));
  v:=List([1..e],k->Filtered([1..e],i->ForAll([1..Length(a)],j->IsSubset(
    pp[k][j],v[i][j]))));
  G:=List(sch,x->Position(sch,x));
  v:=List(Set(G),x->v[x]);
  G:=List(G,x->Position(Set(G),x));
  bad:=[1..Length(v)];
  repeat
    good:=Filtered(bad,i->Length(v[i])=Number([1..e],j->G[j]=i));
    bad:=Difference(bad,good);
    for p in bad do for a in good do v[p]:=Difference(v[p],v[a]);od;od;
  until Length(good)=0;
  if Sum(v,Length)>e then Diagnose("non-unique solution\n");return false;fi;
  schurelm:=para->List([1..e],k->Product(Filtered([1..e],j->j<>k),
    j->CycPol(1-para[k]/para[j])));
  para:=a->List([1..e],i->E(e)^a[i]*q^(den*m[i]));
  res:=List([1..Length(v)],x->[]);
  for p in [1..e] do Add(res[G[p]],p);od;
  pp:=[];for p in [1..Length(res)] do pp{res[p]}:=v[p];od;
  if schurelm(para(pp))<>sch then Diagnose("schur elms don't match\n");return false;fi;
  return [v,G];
end;

#---------------------- SeriesOps ---------------------------------------
# A d-Harish-Chandra series \CE(\BG,(\BL,\lambda) is a record with the 
# following fields
#  .d            d
#  .spets        \BG
#  .levi         \BL
#  .cuspidal     \lambda (index in UnipotentCharacters(.levi))
#  .WGL          W_\BG(\BL,\lambda)
#  .charNumbers  \CE(\BG,(\BL,\lambda) (indices in UnipotentCharacters(.spets))
#  .degRLG       The degree of R_\BL^\BG(\lambda)
#  .assignments  Possible bijections .charNumbers <-> Irr(.WGL)
#  .eps          For each \chi in .charNumbers sign of <\chi,R_\BL^\BG(\lambda)>
#  .dims         \chi(1) for \gamma_\chi

SeriesOps:=OperationsRecord("SeriesOps");

SeriesOps.String:=function(s)local res;
  res:=SPrint(s.d,"-series R^{",ReflectionName(s.spets),
                         "}_{",ReflectionName(s.levi),
             "}[c=",UnipotentCharacters(s.levi).charNames[s.cuspidal],"]");
  if IsBound(s.e) then PrintToString(res," W_G(L,c)=Z",s.e);
  elif IsBound(s.WGL)then 
    if IsBound(s.WGL.reflections) then
      PrintToString(res," W_G(L,c)=",ReflectionName(s.WGL));
    else
      PrintToString(res," |W_G(L,c)|=",Size(s.WGL));
    fi;
  elif IsBound(s.WGLdims) then 
      PrintToString(res," |W_G(L,c)|=",Sum(s.WGLdims,x->x^2));
  fi;
  if IsBound(s.translation) and s.translation<>s.e then
    PrintToString(res," translation=",s.translation);
  fi;
  return res;
end;

TeXSeries:=function(s)local res;
  res:=SPrint(s.d,"-series $R^{G=",ReflectionName(s.spets,rec(TeX:=1)),
              "}_{L=",ReflectionName(s.levi,rec(TeX:=1)),
   "}[\\lambda=",UnipotentCharacters(s.levi).TeXCharNames[s.cuspidal],"]");
  if IsBound(s.e) then PrintToString(res,"\\quad W_G(L,\\lambda)=Z_{",s.e,"}");
  elif IsBound(s.WGL)then 
    if IsBound(s.WGL.reflections) then
      PrintToString(res,"\\quad $W_G$(L,c)=",ReflectionName(s.WGL));
    else
      PrintToString(res,"\\quad |$W_G$(L,c)|=",Size(s.WGL));
    fi;
  elif IsBound(s.WGLdims) then 
      PrintToString(res,"\\quad |$W_G$(L,c)|=",Sum(s.WGLdims,x->x^2));
  fi;
  PrintToString(res,"$");
  if IsBound(s.translation) and s.translation<>s.e then
    PrintToString(res,"\\quad translation=",s.translation);
  fi;
  return res;
end;

SeriesOps.Print:=function(s)Print(String(s));end;

SeriesOps.Display:=function(s,opt)Print(Format(s));end;

SeriesOps.params:=function(s)local e;e:=Length(s.charNumbers);
  if not IsBound(s.assignment) then return false;fi;
  return List([1..e],function(i)local f,mod1;
  mod1:=function(x)x:=x-Int(x);if x<0 then x:=x+1;fi;return x;end;
    f:=mod1(-s.mC[i]/s.d+(s.assignment[i]-1)/e);
    return Mvp("q")^s.mC[i]*E(Denominator(f))^Numerator(f);
    end);
end;

SeriesOps.Format:=function(s,option)local m,f,uw,p,e,t;
  uw:=UnipotentCharacters(s.spets);
  p:=SortingPerm(s.assignment);
  e:=Length(s.charNumbers);
  f:=function(arg)local v,vals;
     if IsBound(option.TeX) then v:=[arg[1]];vals:=arg[Length(arg)];
     else v:=[arg[2]]; vals:=arg[3];fi;
     Append(v,Permuted(List(vals,x->Format(x,option)),p));Add(m,v);
  end;
  m:=[];
  f("N^o","No",s.charNumbers);
  f("\\hbox{Character}","Name",uw.charNames{s.charNumbers},uw.TeXCharNames{s.charNumbers});
# f("\\hbox{degree}","degree",CycPolUnipotentDegrees(s.spets){s.charNumbers});
  f("\\hbox{specializes}","specializes",s.assignment-1);
  f("\\hbox{a+A}","a+A",List(CycPolUnipotentDegrees(s.spets){s.charNumbers},d->
   d.valuation+Degree(d)));
  f("\\varepsilon","eps",s.eps);
  f("\\hbox{eigen}","eigen",s.eigen);
  for t in [0] do f("\\hbox{parameter}","param",SeriesOps.params(s));od;
  f("\\hbox{family \\#}","family #",
   List(s.charNumbers,j->PositionProperty(uw.families,k->j in k.charNumbers)));
  if s.eigen<>s.computedeigen then 
    f("\\hbox{computed/eig}","computed/eig",
      List([1..e],i->s.computedeigen[i]/s.eigen[i]));
    f("\\hbox{permutable}","permutable",List([1..e],i->
      PositionProperty(s.permutable,x->i in x)));
  fi;
  if IsBound(option.TeX) then
       return SPrint(TeXSeries(s),"\n",TeXTable(TransposedMat(m)));
  else return SPrint(s,"\n",StringTable(TransposedMat(m)));
  fi;
end;

SeriesOps.degRLG:=function(s)local q;
  if IsBound(s.degRLG) then return s.degRLG;fi;
  q:=X(Cyclotomics);
  s.degRLG:=GenericSign(s.spets)^-1*GenericOrder(s.spets,q)/
           (GenericSign(s.levi)^-1* GenericOrder(s.levi,q));
  s.degRLG.valuation:=0;
  s.degRLG:=CycPol(s.degRLG)*CycPolUnipotentDegrees(s.levi)[s.cuspidal];
  return s.degRLG;
end;

# Degree in q of the parameters (normalized so the smallest is 0)
SeriesOps.mC:=function(s)local D0,aA,uw,ul,pG,pL,xiL,xiG,h,cn;
  h:=HyperplaneOrbits(s.WGL);
  if Length(h)>1 then return;
  else h:=h[1].e_s;fi;
  uw:=UnipotentCharacters(s.spets);
  cn:=s.charNumbers{Filtered([1..Length(s.dims)],i->s.dims[i]=1)};
  aA:=uw.a{cn}+uw.A{cn};
  ul:=UnipotentCharacters(s.levi);
  if s.principal then # Id in L
    pG:=Sum(ReflectionDegrees(Group(s.spets))+ReflectionCoDegrees(Group(s.spets)));
    pL:=Sum(ReflectionDegrees(Group(s.levi))+ReflectionCoDegrees(Group(s.levi)));
    D0:=pG-pL;
    # compute xi_L
    xiL:=AsRootOfUnity(PhiOnDiscriminant(s.levi))*s.d;
    xiG:=AsRootOfUnity(PhiOnDiscriminant(s.spets))*s.d;
    D0:=D0+xiL-xiG;
    # Id in chars
    if Minimum(aA)<>0 then Error("id not in RLG(1)");fi;
    if not IsInt(D0/s.d) then 
      Diagnose(s,"\n    => (l(\pi_G)=",pG,")-(l(\pi_L)=",pL,")+(xiL*d=",
      xiL,")-(xiG*d=",xiG,") not =0(mod d=",s.d,
      ") Max(aA)-Min(aA)=",Maximum(aA)-Minimum(aA),"\n");
    fi;
  else
    # (JM+Gunter 18-3-2004) in any case we normalize so that
    # the smallest mC is 0 since the above choice gives unpleasant
    # results for G27
    D0:=Maximum(aA)-Minimum(aA); # just so that mC are positive
  fi;
  s.mC:=(D0+Minimum(aA)-aA)/h;
  return s.mC;
end;

SeriesOps.WGLdims:=function(s)local m;
  s.WGLdims:=List(CharTable(s.WGL).irreducibles,x->x[1]);
  if ForAll(s.WGLdims,x->x=1) then s.e:=Length(s.WGLdims);fi;
  return s.WGLdims;
end;

# Find W_G(L,\lambda)
SeriesOps.RelativeGroup:=function(s) 
  local unique,rH,r,c,H,rel,W,ud,eig,WF,L,rel,NLF,NLFb,N,hom,phi,V,m,func,
        mats,d,reflists;
  if IsBound(s.WGL) then return s.WGL;fi;
  W:=Group(s.spets);
  unique:=Set(List(Reflections(W),x->Position(Reflections(W),x)));
  mats:=[];
  mats{unique}:=List(unique,r->MatXPerm(W,Reflection(W,r)));
  rH:=[];
  if s.projector=s.projector^0 then #central series
    s.WGL:=Copy(Group(s.spets));
    s.WGL.mapping:=s.WGL.generators;
    s.WGL.reflists:=List(s.WGL.rootInclusion{s.WGL.generatingReflections},x->[x]);
    return s.WGL;
  fi;
  V:=NullspaceMat(s.projector^0-s.projector);  # The E(d)-eigenspace
  for r in unique do
    H:=NullspaceMat(List(PermRootOps.Coroot(Parent(W),W.rootInclusion[r]),x->[x]));
    H:=SumIntersectionMat(H,V)[2];
    if Length(H)<Length(V) then AddSet(rH,H);fi;
  od;
# Print("rH=",rH,"\n"); # Hyperplanes of W_G(L)
  rel:=[];
  for H in rH do
    if Length(H)=0 then 
      ConjugacyClasses(Group(s.levi)); # hack to fix a (bug in GAP?)
      Add(rel,rec(refs:=W.rootInclusion{W.generatingReflections},
             WH:=Normalizer(W,Group(s.levi))));
    else H:=Filtered(unique,r->H*mats[r]=H);
      H:=ReflectionSubgroup(W,W.rootInclusion{H});
      Add(rel,rec(refs:=H.rootInclusion{H.generatingReflections},
                  WH:=Normalizer(H,Group(s.levi))));
    fi;
  od;
  ud:=CycPolUnipotentDegrees(s.levi);
  eig:=Eigenvalues(UnipotentCharacters(s.levi),[1..Length(ud)]);
  ud:=Filtered([1..Length(ud)],i->ud[i]=ud[s.cuspidal]);
  ud:=Filtered(ud,i->eig[i]=eig[s.cuspidal]);
  if s.spets.phi=() then WF:=W;L:=Group(s.levi);
  else WF:=Group(Concatenation(W.generators,[s.spets.phi]),W.identity);
    L:=Subgroup(WF,Group(s.levi).generators);
  fi;
  if Size(L)=1 then
    s.WGL:=Centralizer(W,s.levi.phi);
    for H in rel do H.WH:=Centralizer(H.WH,s.levi.phi);od;
    func:=x->x;
  else
    NLF:=Normalizer(WF,L);
    if NLF=L then Error(s," N=L\n");return false;fi;
    NLFb:=FactorGroup(NLF,L);
    hom:=NaturalHomomorphism(NLF,NLFb);
    phi:=Image(hom,s.levi.phi);
    d:=ReflectionDegrees(W);
    s.WGLdegrees:=d{Filtered([1..Length(d)],
      x->E(s.d)^d[x]*PhiFactors(s.spets)[x]=1)};
    if SemiSimpleDRank(s.levi,s.d)=0 and Number(s.WGLdegrees,x->x<>1)=1 
    then
      if  s.levi.phi in W and Order(NLFb,phi)=Product(s.WGLdegrees) then 
        s.WGL:=Subgroup(NLFb,[phi]);
      else N:=Centralizer(Subgroup(WF,W.generators),s.levi.phi);
        if Size(N)=Size(L)*Product(s.WGLdegrees)
        then s.WGL:=FactorGroup(N,L);
        fi;
      fi;
    fi;
    if not IsBound(s.WGL) then  N:=Subgroup(WF,W.generators);
      N:=Subgroup(N,Intersection(N,NLF).generators);
      s.WGL:=Centralizer(FactorGroup(N,L),phi);
    fi;
    if Length(rel)=1 and rel[1].WH=NLF then rel[1].WH:=s.WGL;
    else rel:=List(rel,function(x)local N,WH;
      N:=Intersection(s.WGL,FactorGroup(Subgroup(NLF,x.WH.generators),L));
      if IsGroup(N) then x.WH:=N;
      else x.WH:=Subgroup(s.WGL,N);
      fi;
      return x;end);
    fi;
    func:=x->x.element.representative;
  fi;
  if Length(ud)>1 then
    Print("*** warning: need to disambiguate to compute WG(L)\n");
    s.WGL:=Stabilizer(s.WGL,Position(ud,s.cuspidal),function(c,g)
      return c^PermutationOnUnipotents(func(g),s.levi,ud);end);
    for H in rel do H.WH:=Intersection(H.WH,s.WGL);od;
    rel:=Filtered(rel,x->Size(x.WH)>1);
  fi;
  if not ForAll(rel,x->IsCyclic(x.WH)) then Error("a");fi;
  hom:=List(rel,x->First(Elements(x.WH),y->Order(x.WH,y)=Size(x.WH)));
  if Subgroup(s.WGL,hom)<>s.WGL then Error("b");fi;
  reflists:=List(rel,x->x.refs);
  m:=Concatenation(V,NullspaceMat(s.projector))^-1;
  rel:=List(hom,function(x)local M;M:=MatXPerm(W,func(x))^m;
    return M{[1..Length(V)]}{[1..Length(V)]};end);
  rel:=List(rel,AsReflection);
  for c in [1..Length(rel)] do
    r:=AsRootOfUnity(rel[c].eigenvalue);
    m:=Numerator(r);r:=Denominator(r);
    if m<>1 then
      rel[c]:=AsReflection(Reflection(rel[c].root,rel[c].coroot)^(1/m mod r));
    fi;
  od;
  s.WGL:=PermRootGroup(List(rel,x->x.root),List(rel,x->x.coroot));
  s.WGL.mapping:=hom;
  s.WGL.reflists:=reflists{List(s.WGL.roots{s.WGL.generatingReflections},
    x->PositionProperty(rel,y->ProportionalityCoefficient(y.root,x)<>false))};
  return s.WGL;
end;

# takes a d-series s with s.spets split; fills in s.charNumbers, s.eps, s.dims
SeriesOps.chars:=function(s)local ad,q,ud,RLG,v,eig,qdeg,f,c,t,cand,i,check;
  SeriesOps.WGLdims(s);
  q:=X(Cyclotomics);
  ud:=CycPolUnipotentDegrees(s.spets);
  ad:=SemiSimpleDRank(s.levi,s.d);
  s.charNumbers:=Filtered([1..Length(ud)],i->ad=Valuation(ud[i],s.d));
  SeriesOps.degRLG(s);
  RLG:=LusztigInduce(s.spets,UnipChar(s.levi,s.cuspidal));
  if RLG=false then Diagnose(s,":RLG failed\n");
  elif s.degRLG<>CycPol(Degree(RLG))then 
    Diagnose(s,":Deg RLG<>Sum(eps[i]*ud[i])\n");
  fi;
# now use that   S_\phi(q)=\eps_\phi Deg(RLG(\lambda))/Deg(\gamma_\phi)
  cand:=List(s.charNumbers,c->rec(charNumbers:=c,sch:=s.degRLG/ud[c]));
  cand:=Filtered(cand,c->ForAll(c.sch.vcyc,x->x[2]>0)); # positive
# Print("after deg cands=",List(cand,c->c.charNumbers),"\n");
  ad:=CycPol(q-E(s.d))^ad;
  v:=Value(s.degRLG/ad,E(s.d));
  for c in cand do 
    c.span:=Degree(c.sch)-c.sch.valuation;
# now use that 
    f:=Sum(s.WGLdims,x->x^2)*Value(ud[c.charNumbers]/ad,E(s.d))/v;
    c.dims:=AbsInt(f);c.eps:=SignInt(f);
  od;
  cand:=Filtered(cand,c->c.dims in s.WGLdims);
# Print("after eps cands=",List(cand,c->c.charNumbers),"\n");
  eig:=Eigenvalues(UnipotentCharacters(s.levi),[s.cuspidal])[1];
  eig:=eig*List([1..s.d^2],i->E(s.d^2)^i);
  cand:=Filtered(cand,c->Eigenvalues(UnipotentCharacters(s.spets),[c.charNumbers])[1]
     in eig);
# Print("after eig cands=",List(cand,c->c.charNumbers),"\n");
  if Length(cand)<Length(s.WGLdims) then 
    Diagnose(s,":not enough left with predicted eigenvalues in ",
        Format(eig),"\n");return false;
  fi;
  SortParallel(List(cand,x->x.dims),cand);Sort(s.WGLdims);
  for f in [ "charNumbers", "eps", "dims", "span"] do
    s.(f):=List(cand,c->c.(f));
  od;
  check:=function()
    if RLG<>false and (Filtered([1..Length(RLG.v)],
       i->RLG.v[i]<>0)<>Set(s.charNumbers)
       or RLG.v{s.charNumbers}<>Zip(s.dims,s.eps,function(x,y)return x*y;end))
    then Diagnose(s,":RLG does not match");
    fi;
  end;
  if Length(cand)=Length(s.WGLdims) then check();return s;fi;
  ud:=Zip(UnipotentDegrees(s.spets,q){s.charNumbers},s.eps,s.dims,
     function(x,y,z)return x*y*z;end);
  f:=Maximum(List(ud,Degree));
  ud:=List(ud,function(p)local c;
   return Concatenation([1..p.valuation]*0,p.coefficients,[1..f-Degree(p)]*0);
   end);
  qdeg:=Value(s.degRLG,q);
  qdeg:=Concatenation([1..qdeg.valuation]*0,qdeg.coefficients,[1..f-Degree(qdeg)]*0);

  v:=SubsetsSum(qdeg,ud,s.WGLdims,s.dims);
  Print("**** ",Length(v)," found\n");#Error();
  if Length(v)>10000 then
    Print(" after span ",Length(v)," combinations\n");
  fi;

  if Length(v)=0 then
     Diagnose("*** before span ",s," no candidates left\n");
     return false;
  fi;
  
  if IsBound(s.e) then
    SeriesOps.mC(s);
    if Length(v)>1 then
      v:=Filtered(v,a->s.span{a}=
        List(a,i->Sum(Difference(a,[i]),j->AbsInt(s.mC[i]-s.mC[j]))));
      if Length(v)>10000 then
        Print(" after span ",Length(v)," combinations\n");
      fi;
    fi;
  fi;

  if Length(v)=0 then
     Diagnose("*** after span ",s," no candidates left\n");
     return false;
  fi;

#  Print(" after span ",Length(v)," combinations\n");
  if Length(v)>1 then
    Print("**** Warning: using Mackey with tori\n");
    f:=FusionConjugacyClasses(s.levi,s.spets);
    c:=Zip(CharTable(s.spets).centralizers{f},CharTable(s.levi).centralizers,
      function(a,b)return a/b;end);
    t:=Zip(TransposedMat(RwTable(s.levi))[s.cuspidal],c,
      function(a,b)return a*b;end);
    t:=List([1..Length(ConjugacyClasses(s.spets))],
            i->Sum(t{Filtered([1..Length(f)],j->f[j]=i)}));
    c:=TransposedMat(RwTable(s.spets));
    v:=Filtered(v,a->Zip(s.dims{a},s.eps{a},
           function(x,y)return x*y;end)*c{s.charNumbers{a}}=t);
  fi;

#  Print(" after Mackey with tori ",Length(v)," combinations\n");
  if Length(v)>1 then
     Diagnose("*** ",s," ",
     Join(List(FactorsSet(List(v,x->s.charNumbers{x})),x->FormatGAP(x)),"x"),
       " candidates\n");
     return false;
  elif Length(v)=0 then
     Diagnose("*** ",s," no candidates left\n");
     return false;
  fi;
  for f in ["charNumbers","eps","dims","span"] do s.(f):=s.(f){v[1]};od;
  check();return s;
end;


SeriesOps.fill:=function(s)local ud,uw,Schur,r,G,p,ratio,LFrob,predeigen,v,
   series,noncus,unique,eig,cand,i,m,quality;
  if IsBound(s.e) then SeriesOps.mC(s);fi;
  uw:=UnipotentCharacters(s.spets);
  ud:=CycPolUnipotentDegrees(s.spets){s.charNumbers};
  Schur:=List([1..s.e],x->s.degRLG/ud[x]*s.eps[x]);
  s.eigen:=Eigenvalues(uw,s.charNumbers);
  # in this case eigen_j:=E(d)^{-m_j e/d} E(d)^j
  LFrob:=Eigenvalues(UnipotentCharacters(s.levi),[s.cuspidal])[1];
  m:=ReflectionDegrees(Group(s.spets));
  s.delta:=PhiFactors(s.spets){Filtered([1..Length(m)],i->m[i]<>1)};
  s.delta:=Lcm(List(s.delta,x->Denominator(AsRootOfUnity(x))));
  predeigen:=function(m,j)return E(s.d^2)^(s.delta*(s.d*(j-1)-m*s.e))*LFrob;end;
  r:=FitParameter(List(Schur,x->CycPolOps.EnnolaTwist(x,E(s.d))),s.mC);
  if r=false then Diagnose(s," FitParameter failed\n");return false;fi;
  v:=r[1];G:=r[2];cand:=List([1..Length(v)],i->Filtered([1..s.e],j->G[j]=i));
  series:=List(s.charNumbers,x->PositionProperty(uw.harishChandra,y->x in y.charNumbers));
  noncus:=Filtered([1..s.e],i->Length(uw.harishChandra[series[i]].charNumbers)>1);
  if Length(noncus)=0 then noncus:=[1..s.e];fi;
  unique:=Filtered([1..s.e],i->Length(v[G[i]])=1);
  ratio:=List(Intersection(noncus,unique),i->s.eigen[i]/predeigen(s.mC[i],v[G[i]][1]));
# Print("ratio=",ratio,"\n");
  if Length(Set(ratio))>1 then 
    Diagnose("eigenvalue ratios=",ratio,"\n");return false;fi;
  ratio:=AsRootOfUnity(ratio[1])*s.d;
  if not IsInt(ratio) then Diagnose(s,":ratio=",ratio,"\n");return false;fi;
  r:=QuotientMod(ratio,s.delta,s.d);
  if r=false then Diagnose("ratio=",ratio," delta=",s.delta," s.d=",s.d,"\n",
    s,":D0 is not divisible by ",s.d,"\n");
    s.mC:=s.mC+ratio*s.d/(s.e*s.delta);
    ratio:=List(Intersection(noncus,unique),
        i->s.eigen[i]/predeigen(s.mC[i],v[G[i]][1]));
    ratio:=AsRootOfUnity(ratio[1])*s.d;
    r:=QuotientMod(ratio,s.delta,s.d);
    if r=false then return false;fi;
  fi;
  v:=List(v+r-1,x->List(x,y->(y mod s.e)+1));
  eig:=[];for i in [1..s.e] do eig[G[i]]:=s.mC[i];od;
  r:=[];
  for i in [1..Length(v)] do
    p:=PositionProperty(Arrangements(v[i],Length(v[i])),
        a->s.eigen{cand[i]}=List(a,j->predeigen(eig[i],j)));
    if p=false then
      p:=Filtered(Arrangements(v[i],Length(v[i])),
        a->s.eigen{cand[i]}=-List(a,j->predeigen(eig[i],j)));
      if Length(p)>0 then r{cand[i]}:=p[1];
      else r{cand[i]}:=v[i];
      fi;
    else r{cand[i]}:=Arrangements(v[i],Length(v[i]))[p];
    fi;
  od;
  s.assignment:=r;
  s.computedeigen:=List([1..s.e],i->predeigen(s.mC[i],r[i]));
  if s.eigen<>s.computedeigen then
    v:=List([1..s.e],i->s.computedeigen[i]/s.eigen[i]);
    p:=Filtered([1..s.e],i->v[i]<>1);
    p:=StringTable([Concatenation(["char"],uw.charNames{s.charNumbers{p}}),
     Concatenation(["eig"],List(s.eigen{p},Format)),
     Concatenation(["computed"],s.computedeigen{p}),
     Concatenation(["computed/eig"],v{p})
     ]);
    Diagnose(s,": eigenvalue pb\n",p);
  fi;
  s.translation:=Filtered([0..s.e-1],t->s.computedeigen=
        List([1..s.e],i->predeigen(s.mC[i],(r[i]+t)mod s.e)));
  if Length(s.translation)=1 then s.translation:=s.e;
  else p:=s.translation[2];
    if ForAny(s.translation,x->x mod p<>0) or Length(s.translation)<>s.e/p
    then Error();
    fi;
    quality:=List(s.translation,t->List([1..s.e],i->
      (s.assignment[i]+t-1)/s.e-s.mC[i]/s.d));
    quality:=List(quality,y->List(y,x->(Numerator(x) mod Denominator(x))/
      Denominator(x)));
    quality:=List(quality,x->Sum(x,Denominator));
 #  Print("quality:",quality,"\n");
    m:=Position(quality,Minimum(quality));
    s.assignment:=List(s.assignment,x->((x+s.translation[m]-1)mod s.e)+1);
    s.translation:=p;
  fi;
  s.permutable:=Filtered(cand,x->Length(x)>1);
  s.Hecke:=Hecke(s.WGL,[SeriesOps.params(s)]);
  return s;
end;

Series:=function(spets,levi,cuspidal,d)local s;
  s:=rec(spets:=spets,levi:=levi,cuspidal:=cuspidal,d:=d,
   operations:=SeriesOps);
  s.principal:=UnipotentCharacters(s.levi).a[s.cuspidal]=0
               and UnipotentCharacters(s.levi).A[s.cuspidal]=0;
  return s;
end;

# CuspidalPairs(W[,d[,ad]]) returns (L,\lambda,d) for all proper 
# d-cuspidal pairs of WF [with eigenspace of dimension ad]
CuspidalPairs:=function(arg)
  local W,d,l,eig,ad,cl,c,w,V,I,HF,res,unique,H,w,WF,mats,s,f,P,char,cusps,frac;
  W:=arg[1];
  if IsSpets(W) then WF:=W;W:=Group(WF); else WF:=Spets(W); fi;
  if Length(arg)=1 then
    return Concatenation(List(AllOrders(W),d->CuspidalPairs(WF,d)));
  fi;
  d:=arg[2];
  if Length(arg)=2 then
    return Concatenation(List([1..DRank(WF,d)],ad->CuspidalPairs(WF,d,ad)));
  fi;
  ad:=arg[3];
  unique:=Filtered([1..Length(W.roots)],
                  i->Position(W.reflections,Reflection(W,i))=i);
  mats:=List(unique,i->MatXPerm(W,Reflection(W,i)));
  eig:=ReflectionEigenvalues(WF);
  if d=1 then frac:=0;else frac:=1/d;fi;
  cl:=Filtered([1..Length(eig)],j->Number(eig[j],e->frac=e)=ad);
  res:=[];
  while Length(cl)>0 do
    c:=cl[1];
    w:=Representative(ConjugacyClasses(WF)[c]);
    V:=NullspaceMat(MatXPerm(WF,w)-E(d)*IdentityMat(W.rank)); # E(d)-eigenspace
    I:=Filtered(unique,i->V=V*mats[Position(unique,i)]);
#     Print("######## W=",WF," d=",d," cl=",cl," c=",c," I=",I,"\n");
    if I=unique then cl:=Difference(cl,[c]);# we want proper pairs
    else HF:=SubSpets(WF,W.rootInclusion{I},w/WF.phi);
      if not IsRec(HF) then 
       Diagnose("SubSpets(",WF,",",W.rootInclusion{I},",class#",c,") failed\n");
	cl:=Difference(cl,[c]);
      else
	f:=Set(FusionConjugacyClasses(HF,WF));
	f:=Intersection(f,cl);
	if Length(f)=0 then Error("fusion is wrong");return false;fi;
	cl:=Difference(cl,f); # other classes in f give the same subspets
	#Print("hf=",HF,"\n");
	cusps:=CuspidalUnipotentCharacters(HF,d);
	Print("Eigspace=",FormatGAP(CycPol(X(Cyclotomics)-E(d))^ad),
	  " L=",ReflectionName(HF)," classes=",Join(f));
	if Length(cusps)=0 then Print(" no ",d,"-cuspidals\n");
	else 
	  Print(" ",d,"-cuspidals=",Join(UnipotentCharacters(HF).charNames{cusps}));
	  H:=Group(HF);
	  P:=StandardParabolic(W,H);
	  if P[1]<>() or P[2]<>H.rootInclusion{H.generatingReflections} then
	    HF:=SubSpets(WF,P[2],HF.phi^P[1]/WF.phi);
	    w:=w^P[1];
	    cusps:=CuspidalUnipotentCharacters(HF,d);
	    Print(" L=>",ReflectionName(HF),"\n");
	  else Print("\n");
	  fi;
  #  if ForAny(HF.type,t->IsBound(t.scalar) and not ForAll(t.scalar,x->x=1)) 
  #  then Diagnose(ReflectionName(HF),
  #		" Ennola-twisted: eigenvalues may be wrong\n");
  #  fi;
	  c:=Difference(eig[c],[frac]);
	  c:=List(c,x->E(Denominator(x))^Numerator(x));
	  f:=MatXPerm(WF,w);
	  if Length(c)=0 then f:=f^0;
	  else f:=Product(c,x->f-f^0*x)/Product(c,x->E(d)-x);
	  fi;
	  for char in cusps do
	    s:=Series(WF,HF,char,d);s.element:=w;s.projector:=f;Add(res,s);
	  od;
	fi;
      fi;
    fi;
  od;
  return res;
end;

CheckSeries:=function(arg)local W,l,s,c,e;
  W:=arg[1];
  if Length(arg)=2 then l:=CuspidalPairs(W,arg[2]);
  else l:=CuspidalPairs(W);
  fi;
  for s in l do
    Title("Check ",String(s));
    RelativeGroup(s);
    SeriesOps.WGLdims(s);
    if s.principal then
      c:=Value(GenericOrder(s.spets,X(Cyclotomics))/Sum(s.WGLdims,x->x^2)/
      GenericOrder(s.levi,X(Cyclotomics)),E(s.d));
      if c<>1 then  Diagnose(s," => |G|/(|L||WGL|) mod Phi=",c,"\n");fi;
      e:=GenericSign(s.spets)*E(s.d)^(Sum(ReflectionCoDegrees(Group(s.spets)))-
	 Sum(ReflectionCoDegrees(Group(s.levi))))/GenericSign(s.levi);
      if e<>1 then  Diagnose(s," => epsL/c mod Phi=",e,"\n");fi;
    fi;
    if SeriesOps.chars(s)<>false then
      SeriesOps.mC(s);
      if IsBound(s.e) then
	s:=SeriesOps.fill(s);
  #	if s<>false then Display(s);fi;
	Ok();
      elif CHECK.relativeSeries then SeriesOps.RelativeSeries(s);
      fi;
    fi;
  od;
  return l;
end;

findfractionalpowers:=function(W)local uc,h,L,cusp,sers,s,p,fix,i,n,mod1,
  reasons,frac;
  if W.nbGeneratingReflections=0 then return [0];fi;
  mod1:=function(x)x:=x-Int(x);if x<0 then x:=x+1;fi;return x;end;
  uc:=UnipotentCharacters(W);
  uc.fractions:=[];
  reasons:=List(uc.charNames,x->[]);
  for h in uc.harishChandra do
    if Length(h.levi)<>W.nbGeneratingReflections then
      L:=ReflectionSubgroup(W,h.levi);
      cusp:=FindCuspidalInLevi(h,L,W);
      n:=findfractionalpowers(L);
      if IsBound(n[cusp]) then
        uc.fractions{h.charNumbers}:=h.charNumbers*0+n[cusp];
      fi;
    fi;
  od;
  sers:=Set(Filtered(CheckSeries(W),x->IsBound(x.mC)));
  while Length(sers)<>0 do
  for s in sers do
    p:=PositionProperty(s.charNumbers,i->IsBound(uc.fractions[i]));
    if p=false then Print("reticent series",s,"\n");
    else
      frac:=s.mC[p]*s.e/s.d;
      fix:=mod1(uc.fractions[s.charNumbers[p]]-frac);
      if fix<>0 then
        Print(" **** Badly normalized series ", s,
	    " adding ",fix,"\n");
      fi;
      for i in [1..s.e] do
        frac:=mod1(s.mC[i]*s.e/s.d+fix);
        if not IsBound(uc.fractions[s.charNumbers[i]]) then
	  uc.fractions[s.charNumbers[i]]:=frac;
	  Add(reasons[s.charNumbers[i]],s);
        elif uc.fractions[s.charNumbers[i]]<>frac then
	  Print("Failed! ",s.charNumbers[i],"=",uc.charNames[s.charNumbers[i]]," in ",s,
	    "\n     where mC mod1=",frac,"\n conflicts with ",
	     reasons[s.charNumbers[i]],"\n");
	fi;
      od;
      sers:=Difference(sers,[s]);
    fi;
  od;
  od;
  return uc.fractions;
end;

CheckMaximalPairs:=function(W)local divs,l,s,d,c,e,res;
  divs:=AllOrders(W);
  res:=[];
  Title("FindMaximalPairs");
  for d in divs do
    l:=CuspidalPairs(W,d,DRank(W,d));
    if Length(l)=0 then Error("no maximal pair");fi;
    Ok();
    Append(res,l);
  od;
  for s in res do
    SeriesOps.RelativeGroup(s);
    e:=Elements(s.levi);
    ForAny(e,function(x)local Z;
      Z:=Centralizer(Group(s.spets),x);
      Z:=Group(Concatenation(Z.generators,Group(s.levi).generators),());
      if Size(Z)<>Size(s.WGL)*Size(s.levi) then Error();fi;
      return true;
      end);
  od;
  return res;
end;

SeriesOps.RelativeSeries:=function(s)local res,l,R,p,ud,u1,f,o,aA,tt;
  res:=[];
  s.relativeSpets:=[];
  for l in s.WGL.reflists do
    R:=ReflectionSubgroup(Group(s.spets),l);
    if not s.element in R then
      R:=SubSpets(s.spets,l,s.element/s.spets.phi);
      if R=false then return false;fi;
    fi;
    Add(s.relativeSpets,R);
    p:=CuspidalPairs(R,s.d);
    if p=false then return false;fi;
    p:=Filtered(p,x->Size(x.levi)=Size(s.levi) and x.cuspidal=s.cuspidal);
    if Length(p)<>1 then 
      Diagnose("found ",p," as ",s.d,"-cuspidal pairs in Spets",R,"\n");
      return false;
    fi;
    Append(res,p);
  od;
  tt:=CHECK.title;
  for p in res do
    RelativeGroup(p);SeriesOps.WGLdims(p);
    CHECK.title:=Concatenation(tt,String(p));
    if SeriesOps.chars(p)<>false then 
      SeriesOps.mC(p);SeriesOps.fill(p);fi;
  od;
  CHECK.title:=tt;
  p:=List(res,SeriesOps.params);
  if not false in p then s.Hecke:=Hecke(s.WGL,p);
    u1:=List(SchurElements(s.Hecke),
       x->s.degRLG/CycPol(ScalMvp(Value(Mvp(x),["q",X(Cyclotomics)]))));
    ud:=List(CycPolUnipotentDegrees(s.spets){s.charNumbers},
     x->x*SignInt(Value(x/CycPolUnipotentDegrees(s.levi)[s.cuspidal],E(s.d))));
    p:=PermListList(u1,ud);
    if p=false then Error(s.Hecke," wrong set of SchurElements");fi;
    for f in ["span","charNumbers"] do s.(f):=Permuted(s.(f),p);od;
    aA:=List(u1,x->E(s.d^2)^(x.valuation+Degree(x)));
    p:=PositionRegularClass(s.WGL,s.d);
    o:=List(CharTable(s.WGL).irreducibles,x->x[p]/x[1]);
    s.predictedEigen:=List([1..Length(o)],i->aA[i]*o[i]);
  fi;
  return res;
end;

Checkzegen:=function(W)local l,uc,aA,e,z,s,mod1,ucL,aAL,Conc;
  mod1:=function(x)x:=x-Int(x);if x<0 then x:=x+1;fi;return x;end;
  Conc:=function(a,b)return Concatenation([a],b);end;
  l:=CheckSeries(W);
# for s in l do SeriesOps.RelativeSeries(s);od;
  Print("with  no Hecke:",Filtered(l,s->not IsBound(s.Hecke)),"\n");
  Print("    center=",OrderCenter(W),"\n");
  l:=Filtered(l,s->IsBound(s.Hecke));
  uc:=UnipotentCharacters(W);
  aA:=uc.a+uc.A;
  for s in l do
    e:=gete(s.Hecke);
    z:=OrderCenter(s.WGL);
    ucL:=UnipotentCharacters(s.levi);
    aAL:=ucL.A+ucL.a;aAL:=aAL[s.cuspidal];
    Print(s,"\n");
    Print(StringTable([Conc("(aA-aAL)/z",List((aA{s.charNumbers}-aAL)/z,mod1)),
          Conc("aA/z",List(aA{s.charNumbers}/z,mod1)),
          Conc("local Hecke irr",List(e,x->mod1(1/x)))]));
  od;
end;
