#############################################################################
##
#A  global.g       VKCURVE package         Jean Michel
##
#Y  Copyright (C) 2001 - 2002  University Paris VII, France.
##
##  This file holds the controlling functions which calls the various
##  steps of the Van Kampen algorithm.
## 
#############################################################################

# VKCURVE.Loops, si les champs .loops, .segments, .points ne sont pas remplis
# appelle LoopsAroundPunctures qui les remplit en construisant ainsi une
# serie de lacets autour des points r.roots.
# .loops --  une liste qui decrit ces lacets par des listes d'indices d'elements#            de r.segments (les indices representent  des segments orientes, un
#            indice negatif represente un segment parcouru a l'envers).
# .segments -- une liste de paires d'indices dans r.points.
# .points -- une liste de complexes decimaux.
#
VKCURVE.Loops:=function(r)local m,i,P,l,segmentNumbers,pointNumbers;
  Inherit(r,LoopsAroundPunctures(r.roots));
  # here we  have loops around the  'true' roots and around  the 'extra'
  # roots Difference(r.roots,r.trueroots). We get rid of the extra loops
  # and the associated segments and points, first saving the basepoint.
  # (its location is known now, and maybe not later? J.M.) 
  if r.loops[1][1]<0 then r.basepoint:=r.segments[-r.loops[1][1]][2];
                     else r.basepoint:=r.segments[r.loops[1][1]][1];
  fi;
  if r.ismonic
      then r.loops:=r.loops{List(r.roots,z->Position(r.roots,z))};
      else r.loops:=r.loops{List(r.trueroots,z->Position(r.roots,z))};
  fi;
  segmentNumbers:=Set(Concatenation(List(r.loops,x->List(x,AbsInt))));
  r.loops:=List(r.loops,x->List(x,function(y)
    if y<0 then return -Position(segmentNumbers,-y);
           else return Position(segmentNumbers,y);
    fi;
    end));
  r.segments:=r.segments{segmentNumbers};
  pointNumbers:=Set(Concatenation(r.segments));
  r.segments:=List(r.segments,x->List(x,y->Position(pointNumbers,y)));
  r.points:=r.points{pointNumbers};
  r.basepoint:=Position(pointNumbers,r.basepoint);
  if VKCURVE.showSegments then
    Print("# There are ",Length(r.segments),
          " segments in ",Length(r.loops)," loops\n");
  fi;
  if VKCURVE.showWorst then
    l:=List([1..Length(r.segments)],function(i)local m,s;
     s:=r.segments[i];
     m:=MinPos(List(r.roots,z->DistSeg(z,r.points[s[1]],r.points[s[2]])));
     return [m[1],i,m[2]];
    end);
    Sort(l);
    Print("worst segments:\n");
    for i in [1..Minimum(5,Length(l))] do
      P:=l[i];
      Print("segment ",P[2],"=",r.segments[P[2]]," dist to ",P[3],
	     "-th root is ",P[1],"\n");
    od;
  fi;
# find the minimum distance m betwenn two roots
  if Length(r.roots)>1 then
    m:=Dispersal(r.roots);
    if VKCURVE.showRoots then
      Print("\nMinimum distance=",evalf(m[1])," between roots ",m[2][1]," and ",
	 m[2][2]," of discriminant\n");
    fi;
    r.dispersal:=m[1];
  else
    r.dispersal:=1/1000;
  fi;
# and round points to m/100
  m:=-DecimalLog(DecimalToRat(r.dispersal)/100);
  r.points:=List(r.points,y->evalf(y,m));
end;

# si le champ zeros n'est pas rempli le remplit (.points doit etre rempli)
VKCURVE.Zeros:=function(r)local i,min,m;
  if VKCURVE.showRoots then
    Print("Computing zeros of curve at the ",Length(r.points),
          " segment extremities...\n"); 
  fi;
  r.zeros:=[];min:=[];
  for i in [1..Length(r.points)] do
    if VKCURVE.showZeros then Print("<",i,"/",Length(r.points),">");fi;
    r.zeros[i]:=SeparateRoots(Value(r.curve,
			 ["y",ComplexToComplexRat(r.points[i])]),1000);
    if Length(r.zeros[i])>1 then
      m:=Dispersal(r.zeros[i]);
      m[1]:=evalf(m[1],-DecimalLog(m[1]/1000));
      if VKCURVE.showZeros then Print(" d=",m,"\n");fi;
      min[i]:=[m[1],i];
    fi;
  od;
  if VKCURVE.showWorst and Length(r.zeros[1])>1 then
    Sort(min);
    Print("worst points:\n");
    for i in [1..Minimum(5,Length(min))] do
      Print(min[i][2],": mindist(zeros)=",min[i][1],"\n");
    od;
  fi;
end;

VKCURVE.PrepareCurve:=function(curve)local d;
#  This makes sure  r.curve is an Mvp, whose  coefficients are rationals
#  or gaussians, and which is quadratfrei
  if IsPolynomial(curve) then 
    curve.coefficients:=List(curve.coefficients,z->Value(z,Mvp("x")));
    curve:=Value(curve,Mvp("y"));
  fi;
  if ForAny(curve.coeff,IsComplex) then
    curve.coeff:=List(curve.coeff,x->ComplexRatToGaussian(Complex(x)));
  fi;
  d:=Discy(curve);
  if d=0*d then 
    Print("**** Warning: curve is not quadratfrei\c");
    d:=MvpGcd(curve,Derivative(curve,"y"));
    Print(": dividing by ",d,"\n");
    curve:=curve/d;
  fi;
  return rec(curve:=curve);
end;

VKCURVE.Discy:=function(r)local d,l,F,gcd,i,common;
# Here r.curve  is considered as  a polynomial in x  (whose coefficients
# are polynomials in y).
# The zeros of the discriminant  of this polynomial are computed. First,
# d=Gcd of  the coefficients is computed,  and if its zeros  are stored.
# Then  disc=discriminant  of the  r.curve/d  is  computed. It  is  made
# 'quadratfrei'  (disc=disc/Gcd(disc,disc'))  and  stripped  of  factors
# common with d  and then factored (if possible which  in GAP happens if
# it is  a polynomial over  the rationals (not gaussian  rationals), and
# stored in r.discy as a list of polynomials in x.
# Some of  these computations may be  too costly for GAP,  in which case
# one has a better hope to complete them in MAPLE.

# write curve as \sum_i x^i l_i(y)
  l:=List(Coefficients(r.curve,"x"));
  gcd:=l[1];for i in [2..Length(l)] do gcd:=MvpGcd(l[i],gcd);od;
  # gcd is here the Gcd of the l_i(y)
  l:=ScalMvp(Coefficients(gcd,"y"));
  r.curveVerticalPart:=Polynomial(DefaultField(l),l);
  if VKCURVE.showRoots and Degree(r.curveVerticalPart)>0 then
    Print("Curve has ",Degree(r.curveVerticalPart)," linear factors in y\n"); 
  fi;
  r.nonVerticalPart:=r.curve/gcd;
  d:=Discy(r.nonVerticalPart); # an Mvp in x
  # d=Discriminant of the curve minus vertical components
  if d=0*d then Error("Discriminant is 0:  should not happen"); 
   # curve should be quadratfrei here
  fi;
  if VKCURVE.showRoots then
    Print("Discriminant has ",Degree(d)," roots, "); 
  fi;
  d:=d/MvpGcd(d,Derivative(d,"x")); # make sure roots have multiplicity 1
  if VKCURVE.showRoots then
    Print(" of which ",Degree(d)," are distinct\n"); 
  fi;
  common:=MvpGcd(d,Value(gcd,["y",Mvp("x")]));
  if VKCURVE.showRoots and Degree(common)>0 then
    Print(" and of which ",Degree(common)," are roots of linear factors\n"); 
  fi;
  d:=d/common; # make sure roots contain no vertical component
  d:=d/d.coeff[Length(d.coeff)];
  l:=ScalMvp(Coefficients(d,"x"));
  F:=DefaultField(l);
  r.discy:=Polynomial(F,l);
  if F=GaussianRationals then
    d.coeff:=List(d.coeff,Complex);
    r.discyFactored:=[d];
  else r.discyFactored:=List(Factors(r.discy),x->Value(x,Mvp("x")));
  fi;
end;

VKCURVE.GetDiscyRoots:=function(r)local m;
  if VKCURVE.showRoots then
    Print("Computing roots of discriminant...\n"); 
  fi;
  if Degree(r.curveVerticalPart)>0 then 
       r.roots:=SeparateRoots(Value(r.curveVerticalPart,Mvp("x")),1000);
  else r.roots:=[];
  fi;
  Append(r.roots,Concatenation(List(r.discyFactored,p->SeparateRoots(p,1000))));
end;

VKCURVE.Braids:=function(r)local i,l,bb,pr;
  if VKCURVE.showgetbraid then pr:=Print;else pr:=Ignore;fi;
  pr("# Computing monodromy braids\n");
  r.braids:=[];
  for i in [1..Length(r.loops)] do
    l:=Filtered(List(r.loops[i],AbsInt),s->not IsBound(r.monodromy[s]));
    if Length(l)>0 then 
      pr("# loop[",i,"] missing segments ",l,"\n");
    else
      bb:=Product(r.loops[i],function(s)if s<0 then return r.monodromy[-s]^-1;
	                                    else return r.monodromy[s];fi;end);
      pr("# loop[",i,"]=",bb,"\n");
      Add(r.braids,bb);
     fi;
  od;
  if VKCURVE.shrinkBraid then
    r.rawBraids:=r.braids;
    r.braids:=ShrinkGarsideGeneratingSet(r.braids);
  fi;
end;

Segment:=function(r,segno)local pr,tm;
  if IsBound(r.name) then 
    PrintTo(SPrint(r.name,".",segno),"");
    pr:=function(arg)
      ApplyFunc(AppendTo,Concatenation([SPrint(r.name,".",segno)],arg));
      ApplyFunc(Print,arg);
    end;
  elif VKCURVE.showSegments then pr:=Print;
  else pr:=Ignore;
  fi;
  tm:=Runtime();
  if VKCURVE.monodromyApprox then
    r.monodromy[segno]:=ApproxFollowMonodromy(r,segno,pr);
  else
    r.monodromy[segno]:=FollowMonodromy(r,segno,pr);
  fi;
  tm:=Runtime()-tm;
  if IsBound(r.name) then pr(r.name,".");fi;
  pr("monodromy[",segno,"]:=");
  if IsBound(r.name) then pr(r.name,".");fi;
  pr(Format(r.monodromy[segno],rec(GAP:=1,Fraction:=1)),";\n");
  pr("# segment ",segno,"/",Length(r.segments),
     " Time=",evalf(tm/1000,1),"sec\n");
end;

VKCURVE.SetPrintLevel:=function(printlevel)
  if printlevel=0 then
    VKCURVE.showSingularProj:=false;
    VKCURVE.showBraiding:=false;
    VKCURVE.showLoops:=false;
    VKCURVE.showAction:=false;
    VKCURVE.showSegments:=false;
    VKCURVE.showInsideSegments:=false;
    VKCURVE.showWorst:=false;
    VKCURVE.showZeros:=false;
    VKCURVE.showNewton:=false;
    VKCURVE.showgetbraid:=false;
    VKCURVE.showRoots:=false;
  elif printlevel=1 then
    VKCURVE.showSegments:=true;
    VKCURVE.showgetbraid:=true;
    VKCURVE.showSingularProj:=false;
    VKCURVE.showBraiding:=false;
    VKCURVE.showLoops:=false;
    VKCURVE.showAction:=false;
    VKCURVE.showInsideSegments:=false;
    VKCURVE.showWorst:=false;
    VKCURVE.showZeros:=false;
    VKCURVE.showNewton:=false;
    VKCURVE.showRoots:=false;
  elif printlevel=2 then
    VKCURVE.showSingularProj:=true;
    VKCURVE.showBraiding:=true;
    VKCURVE.showLoops:=true;
    VKCURVE.showAction:=true;
    VKCURVE.showSegments:=true;
    VKCURVE.showInsideSegments:=true;
    VKCURVE.showWorst:=false;
    VKCURVE.showZeros:=true;
    VKCURVE.showNewton:=true;
    VKCURVE.showgetbraid:=true;
    VKCURVE.showRoots:=true;
  fi;
end;

VKCURVE.TrivialCase:=function(r)
  r.presentation:=PresentationFpGroup(FreeGroup(Degree(r.curve)));
  r.operations:=rec(Print:=function(r)DisplayPresentation(r.presentation);end);
  return r;
end;

# Segments(name[,range])
# calcule les fichiers decrivant la monodromie le long des segments de
# r.segments (tous les segments si un seul argument, ceux donnes
# par range si 2 arguments).
# si r contient un champ .name alors les segments sont ecrits sur un fichier
# commencant par ce nom, sinon sont mis dans r.monodromy
# if r does not have the right fields then one tires to read (r.name).tmp
Segments:=function(arg)local r,range,segno,tm,pr;
  r:=arg[1];
  Read(Concatenation(r.name,".tmp"));
  r.B:=Braid(CoxeterGroupSymmetricGroup(Length(r.zeros[1])));
  if not IsBound(r.monodromy) then r.monodromy:=[];fi;
  if Length(arg)>1 and IsList(arg[2]) then range:=arg[2];
  else range:=[1..Length(r.segments)];
  fi;
  for segno in range do Segment(r,segno);od;
end;

VKCURVE.Finish:=function(r)
  VKCURVE.Braids(r);
  if r.ismonic then r.rawPresentation:=VKQuotient(r.braids);
               else r.rawPresentation:=DBVKQuotient(r);
  fi;
  r.presentation:=PresentationFpGroup(r.rawPresentation);
  ShrinkPresentation(r.presentation);
  r.rawPresentation:=PresentationFpGroup(r.rawPresentation);
  r.operations:=rec(Print:=function(r)DisplayPresentation(r.presentation);end);
  return r;
end;

VKCURVE.SearchHorizontal:=function(r0)local height,intersection,r;
  # Searching for a good horizontal
  height:=9;
  repeat height:=height+1;
    intersection:=ScalMvp(Coefficients(Value(r0.curve,["x",height]),"y"));
    intersection:=Polynomial(DefaultField(intersection),intersection);
    intersection:=intersection/Gcd(Derivative(intersection),intersection);
  until  Degree(intersection)=Length(Coefficients(r0.curve,"y"))-1
     and Degree(Gcd(r0.discy,intersection))=0;
  intersection:=intersection/Gcd(intersection,r0.curveVerticalPart);
  Print("Curve is not monic -- Trivializing along the horizontal line x = ",
          height,"\n");
  r:=rec(height:=height,input:=r0.curve,curve:=r0.curve*(x-height));
  if IsBound(r0.name) then r.name:=r0.name;fi;
  VKCURVE.Discy(r);
  # set trueroots  to roots  of Discy  which do  not only  correspond to
  # intersections of the curve with the chosen horizontal
  r.trueroots:=r0.roots;
  r.verticallines:=r0.roots{[1..Degree(r0.curveVerticalPart)]};
  r.roots:=ShallowCopy(r.trueroots);
  r.ismonic:=false;
  Append(r.roots,SeparateRoots(Value(intersection,Mvp("x")),1000));
  return r;
end;

# curve --  un Mvp en x et y "monique en x" decrivant une courbe P(x,y)
FundamentalGroup:=function(arg)local r,i;#height,intersection;
  if Length(arg)=2 then VKCURVE.SetPrintLevel(arg[2]);
                   else VKCURVE.SetPrintLevel(0);fi;
  r:=VKCURVE.PrepareCurve(arg[1]);
  VKCURVE.Discy(r);VKCURVE.GetDiscyRoots(r);
  if Length(r.roots)=0 then return VKCURVE.TrivialCase(r);fi;
  r.ismonic:=Length(Coefficients(Reversed(Coefficients(r.curve,"x"))[1],"y"))=1;
  if not r.ismonic then r:=VKCURVE.SearchHorizontal(r); fi;
  VKCURVE.Loops(r); VKCURVE.Zeros(r);
  if Length(r.zeros[1])=0 then return VKCURVE.TrivialCase(r);fi;
  r.B:=Braid(CoxeterGroupSymmetricGroup(Length(r.zeros[1])));
  r.monodromy:=[];
  for i in [1..Length(r.segments)] do Segment(r,i);od;
  return VKCURVE.Finish(r);
end;

#Garanties sur LoopsAroundPunctures:
#Pour un ensemble Z de zeros et z dans Z, on definit R(z)=1/2 dist(z,Z-z).
#L'input de LoopsAroundPunctures est un ensemble Z de zeros approches de r.discy
#tels que pour tout z il y ait un zero plus pres que R(z)/S ou S est une 
#constante globale du programme (en pratique nous prendrons sans doute S=100).
# Soit d=inf_{z in Z}(R(z))
# en output on retourne des points de denominateur 10^-k ou 10^-k<d/S'
# (en pratique on prendra S'=100) et tels que la distance d'un segment a
# un zero de r.discy soit garantie >= d-d/S'-d/S

# curve --  un Mvp en x et y "monique en x" decrivant une courbe P(x,y)
PrepareFundamentalGroup:=function(curve,name)local r,i,f;
  r:=VKCURVE.PrepareCurve(curve);r.name:=name;
  VKCURVE.Discy(r);VKCURVE.GetDiscyRoots(r);
  if Length(r.roots)=0 then return VKCURVE.TrivialCase(r);fi;
  r.ismonic:=Length(Coefficients(Reversed(Coefficients(r.curve,"x"))[1],"y"))=1;
  if not r.ismonic then r:=VKCURVE.SearchHorizontal(r); fi;
  VKCURVE.Loops(r);VKCURVE.Zeros(r);
  if Length(r.zeros[1])=0 then return VKCURVE.TrivialCase(r);fi;
  PrintTo(Concatenation(r.name,".tmp"),"");
  for f in ["curve","discy","roots","loops","segments","points","zeros"] do
    Cut(SPrint(r.name,".",f,":=",Format(r.(f),rec(GAP:=1)),";\n"),SizeScreen()[1],
      ",+*",Concatenation(r.name,".tmp"));
  od;
  Print("    ----------------------------------\n");
  Print("Data saved in ",r.name,".tmp\n");
  Print("You can now compute segments 1 to ",Length(r.segments),"\n");
  Print("in different GAP sessions by doing in each of them:\n");
  Print("    ",r.name,":=rec(name:=\"",r.name,"\");\n");
  Print("    Segments(",r.name,",[1..",Length(r.segments),"]);\n");
  Print("(or some other range depending on the session)\n");
  Print("Then when all files ",r.name,".xx have been computed finish by\n");
  Print("    ",r.name,":=rec(name:=\"",r.name,"\");\n");
  Print("    FinishFundamentalGroup(",r.name,");\n");
end;

FinishFundamentalGroup:=function(r)local i;
  Read(Concatenation(r.name,".tmp"));
  if not IsBound(r.monodromy) then r.monodromy:=[];fi;
  r.B:=Braid(CoxeterGroupSymmetricGroup(Length(r.zeros[1])));
  for i in [1..Length(r.segments)] do
    Read(SPrint(r.name,".",i));
   if not IsBound(r.monodromy[i]) then Print("***** ",i," missing ****\n");fi;
  od;
  return VKCURVE.Finish(r);
end;
